package it.unibz.inf.isochrone;

import it.unibz.inf.isochrone.algorithm.AbstractAlgorithm;
import it.unibz.inf.isochrone.algorithm.AlgorithmException;
import it.unibz.inf.isochrone.algorithm.AlgorithmFactory;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.MemoryOutput;
import it.unibz.inf.isochrone.util.BigDecimalOptionHandler;
import it.unibz.inf.isochrone.util.LocalDateTimeOptionHandler;
import it.unibz.inf.isochrone.util.QueryPointOptionHandler;

import java.awt.geom.Point2D;
import java.io.ByteArrayOutputStream;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.time.Month;
import java.util.List;
import java.util.Set;

import javax.validation.ConstraintViolation;
import javax.validation.Validation;
import javax.validation.Validator;
import javax.validation.constraints.DecimalMin;
import javax.validation.constraints.Min;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.kohsuke.args4j.CmdLineException;
import org.kohsuke.args4j.CmdLineParser;
import org.kohsuke.args4j.Option;

public final class AlgorithmRunner {
	private static final Logger LOGGER = LogManager.getLogger(AlgorithmRunner.class.getName());
	// CHECKSTYLE:OFF MagicNumber
	@Option(name = "-a", usage = "name of the algorithm")
	private AlgorithmFactory algorithmName = AlgorithmFactory.MINE;
	@Option(name = "-d", usage = "the direction of the isochrone")
	private Direction direction = Direction.INCOMING;
	@Option(name = "-ds", usage = "name of the dataset to be used")
	private String datasetName = "bz";
	@Option(name = "-dt", handler = LocalDateTimeOptionHandler.class, usage = "the date and time to start the computation for (in format \"yyyy-MM-ddTHH:mm\")")
	private LocalDateTime dateTime = LocalDateTime.of(2011, Month.JULY, 1, 12, 0, 0);
	@Min(1L)
	@Option(name = "-dmax", usage = "the isochrones runtime (in seconds; minimal valid value is 1 second) ")
	private Long dmax = 910L;
	@Option(name = "-h", usage = "prints this help")
	private boolean showHelp = false;
	@Min(0L)
	@Option(name = "-mm", usage = "the number of nodes that are allowed to be in memory. Only useful for MINER/MINERX algorithm. 0 value means \"unlimited\".")
	private Long maxMemorySize = 0L;
	@Option(name = "-m", usage = "the network mode to use for calculation")
	private Mode mode = Mode.MULTIMODAL;
	@Option(name = "-q", handler = QueryPointOptionHandler.class, usage = "node to start the calculation from (format \"lng,lat\").")
	private Point2D.Double queryPoint = new Point2D.Double(11.3525539631585d, 46.4975205d);
	@DecimalMin("0.1")
	@Option(name = "-ws", handler = BigDecimalOptionHandler.class, usage = "the walking speed in km/h (minimal valid value is 0.1 km/h)")
	private BigDecimal walkingSpeed = BigDecimal.valueOf(4.5f);
	private Throwable error;
	private String usage;
	// CHECKSTYLE:ON MagicNumber

	// Constructor

	protected AlgorithmRunner(final String[] args) {
		final CmdLineParser parser = new CmdLineParser(this);
		try {
			parser.parseArgument(args);
		} catch (final CmdLineException e) {
			LOGGER.error("Invalid usage: {}", e.getMessage());
			error = e;
		}

		final Validator validator = Validation.buildDefaultValidatorFactory().getValidator();
		final Set<ConstraintViolation<AlgorithmRunner>> validationErrors = validator.validate(this);
		if (!validationErrors.isEmpty()) {
			final String msg = "Invalid usage: parameter validation failed";
			LOGGER.error(msg);
			for (final ConstraintViolation<AlgorithmRunner> constraintViolation : validationErrors) {
				LOGGER.error(" - {}", constraintViolation.getMessage());
			}
			error = new IllegalArgumentException(msg);
		}

		if (showHelp) {
			final ByteArrayOutputStream out = new ByteArrayOutputStream();
			parser.printUsage(out);

			try {
				usage = out.toString("UTF-8");
			} catch (final UnsupportedEncodingException e) {
				usage = "Usage information could not be calculated";
			}
		}
	}

	// Public static void main method

	public static void main(final String[] args) {
		final AlgorithmRunner runner = new AlgorithmRunner(args);
		if (!runner.isValid() || runner.showHelp) {
			runner.printHelp();
			return;
		}
		
		runner.run();
	}

	// Protected methods

	protected Throwable getError() {
		return error;
	}

	protected boolean isValid() {
		return error == null;
	}

	// Private methods

	private void printHelp() {
		System.out.println(usage);
	}

	private void run() {
		final ConfigDataset dataset = ConfigDataset.getInstance(datasetName);
		if (dataset == null) {
			throw new IllegalArgumentException("Dataset could not be created from parameters");
		}

		final ConfigAlgorithmBuilder configBuilder = new ConfigAlgorithmBuilder(queryPoint)
			.dateTime(dateTime)
			.direction(direction)
			.duration(dmax)
			.mode(mode)
			.walkingSpeed(walkingSpeed);

		final ConfigAlgorithm algorithmConfig = configBuilder.build();
		try (final Database database = new Database(dataset, algorithmConfig.getMode(), algorithmConfig.getDirection())) {
			final long l0 = System.currentTimeMillis();
			final AbstractAlgorithm algorithm = algorithmName.getInstance(database, algorithmConfig, maxMemorySize);
			final MemoryOutput output = algorithm.getResult();
			LOGGER.info("Calculation time (in msec): {}", (System.currentTimeMillis() - l0));
			LOGGER.info("#links: {}", output.getLinks().size());
			LOGGER.info("#nodes: {}", output.getNodes().size());
			final List<Long> coordinateTimes = database.getCoordinateTime;
			if(coordinateTimes.size() > 0){
				long c = 0;
				for(long l : coordinateTimes){
					c += l;
				}
				LOGGER.info("getCoordinate Time: first: " + coordinateTimes.get(0) + ", average: " + c/coordinateTimes.size());
			}
			final List<Long> calendarTimes = database.getCalendarTime;
			if(calendarTimes.size() > 0){
				long c = 0;
				for(long l : calendarTimes){
					c += l;
				}
				LOGGER.info("getCalendar Time: first: " + calendarTimes.get(0) + ", average: " + c/calendarTimes.size());
			}
			final List<Long> linkTimes = database.getLinkTime;
			if(linkTimes.size() > 0){
				long c = 0;
				for(long l : linkTimes){
					c += l;
				}
				LOGGER.info("getLink Time: first: " + linkTimes.get(0) + ", average: " + c/linkTimes.size());
			}
			final List<Long> linksTimes = database.getLinksTime;
			if(linksTimes.size() > 0){
				long c = 0;
				for(long l : linksTimes){
					c += l;
				}
				LOGGER.info("getLinks Time: first: " + linksTimes.get(0) + ", average: " + c/linksTimes.size());
			}
			final List<Long> linksInRangeTimes = database.getLinksInRangeTime;
			if(linksInRangeTimes.size() > 0){
				long c = 0;
				for(long l : linksInRangeTimes){
					c += l;
				}
				LOGGER.info("getLinksInRange Time: first: " + linksInRangeTimes.get(0) + ", average: " + c/linksInRangeTimes.size());
			}
			final List<Long> nodeTravellingCostsTimes = database.getNodeTravellingCostsTime;
			if(nodeTravellingCostsTimes.size() > 0){
				long c = 0;
				for(long l : nodeTravellingCostsTimes){
					c += l;
				}
				LOGGER.info("getNodeTravellingCostsTime: first: " + nodeTravellingCostsTimes.get(0) + ", average: " + c/nodeTravellingCostsTimes.size());
			}
			final List<Long> rangeByNodeCountTimes = database.getRangeByNodeCountTime;
			if(rangeByNodeCountTimes.size() > 0){
				long c = 0;
				for(long l : rangeByNodeCountTimes){
					c += l;
				}
				LOGGER.info("rangeByNodeCountTimes Time: first: " + rangeByNodeCountTimes.get(0) + ", average: " + c/rangeByNodeCountTimes.size());
			}
			final List<Long> getProjectOnLinksTimes = database.getProjectOnLinksTime;
			if(getProjectOnLinksTimes.size() > 0){
				long c = 0;
				for(long l : getProjectOnLinksTimes){
					c += l;
				}
				LOGGER.info("getProjectOnLinksTimes Time: first: " + getProjectOnLinksTimes.get(0) + ", average: " + c/getProjectOnLinksTimes.size());
			}
			final List<Long> readNetworkTimes = database.readNetworkTime;
			if(readNetworkTimes.size() > 0){
				long c = 0;
				for(long l : readNetworkTimes){
					c += l;
				}
				LOGGER.info("rangeByNodeCountTimes Time: first: " + readNetworkTimes.get(0) + ", average: " + c/readNetworkTimes.size());
			}
			
		} catch (final AlgorithmException e) {
			throw new IllegalStateException(e);
		}
	}
}
