package it.unibz.inf.isochrone.algorithm;

import java.awt.geom.Point2D;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.PriorityQueue;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.tools.ant.types.CommandlineJava.SysProperties;

import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.AbstractOutput;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Location;
import it.unibz.inf.isochrone.network.MemoryOutput;
import it.unibz.inf.isochrone.network.Node;
import it.unibz.inf.isochrone.network.NodeConnection;

/**
 * The isochrone algorithm. As the different algorithms all have
 * the same expansion algorithm, and differ only in the way the
 * data is loaded from the database. Therefore this is an abstract
 * class describing the expansion technique, which can be
 * instantiated by different algorithms that define the loading of
 * the data from the database.
 */
public abstract class AbstractAlgorithm {
	public static final double CONVERSION_FACTOR_TO_METER_PER_SEC = 3.6d;
	private static final AlgorithmCache CACHE;
	private static final String CACHE_IDENTIFIER = "default";
	private static final Logger LOGGER = LogManager.getLogger(AbstractAlgorithm.class);
	private static final Predicate<Link> PREDICATE_LINK_CONT = new Predicate<Link>() {
		private final Double dMin = Double.MIN_VALUE;

		@Override
		public boolean test(final Link link) {
			return link.isContinuous() && !dMin.equals(link.getStartOffset());
		}
	};
	protected final Long cDuration;
	protected final Long cMaxNodes;
	protected final boolean cIsIncoming;
	protected final long cFromTime;
	protected final boolean cIsMultimodal;
	protected final Collection<Point2D.Double> cQueryPoints;
	protected final Integer cQueryPointsSRID;
	protected final long cToTime;
	protected final double cWalkingSpeed;
	protected final Database database;
	private final Collection<Integer> serviceIds;
	private final ConfigAlgorithm config;
	private final PriorityQueue<Node> nodeQueue;
	private final AtomicBoolean running = new AtomicBoolean(false);
	private long expandedNodes;
	private AlgorithmStatistics stat;

	// Static block

	static {
		// set this to null to deactivate isochrone caching
		CACHE = AlgorithmCache.getInstance(CACHE_IDENTIFIER);
	}

	// Constructors

	/**
	 * Instantiates a new isochrone object.
	 * Note that the isochrone algorithm does not handle database connections.
	 * It does not close any of the connections that are opened during computation.
	 * Therefore, be sure to call database.close() as soon as the computation finished
	 * (as long as you don't need the database connections for something else).
	 *
	 * @param database used for database communication
	 * @param config the parameters for the isochrone
	 * @throws AlgorithmException thrown if there are no date codes in the database for the specified query
	 */
	public AbstractAlgorithm(final Database database, final ConfigAlgorithm config) throws AlgorithmException {
		if (config.getDateTime() == null) {
			throw new AlgorithmException("Invalid algorithm configuration given!");
		}

		if (LOGGER.isTraceEnabled()) {
			LOGGER.trace("Creating algorithm {}", this.getClass().getSimpleName());
		}

		this.config = config;
		this.cDuration = config.getDuration();
		this.cIsIncoming = config.isIncoming();
		this.cMaxNodes = config.getNodes();
		this.cQueryPoints = config.getQueryPoints();
		this.cQueryPointsSRID = config.getQueryPointSRID();
		this.cToTime = config.getTimeTo();
		this.cFromTime = config.getTimeFrom();
		this.cWalkingSpeed = config.getWalkingSpeed().doubleValue() / CONVERSION_FACTOR_TO_METER_PER_SEC;
		this.cIsMultimodal = config.isMultimodal();
		this.database = database;
		if (cDuration == null || !cIsMultimodal) {
			// we do not need serviceIds from the calendar table (since unimodal mode is used or no maxDuration is set)
			this.serviceIds = null;
		} else {
			this.serviceIds = getServiceIds();
			if (serviceIds == null || serviceIds.isEmpty()) {
				// empty calendars were solved in uniBz version by switching computation mode to unimodal mode...
				// ... we do not do this any more (since user does not get informed about the change)
				//  => we throw this (new) exception here
				throw new AlgorithmException("Could not find calendars in the database for the given date");
			}
		}

		this.nodeQueue = new PriorityQueue<>();
	}

	// Public abstract methods

	/**
	 * Closes all the resources used and removes calculation nodes/links from the memory.
	 */
	public abstract void clear();

	// Public methods

	/**
	 * Activates the recording of the algorithm's statistics.
	 * Note that this can slow down the computation.
	 */
	public void activateStatistics() {
		stat = new AlgorithmStatistics();
		database.activateStatistics();
	}

	/**
	 * Computes the isochrone and returns the result (as MemoryOutput object).
	 *
	 * @return the generated output (as MemoryOutput object) or null if the calculation has been stopped.
	 */
	public MemoryOutput getResult() {
		return getResult(MemoryOutput.class);
	}

	/**
	 * Computes the isochrone and returns the result (of type clazz).
	 *
	 * @param <T> subclass of MemoryOutput. In this class the result will be returned.
	 * @param clazz the calculation result will be returned in an object created from this class
	 * @return the generated output (of class specified by parameter clazz) or null if the calculation has been stopped.
	 */
	public <T extends MemoryOutput> T getResult(final Class<T> clazz) {
		if (LOGGER.isTraceEnabled()) {
			LOGGER.trace("Getting result from algorithm {} and returning it as {} object", this.getClass().getSimpleName(), clazz.getSimpleName());
		}

		if (clazz == null) {
			throw new NullPointerException("Can not compute with null output");
		}

		final CacheKey cacheKey = new CacheKey(this.getClass(), clazz, config);
		T output = getCachedResult(cacheKey);
		if (output != null) {
			return output;
		}

		output = initializeQueryPoints(clazz, cQueryPoints);
		if (!calculate(output)) {
			return null;
		}

		if (CACHE != null) {
			LOGGER.trace("   caching result for future references");
			CACHE.put(cacheKey, output);
		}

		return output;
	}

	/**
	 * Gets the statistics created during the algorithm execution.
	 * Note that {@link #activateStatistics()} has to be called before starting the execution in order to return a non-null result here.
	 *
	 * @return statistics object containing information about the algorithm execution
	 */
	public AlgorithmStatistics getStatistics() {
		if (stat == null) {
			return null;
		}

		// merge database statistics
		final Map<String, Long> dbStats = database.getStatistics();
		if (dbStats != null) {
			final Set<Entry<String, Long>> e = dbStats.entrySet();
			for (final Entry<String, Long> entry : e) {
				stat.setNumericStat(entry.getKey(), entry.getValue());
			}
		}

		return stat;
	}

	public boolean isRunning() {
		return running.get();
	}

	public void stop() {
		running.set(false);
	}

	// Public static methods

	public static void clearCache() {
		if (CACHE == null) {
			return;
		}

		LOGGER.debug("Clearing algorithm result cache");
		CACHE.clear();
	}

	// Protected abstract methods

	/**
	 * Gets a single link from the network.
	 *
	 * @param linkId the id of the requested link
	 * @return the requested link
	 */
	protected abstract Link getLink(int linkId);

	/**
	 * Gets all links to (or from) a given node.
	 *
	 * @param node the node for which the links should be retrieved.
	 * @return the link collection
	 */
	protected abstract Collection<Link> getLinks(final Node node);

	/**
	 * Gets a single node from the network.
	 *
	 * @param nodeId the id of the requested node
	 * @return the node requested
	 */
	protected abstract Node getNode(int nodeId);

	/**
	 * Remove a node from the in-memory network, only used in the
	 * MineX and MineRX algorithms.
	 *
	 * @param node The node to be removed (identified by id)
	 */
	protected abstract void removeNode(final Node node);

	/**
	 * Indicates if nodes for this algorithm expires.
	 *
	 * @return flag indicating if nodes expire for this algorithm
	 */
	protected abstract boolean shouldNodesExpire();

	// Protcted methods

	protected boolean isStatisticsActive() {
		return stat != null;
	}

	// Private methods

	private void addTargetNode(final NodeConnection nConnection, final Node node, final Link link) {
		final Node targetNode = getNode(link.getOppositeOf(node));
		targetNode.visitLink();

		// TODO: Are there cases, where closed target nodes should be added (maybe in the case when two distinct isochrone areas are merged?)
		if (targetNode.isClosed()) {
			if (shouldNodesExpire() && targetNode.isExpired()) {
				removeNode(targetNode);
			}
		} else {
			nConnection.addLink(targetNode, link);
		}
	}

	/**
	 * Calculates the isochrone and stores the result in the given output.
	 * Note that the output may not be valid if the result if this is false.
	 *
	 * @param output the output to store the result in
	 * @return boolean flag indicating if the calculation finished (true) or if the calculation has been stopped (false)
	 */
	private <T extends MemoryOutput> boolean calculate(final T output) {
		LOGGER.info("Calculating new isochrone");
		LOGGER.debug("   configuration is: {}", config.toString());

		running.set(true);
		while (running.get() && !nodeQueue.isEmpty() && (cMaxNodes == null || expandedNodes < cMaxNodes)) {
			expandNode(nodeQueue.poll(), output);
		}

		return running.getAndSet(false);
	}

	private <T extends MemoryOutput> T calculateIncremental(final T cachedOutput, final long offset) {
		@SuppressWarnings("unchecked")
		final T output = (T) createOutput(cachedOutput.getClass());
		final Collection<Link> links = cachedOutput.getLinks();
		final Collection<Node> nodes = cachedOutput.getNodes();
		if (offset == 0) {
			output.addLinks(links);
			output.addNodes(nodes);
		} else if (offset > 0) {
			// duration in cache is bigger than requested duration => we need to shrink the already known isochrone
			for (final Node node : nodes) {
				final double nDistance = node.getDistance();
				if (cDuration == null || nDistance <= cDuration) {
					output.addNode(node);
					final Collection<Link> nodeLinks = node.getLinks();
					if (nodeLinks != null) {
						nodeLinks.stream().forEach(l -> {
							prepareLinkOffsets(l, nDistance);
							output.addLink(l);
						});
					}
				}
			}
		} else {
			// duration in cache is smaller than requested duration => we need to expand the already known isochrone
			// we also have to take care about bus stations and other discrete links
			output.addLinks(links);
			output.addNodes(nodes);

			for (final Node node : nodes) {
				if (!node.isClosed()) {
					nodeQueue.add(node);
				}
			}

			// TODO: Re-calculate discrete nodes (such as bus stations, ...)
//			calculate(output);
			return null;
		}

		return output;
	}

	private <T extends AbstractOutput> T createOutput(final Class<T> clazz, final Object... params) {
		T output = AbstractOutput.createInstance(clazz, params);
		if (output == null) {
			// prepend database object to params
			final Object[] extendedParams = new Object[params.length + 1];
			extendedParams[0] = database;
			System.arraycopy(params, 0, extendedParams, 1, params.length);

			output = AbstractOutput.createInstance(clazz, extendedParams);
		}

		return output;
	}

	/**
	 * Expand the given continuous link (that is connected to the given node)
	 * and adjusts the link offsets based on the distance of the current Node
	 * and the maximum duration of the query.
	 *
	 * @param nConnection information about the the sourceNode, its links and routes
	 * @return another node, if it can be reached in the available time (or null if time has run out)
	 */
	private Set<Node> expandContinuousLinks(final NodeConnection nConnection) {
		if (!nConnection.containsTargets()) {
			return Collections.emptySet();
		}

		final Node sourceNode = nConnection.getSourceNode();
		final double sourceDistance = sourceNode.getDistance();
		final Map<Node, Collection<Link>> targets = nConnection.getTargetEntries();
		final Collection<Entry<Node, Collection<Link>>> entries = targets.entrySet();

		final Set<Node> results = new LinkedHashSet<>(targets.size());
		for (final Entry<Node, Collection<Link>> e : entries) {
			final Node targetNode = e.getKey();
			final Collection<Link> targetLinks = e.getValue();
			for (final Link link : targetLinks) {
				final double newDistance = sourceDistance + link.getLength() / cWalkingSpeed;
				if ((cDuration == null || newDistance <= cDuration) && (newDistance < targetNode.getDistance())) {
					targetNode.setDistance(newDistance);
					results.add(targetNode);
				}
			}
		}

		return results;
	}

	/**
	 * Expand the given discrete link (that is connected to the
	 * given node and adjust the link offsets based on the
	 * distance of the current Node and the maximum duration of
	 * the query.
	 *
	 * @param nConnection information about the the sourceNode, its links and routes
	 * @return all nodes that can be reached in the available time (or an empty set if there is no time to reach another node)
	 */
	private Set<Node> expandDiscreteLinks(final NodeConnection nConnection) {
		if (!nConnection.containsTargets()) {
			return Collections.emptySet();
		}

		final Map<Node, Long> newDistances = getNodeTravellingCosts(nConnection);
		final Collection<Entry<Node, Long>> entries = newDistances.entrySet();

		final Set<Node> results = new LinkedHashSet<>(newDistances.size());
		for (final Entry<Node, Long> e : entries) {
			final Node targetNode = e.getKey();
			final long newDistance = e.getValue();
			if ((cDuration == null || newDistance <= cDuration) && newDistance < targetNode.getDistance()) {
				targetNode.setDistance(newDistance);
				results.add(targetNode);
			}
		}

		return results;
	}

	private <T extends AbstractOutput> void expandNode(final Node node, final T output) {
		final Collection<Link> links = getLinks(node);
		node.close();
		final NodeConnection cConnections = prepareContinuousLinks(node, links);
		updateNodeQueue(expandContinuousLinks(cConnections));

		if (cIsMultimodal) {
			final NodeConnection dConnections = prepareDiscreteLinks(node, links);
			updateNodeQueue(expandDiscreteLinks(dConnections));
		}

		if (shouldNodesExpire() && node.isExpired()) {
			removeNode(node);
		}

		final Collection<Link> filteredLinks = links.parallelStream().filter(PREDICATE_LINK_CONT).collect(Collectors.toList());
		node.addLinks(filteredLinks);
		output.addLinks(filteredLinks);

		final boolean changed = output.addNode(node);
		if (changed) {
			++expandedNodes;
		}
	}

	private <T extends MemoryOutput> T getCachedResult(final CacheKey cacheKey) {
		if (config.isCacheActive()) {
			// get result from cache (exact match)
			final T cachedOutput = (CACHE == null) ? null : CACHE.getExact(cacheKey);
			if (cachedOutput != null) {
				final T output = calculateIncremental(cachedOutput, 0);
				if (output != null) {
					LOGGER.info("Returning isochrone from cached result (exact match)");
					return output;
				}
			}
		}

		if (config.isCacheIncrementalActive()) {
			// no exact cached result found... try to find similar result and perform incremental isochrone calculation
			final Map.Entry<Long, T> similarOutput = (CACHE == null) ? null : CACHE.getSimilar(cacheKey);
			if (similarOutput != null) {
				final T output = calculateIncremental(similarOutput.getValue(), similarOutput.getKey());
				if (output != null) {
					LOGGER.info("Returning isochrone from cached result (incremental calculated)");
					LOGGER.trace("   caching result for future references");

					CACHE.put(cacheKey, output);
					return output;
				}
			}
		}

		return null;
	}

	/**
	 * Gets valid service ids (from the calendar table) for a given dateTime.
	 *
	 * @return returns a set of date codes
	 */
	private Collection<Integer> getServiceIds() {
		if (cDuration == null) {
			return null;
		}

		final LocalDateTime dateTime = config.getDateTime();
		final LocalDate endDate = (cIsIncoming) ? dateTime.toLocalDate() : dateTime.plus(cDuration, ChronoUnit.SECONDS).toLocalDate();
		final LocalDate startDate = (cIsIncoming) ? dateTime.minus(cDuration, ChronoUnit.SECONDS).toLocalDate() : dateTime.toLocalDate();

		return database.getCalendar(startDate, endDate);
	}

	/**
	 * Get the cost to travel to an adjacent node.
	 *
	 * @param nConnection information about the the sourceNode, its links and routes
	 * @return the cost of traveling from the node to the adjacent nodes (as map)
	 */
	private Map<Node, Long> getNodeTravellingCosts(final NodeConnection nConnection) {
		return database.getNodeTravellingCosts(nConnection, serviceIds, cFromTime, cToTime);
	}

	/**
	 * Initializes the queryPoint from which the isochrone should be calculated.
	 *
	 * @param queryPoints The queryPoint from which the isochrone should start.
	 * @return the output with the start locations initialized
	 */
	private <T extends AbstractOutput> T initializeQueryPoints(final Class<T> clazz, final Collection<Point2D.Double> queryPoints) {
		LOGGER.debug("Initializing queryPoint locations");
		final T output = createOutput(clazz);
		final double walkingDuration = (cDuration == null) ? 0 : cDuration * cWalkingSpeed;

		final Set<Location> locations = new HashSet<>(queryPoints.size() * 2);
		for (final Point2D qPoint : queryPoints) {
			LOGGER.debug("   {}", qPoint.toString());
			if (cQueryPointsSRID == null) {
				locations.addAll(database.projectOnLinks(qPoint));
			} else {
				locations.addAll(database.projectOnLinks(qPoint, cQueryPointsSRID));
			}
		}
		for (final Location location : locations) {
			final Link link = getLink(location.getLinkId());
			final double locationOffset = location.getOffset();
			double distance;
			Node node;

			if (cIsIncoming) {
				node = getNode(link.getStartNode());
				distance = locationOffset / cWalkingSpeed;
				if (locationOffset > 0 && locationOffset < link.getLength()) {
					link.setStartOffset(Math.max(0, locationOffset - walkingDuration));
					link.setEndOffset(locationOffset);
					output.addLink(link);
				}
			} else {
				final double d = link.getLength() - locationOffset;
				node = getNode(link.getEndNode());
				distance = d / cWalkingSpeed;
				if (locationOffset > 0 && locationOffset < link.getLength()) {
					link.setStartOffset(locationOffset);
					link.setEndOffset(Math.min(link.getLength(), Math.abs(d - walkingDuration)));
					output.addLink(link);
				}
			}
			if ((cDuration == null || distance <= cDuration) && distance < node.getDistance()) {
				node.setDistance(distance);
				updateNodeQueue(node);
			}
		}

		return output;
	}

	private NodeConnection prepareContinuousLinks(final Node node, final Collection<Link> links) {
		final NodeConnection nConnection = new NodeConnection(node);
		if (links == null || links.isEmpty()) {
			return nConnection;
		}

		final double nodeDistance = node.getDistance();
		links.stream().filter(l -> l.isContinuous()).forEach(l -> {
			prepareLinkOffsets(l, nodeDistance);
			addTargetNode(nConnection, node, l);
		});

		return nConnection;
	}

	private NodeConnection prepareDiscreteLinks(final Node node, final Collection<Link> links) {
		final NodeConnection nConnection = new NodeConnection(node);
		if (links == null || links.isEmpty()) {
			return nConnection;
		}

		links.stream().filter(l -> !l.isContinuous()).forEach(l -> {
			addTargetNode(nConnection, node, l);
		});

		return nConnection;
	}

	private void prepareLinkOffsets(final Link link, final double nodeDistance) {
		final double linkLength = link.getLength();
		final double maxDistance = (cDuration == null) ? 0 : (cDuration - nodeDistance);

		if (cIsIncoming) {
			link.setStartOffset(Math.max(0, linkLength - maxDistance * cWalkingSpeed));
			link.setEndOffset(linkLength);
		} else {
			final double remainingDistance = Math.max(0, maxDistance);
			link.setStartOffset(0);
			link.setEndOffset(Math.min(linkLength, remainingDistance * cWalkingSpeed));
		}
	}

	/**
	 * Sets an updated node into the right place in the priority queue.
	 *
	 * @param node The node that was changed
	 */
	private void updateNodeQueue(final Node node) {
		if (node == null) {
			return;
		}
		nodeQueue.remove(node);
		nodeQueue.add(node);
	}

	/**
	 * Sets updated nodes into the right place in the priority queue.
	 *
	 * @param nodes The nodes that were changed
	 */
	private void updateNodeQueue(final Set<Node> nodes) {
		if (nodes == null || nodes.isEmpty()) {
			return;
		}

		nodeQueue.removeAll(nodes);
		nodeQueue.addAll(nodes);
	}

}
