package it.unibz.inf.isochrone.algorithm;

import it.unibz.inf.isochrone.network.AbstractOutput;

import java.util.AbstractMap.SimpleEntry;
import java.util.Comparator;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.commons.jcs.JCS;
import org.apache.commons.jcs.access.CacheAccess;
import org.apache.commons.jcs.access.exception.CacheException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

final class AlgorithmCache {
	private static final Logger LOGGER = LogManager.getLogger(AlgorithmCache.class);
	private static final Comparator<Number> COMPARATOR_ABS_VALUE = new Comparator<Number>() {
		@Override
		public int compare(final Number o1, final Number o2) {
			return Math.abs(o1.intValue()) - Math.abs(o2.intValue());
		}
	};
	private static final String PATTERN_DURATION = "duration=[0-9]+";
	private final CacheAccess<String, AbstractOutput> cache;

	// Constructor

	private AlgorithmCache(final String cacheIdentifier) {
		cache = init(cacheIdentifier);
	}

	// Public static methods

	public static AlgorithmCache getInstance(final String cacheIdentifier) {
		return new AlgorithmCache(cacheIdentifier);
	}

	// Public methods

	public void clear() {
		if (cache == null) {
			return;
		}

		try {
			cache.clear();
		} catch (final CacheException e) {
			// deactivates the cache
			LOGGER.warn("Cache could not be cleared... deactivating cache");
		}
	}

	public void put(final CacheKey cacheKey, final AbstractOutput output) {
		if (cache == null || output == null) {
			return;
		}

		try {
			final String key = cacheKey.toString();
			LOGGER.trace("Storing output into cache: {}", key);
			cache.put(key, output);
		} catch (final CacheException e) {
			// could not cache output... no problem, so do not do anything but logging
			LOGGER.warn("Could not cache calculation result. This can lead to worse performance");
			LOGGER.debug(e);
		}
	}

	public <T extends AbstractOutput> T getExact(final CacheKey cacheKey) {
		if (cache == null) {
			return null;
		}

		final String key = cacheKey.toString();
		LOGGER.trace("Retrieving output from cache: {}", key);

		@SuppressWarnings("unchecked")
		final T t = (T) cache.get(key);
		return t;
	}

	public <T extends AbstractOutput> Map.Entry<Long, T> getSimilar(final CacheKey cacheKey) {
		if (cache == null) {
			return null;
		}

		final String key = cacheKey.toString();
		LOGGER.trace("Retrieving similar outputs (similar in duration) from cache: {}", key);
		final SortedMap<Long, String> similarKeys = getSimilarDurationKeys(cacheKey);
		if (similarKeys == null || similarKeys.isEmpty()) {
			LOGGER.trace("No similar keys found");
			return null;
		}

		if (LOGGER.isTraceEnabled()) {
			LOGGER.trace("Found {} similar keys", similarKeys.size());
		}

		final Long firstKey = similarKeys.firstKey();
		final String firstValue = similarKeys.get(firstKey);

		@SuppressWarnings("unchecked")
		final T t = (T) cache.get(firstValue);
		return new SimpleEntry<Long, T>(firstKey, t);
	}

	// Private static methods

	/**
	 * Gets keys for the cache in a sorted map.
	 * The map is sorted by the absolute duration difference to the given queryKey.
	 *
	 * @param key the cache key to get similar durations for
	 * @return sorted map with similar duration keys (sorted by the duration difference to the given key)
	 */
	private SortedMap<Long, String> getSimilarDurationKeys(final CacheKey key) {
		return getSimilarKeys(key, PATTERN_DURATION);
	}

	private SortedMap<Long, String> getSimilarKeys(final CacheKey key, final String pattern) {
		if (cache == null) {
			return null;
		}

		final String keyStr = key.toString();
		final String keyPattern = keyStr.replaceAll("\\[", "\\\\[").replaceAll("\\]", "\\\\]").replaceAll("\\.", "\\\\.").replaceAll(pattern, pattern);
		final Map<String, AbstractOutput> matchingMap = cache.getMatching(keyPattern);
		if (matchingMap.isEmpty()) {
			return null;
		}

		final SortedMap<Long, String> resultMap = new TreeMap<>(COMPARATOR_ABS_VALUE);
		final long queryDuration = CacheKey.getDurationFromString(keyStr);
		final Set<String> keySet = matchingMap.keySet();
		for (final String k : keySet) {
			final long duration = CacheKey.getDurationFromString(k);
			resultMap.put(duration - queryDuration, k);
		}

		return resultMap;
	}

	/**
	 * Initializes the JCS cache.
	 *
	 * @return the initialized cache (or null if it could no be initialized)
	 */
	private CacheAccess<String, AbstractOutput> init(final String cacheIdentifier) {
		try {
			return JCS.getInstance(cacheIdentifier);
		} catch (final CacheException e) {
			LOGGER.warn("Could not initialize isochrone cache (for incremental calculation). This can lead to worse performance.");
			return null;
		}
	}
}
