package it.unibz.inf.isochrone.algorithm;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics.Statistic;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Node;

/**
 * The Dijkstra isochrone algorithm.
 * Loads all data at the beginning and then takes the data out of maps for consumption.
 * Especially good for isochrones that span a big area of the data that is in the database.
 */
public class MDijkstra extends AbstractAlgorithm {
	private final Map<Integer, Link> links;
	private final Map<Integer, Node> nodes;

	// Constructors

	public MDijkstra(final Database db, final ConfigAlgorithm config) throws AlgorithmException {
		super(db, config);

		links = new LinkedHashMap<>();
		nodes = new HashMap<>();

		database.readNetwork(nodes, links);
	}

	// Public methods

	@Override
	public void clear() {
		links.clear();
		nodes.clear();
	}

	@Override
	public AlgorithmStatistics getStatistics() {
		final AlgorithmStatistics stat = super.getStatistics();
		if (stat == null) {
			return null;
		}

		final int nodeSize = nodes.size();

		stat.setNodes(nodes.values());
		stat.setNumericStat(Statistic.LOADED_NODES_CURRENT, nodeSize);
		stat.setNumericStat(Statistic.LOADED_NODES_MAX, nodeSize);
		stat.setNumericStat(Statistic.LOADED_NODES_TOTAL, nodeSize);

		return stat;
	}

	// Protected methods

	@Override
	protected Link getLink(final int linkId) {
		final Link link = links.get(linkId);
		if (link == null) {
			throw new RuntimeException("LinkId not present: " + linkId);
		}

		return link;
	}

	@Override
	protected Collection<Link> getLinks(final Node node) {
		final Collection<Link> l = getNode(node.getId()).getLinks();
		if (l == null) {
			return Collections.emptyList();
		}

		return l;
	}

	@Override
	protected Node getNode(final int nodeId) {
		final Node node = nodes.get(nodeId);
		if (node == null) {
			throw new RuntimeException("NodeId not present: " + nodeId);
		}

		return node;
	}

	@Override
	protected void removeNode(final Node node) {
		// Nothing to do here
//		nodes.remove(node.getId());
	}

	@Override
	protected boolean shouldNodesExpire() {
		return false;
	}

}
