package it.unibz.inf.isochrone.algorithm;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics.Statistic;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Node;

/**
 * The Mine isochrone algorithm.
 * Loads the nodes and links from the database as soon as they are needed.
 */
public class Mine extends AbstractAlgorithm {
	protected final Map<Integer, Node> nodes;
	private long loadedNodesMax;
	private long loadedNodesTotal;

	// Constructors

	public Mine(final Database db, final ConfigAlgorithm config) throws AlgorithmException {
		super(db, config);
		nodes = new HashMap<>();
	}

	// Public methods

	@Override
	public void clear() {
		nodes.clear();
	}

	@Override
	public AlgorithmStatistics getStatistics() {
		final AlgorithmStatistics stat = super.getStatistics();
		if (stat == null) {
			return null;
		}

		final int nodeSize = nodes.size();

		stat.setNodes(nodes.values());
		stat.setNumericStat(Statistic.LOADED_NODES_CURRENT, nodeSize);
		stat.setNumericStat(Statistic.LOADED_NODES_MAX, loadedNodesMax);
		stat.setNumericStat(Statistic.LOADED_NODES_TOTAL, loadedNodesTotal);

		return stat;
	}

	// Protected methods

	@Override
	protected Link getLink(final int linkId) {
		final Link link = database.getLink(linkId, nodes);
		if (link == null) {
			throw new RuntimeException("LinkId not present: " + linkId);
		}

		return link;
	}

	@Override
	protected Collection<Link> getLinks(final Node node) {
		final int numBefore = (isStatisticsActive() ? nodes.size() : -1);
		final Collection<Link> resultLinks = database.getLinks(node, nodes);

		if (isStatisticsActive()) {
			updateLoadingStatistics(numBefore, nodes.size());
		}

		return resultLinks;
	}

	@Override
	protected Node getNode(final int nodeId) {
		final Node node = nodes.get(nodeId);
		if (node == null) {
			throw new RuntimeException("NodeId not present: " + nodeId);
//			node = database.getNode(nodeId);
//			if (node == null) {
//				throw new RuntimeException("NodeId not present: " + nodeId);
//			}
//
//			nodes.put(nodeId, node);
		}

		return node;
	}

	@Override
	protected void removeNode(final Node node) {
		nodes.remove(node.getId());
	}

	@Override
	protected boolean shouldNodesExpire() {
		return false;
	}

	protected int updateLoadingStatistics(final int numBefore, final int numCurrent) {
		if (!isStatisticsActive()) {
			return -1;
		}

		final int addedNodes = numCurrent - numBefore;
		loadedNodesTotal = Math.max(loadedNodesTotal, numBefore);
		loadedNodesMax = Math.max(loadedNodesMax, numCurrent);
		loadedNodesTotal += addedNodes;

		return addedNodes;
	}

}
