package it.unibz.inf.isochrone.algorithm;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.tools.ant.types.CommandlineJava.SysProperties;

import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Node;
import it.unibz.inf.isochrone.network.Seed;

/**
 * The MineR isochrone algorithm.
 * Loads the data in circular ranges that surround the currently explored node.
 */
public class MineR extends Mine {
	private static final Logger LOGGER = LogManager.getLogger(MineR.class);
	/** Memory that is used at max by nodes (other memory is reserved for route collections, links, ...). */
	private static final float MAX_MEMORY_PERCENT = 0.3f;
	private final Map<Integer, Collection<Link>> loadedLinks;
	private final Set<Seed> loadingSeeds;
	private final Long nodeLimit;
	private final long totalMemory;

	// Constructors

	public MineR(final Database db, final ConfigAlgorithm config) throws AlgorithmException {
		this(db, config, null);
	}

	public MineR(final Database db, final ConfigAlgorithm config, final Long nodeLimit) throws AlgorithmException {
		super(db, config);
		this.nodeLimit = (nodeLimit == null || nodeLimit <= 0) ? null : nodeLimit;
		this.totalMemory = Runtime.getRuntime().totalMemory();

		loadedLinks = new HashMap<>();
		loadingSeeds = new LinkedHashSet<>();
	}

	// Public methods

	@Override
	public void clear() {
		super.clear();

		loadedLinks.clear();
		loadingSeeds.clear();
	}

	public Collection<Seed> getLoadingSeeds() {
		// FIXME: Handle incremental caching... list is empty when result is loaded from incremental cache
		return new ArrayList<>(loadingSeeds);
	}

	// Protected methods

	@Override
	protected Collection<Link> getLinks(final Node node) {
		final int nodeId = node.getId();
		final Collection<Link> links = loadedLinks.get(nodeId);
		if (links != null) {
			/*
			 * Check if continuous links are empty and the
			 * node is pre-explored.  In that case load the
			 * continuous links, because density may have
			 * been to small, so only the discrete links
			 * were loaded.
			 *
			 * Most likely this is a non-issue, because
			 * the old code checks if a node is pre-explored,
			 * but that's always false (java initializes
			 * variables as false and it's never set) and
			 * therefore continuous links are never added
			 * after getting the links.
			 *
			 * Also I like long comments :(
			 */
			return links;
		}

		final Seed seed = new Seed(database.getCoordinates(nodeId), getLoadableRadius(node));
		final Collection<Seed> seedsOverlapping = getOverlappingRegionSeeds(seed);
		final Map<Integer, Collection<Link>> mLoaded = loadRange(seed, seedsOverlapping);
		loadedLinks.putAll(mLoaded);
		final Collection<Link> result = mLoaded.get(nodeId);
		return (result == null) ? Collections.emptyList() : result;
	}

	@Override
	protected boolean shouldNodesExpire() {
		return false;
	}

	// Private methods

	/**
	 * Gets the maximal radius that should be used for range queries.
	 * This takes care about the total memory when calculating the radius.
	 * To do so the database is accessed (using {@link Database#getRangeByNodeCount(int, int)}) if (and only if) the nodeLimit has been set.
	 *
	 * @param node The node for which the radius should be calculated
	 * @return the radius that should be used for range queries (max == remaining distance; actual value depends on free memory)
	 */
	private double getLoadableRadius(final Node node) {
		final double radius = getRemainingDistance(node);
		final Long currentNodeLimit = getCurrentNodeLimit();
		if (currentNodeLimit == null) {
			return radius;
		}
		return Math.min(radius, database.getRangeByNodeCount(node.getId(), currentNodeLimit));
	}

	/**
	 * Returns the maximum count of nodes that are allowed to be loaded into memory.
	 *
	 * @return the maximum count of nodes that are allowed to be loaded. Null if there is no limit set.
	 */
	private Long getCurrentNodeLimit() {
		if (nodeLimit == null) {
			return null;
		}

		final long nodeCount = Math.max(0, nodeLimit - nodes.size());
		final double allowedMemory = totalMemory * MAX_MEMORY_PERCENT;
		final long spaceNeeded = Math.max(1, nodeCount * Node.sizeOf());
		if (spaceNeeded > allowedMemory) {
			return (long) Math.floor(allowedMemory / Node.sizeOf());
		}

		return nodeCount;
	}

	private Collection<Seed> getOverlappingRegionSeeds(final Seed seed) {
		final int loadedSize = loadingSeeds.size();
		final double radius = seed.getRadius();

		final Collection<Seed> intersectionSeeds = new ArrayList<>(loadedSize + 1);
		for (final Seed loadedSeed : loadingSeeds) {
			final double distance = seed.euclideanDistance(loadedSeed.getCenter());
			if (distance < radius + loadedSeed.getRadius()) {
				intersectionSeeds.add(loadedSeed);
			}
		}
		return intersectionSeeds;
	}

	/**
	 * Gets the remaining distance for a node.
	 *
	 * @param node The node for which the remaining distance should be calculated
	 * @return the remaining distance
	 */
	private double getRemainingDistance(final Node node) {
		if (cDuration == null) {
			return Double.MAX_VALUE;
		}

		return (cDuration - node.getDistance()) * cWalkingSpeed;
	}

	/**
	 * Loads the links in the circle that should be loaded from the database.
	 * Also adds the loaded nodes to the nodes collection of the class.
	 *
	 * @param seed the seed, which is in the center of the stuff to be loaded
	 * @param intersectionSeeds locations that are already loaded
	 * @return the links loaded (as map; key == nodeId; value == collection of links)
	 */
	private Map<Integer, Collection<Link>> loadRange(final Seed seed, final Collection<Seed> intersectionSeeds) {
		if (LOGGER.isTraceEnabled()) {
			LOGGER.trace("Starting new range query");
			LOGGER.trace("   seed: {}", seed.toString());
			LOGGER.trace("   #seedsOverlapping: {}", intersectionSeeds.size());
		}

		loadingSeeds.add(seed);

		final int numBefore = (isStatisticsActive() ? nodes.size() : -1);
		final Map<Integer, Collection<Link>> resultLinks = database.getLinksInRange(seed, intersectionSeeds, nodes);
		
		if (isStatisticsActive()) {
			final int nodesAdded = updateLoadingStatistics(numBefore, nodes.size());
			LOGGER.trace("   {} new nodes were added by the loaded range", nodesAdded);
		}
		return resultLinks;
	}

}
