package it.unibz.inf.isochrone.config;

import java.awt.geom.Point2D;
import java.io.Serializable;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.time.Month;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.stream.Collectors;

/**
 * A query is used to compute different isochrones.  It is fed to the
 * Isochrone algorithm which will use the data to figure out what type
 * of isochrone is to be computed.
 *
 * <a href="mailto:nikolaus.krismer@uibk.ac.at">Nikolaus Krismer</a>
 */
public final class ConfigAlgorithm implements Serializable {
	public enum Direction { INCOMING, OUTGOING }
	public enum Mode { UNIMODAL, MULTIMODAL };
	public static final int MSEC_FACTOR = 1000;
	private static final long serialVersionUID = 701390363941349843L;
	private final LocalDateTime dateTime;
	private final Direction direction;
	private final Long duration;
	private final Long nodes;
	private final Mode mode;
	private final Collection<Point2D.Double> queryPoints;
	private final Integer queryPointSRID;
	private final boolean useCache;
	private final boolean useIncremetalCache;
	private final BigDecimal walkingSpeed;
	private Long toTime;
	private Long fromTime;

	// Constructors

	public static class ConfigAlgorithmBuilder implements Cloneable {
		private final Collection<Point2D.Double> queryPoints;
		// CHECKSTYLE:OFF MagicNumber
		private Direction direction = Direction.INCOMING;
		private LocalDateTime dateTime = LocalDateTime.of(2011, Month.OCTOBER, 20, 15, 0, 0, 0);
		private Long duration = null;
		private Long nodes = null;
		private Mode mode = Mode.MULTIMODAL;
		private Integer queryPointSRID = null;
		private boolean useCache = false;
		private boolean useIncremetalCache = false;
		private BigDecimal walkingSpeed = BigDecimal.valueOf(4.5d);
		// CHECKSTYLE:ON MagicNumber

		public ConfigAlgorithmBuilder(final Collection<Point2D.Double> queryPoints) {
			this.queryPoints = queryPoints;
		}

		public ConfigAlgorithmBuilder(final Point2D.Double l0, final Point2D.Double... l) {
			this.queryPoints = new LinkedHashSet<>((l == null) ? 1 : l.length + 1);
			this.queryPoints.add(l0);
			if (l != null) {
				this.queryPoints.addAll(Arrays.asList(l));
			}
		}

		public ConfigAlgorithm build() {
			if (duration == null && nodes == null) {
				throw new IllegalStateException("Either duration or nodes (or both) has to be set!");
			}

			return new ConfigAlgorithm(this);
		}

		@Override
		public ConfigAlgorithmBuilder clone() throws CloneNotSupportedException {
			return (ConfigAlgorithmBuilder) super.clone();
		}

		// CHECKSTYLE:OFF HiddenField
		public ConfigAlgorithmBuilder dateTime(final LocalDateTime dateTime) {
			this.dateTime = dateTime;
			return this;
		}

		public ConfigAlgorithmBuilder direction(final Direction direction) {
			this.direction = direction;
			return this;
		}

		public ConfigAlgorithmBuilder duration(final long duration) {
			this.duration = duration;
			return this;
		}

		public ConfigAlgorithmBuilder nodes(final long nodes) {
			this.nodes = nodes;
			return this;
		}

		public ConfigAlgorithmBuilder mode(final Mode mode) {
			this.mode = mode;
			return this;
		}

		public ConfigAlgorithmBuilder queryPointSRID(final int queryPointSRID) {
			this.queryPointSRID = queryPointSRID;
			return this;
		}

		public ConfigAlgorithmBuilder useCache() {
			this.useCache = true;
			return this;
		}

		public ConfigAlgorithmBuilder useIncrementalCache() {
			this.useIncremetalCache = true;
			return this;
		}

		public ConfigAlgorithmBuilder walkingSpeed(final BigDecimal walkingSpeed) {
			this.walkingSpeed = walkingSpeed;
			return this;
		}
		// CHECKSTYLE:OFF HiddenField
	}

	private ConfigAlgorithm(final ConfigAlgorithmBuilder builder) {
		this.dateTime = builder.dateTime;
		this.direction = builder.direction;
		this.duration = builder.duration;
		this.nodes = builder.nodes;
		this.mode = builder.mode;
		this.queryPointSRID = builder.queryPointSRID;
		this.useCache = builder.useCache;
		this.useIncremetalCache = builder.useIncremetalCache;
		this.queryPoints = builder.queryPoints;
		this.walkingSpeed = builder.walkingSpeed;

		initDateMillis();
	}

	// Getters

	public LocalDateTime getDateTime() {
		return dateTime;
	}

	public Direction getDirection() {
		return direction;
	}

	public Long getDuration() {
		return duration;
	}

	public Mode getMode() {
		return mode;
	}

	public Long getNodes() {
		return nodes;
	}

	public Collection<Point2D.Double> getQueryPoints() {
		return queryPoints;
	}

	public Integer getQueryPointSRID() {
		return queryPointSRID;
	}

	public Long getTimeFrom() {
		return fromTime;
	}

	public Long getTimeTo() {
		return toTime;
	}

	public BigDecimal getWalkingSpeed() {
		return walkingSpeed;
	}

	public boolean isCacheActive() {
		return useCache;
	}

	public boolean isCacheIncrementalActive() {
		return useIncremetalCache;
	}

	public boolean isIncoming() {
		return direction == Direction.INCOMING;
	}

	public boolean isMultimodal() {
		return mode == Mode.MULTIMODAL;
	}

	// (Private) Setters

	private void initDateMillis() {
		if (dateTime == null) {
			return;
		}

		final long timeDiff = (duration == null) ? 0 : duration;
		if (getDirection() == Direction.INCOMING) {
			toTime = getSecondsTillMidnight(dateTime);
			fromTime = Math.max(0, toTime - timeDiff);
		} else {
			fromTime = getSecondsTillMidnight(dateTime);
			toTime = fromTime + timeDiff;
		}
	}

	// Public methods

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		// CHECKSTYLE:OFF MagicNumber
		result = prime * result + ((dateTime == null) ? 0 : dateTime.hashCode());
		result = prime * result + ((direction == null) ? 0 : direction.hashCode());
		result = prime * result + ((duration == null) ? 0 : duration.hashCode());
		result = prime * result + ((mode == null) ? 0 : mode.hashCode());
		result = prime * result + ((nodes == null) ? 0 : nodes.hashCode());
		result = prime * result + ((queryPointSRID == null) ? 0 : queryPointSRID.hashCode());
		result = prime * result + ((queryPoints == null) ? 0 : queryPoints.hashCode());
		result = prime * result + (useCache ? 1231 : 1237);
		result = prime * result + (useIncremetalCache ? 1231 : 1237);
		result = prime * result + ((walkingSpeed == null) ? 0 : walkingSpeed.hashCode());
		// CHECKSTYLE:ON MagicNumber

		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final ConfigAlgorithm other = (ConfigAlgorithm) obj;
		if (dateTime == null) {
			if (other.dateTime != null) {
				return false;
			}
		} else if (!dateTime.equals(other.dateTime)) {
			return false;
		}
		if (direction != other.direction) {
			return false;
		}
		if (duration == null) {
			if (other.duration != null) {
				return false;
			}
		} else if (!duration.equals(other.duration)) {
			return false;
		}
		if (mode != other.mode) {
			return false;
		}
		if (nodes == null) {
			if (other.nodes != null) {
				return false;
			}
		} else if (!nodes.equals(other.nodes)) {
			return false;
		}
		if (queryPointSRID == null) {
			if (other.queryPointSRID != null) {
				return false;
			}
		} else if (!queryPointSRID.equals(other.queryPointSRID)) {
			return false;
		}
		if (queryPoints == null) {
			if (other.queryPoints != null) {
				return false;
			}
		} else if (!queryPoints.equals(other.queryPoints)) {
			return false;
		}
		if (useCache != other.useCache) {
			return false;
		}
		if (useIncremetalCache != other.useIncremetalCache) {
			return false;
		}
		if (walkingSpeed == null) {
			if (other.walkingSpeed != null) {
				return false;
			}
		} else if (!walkingSpeed.equals(other.walkingSpeed)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "ConfigAlgorithm [direction=" + direction
			+ ", date=" + dateTime.toString()
			+ ", maxNodes=" + nodes
			+ ", mode=" + mode
			+ ", walkingSpeed=" + walkingSpeed
			+ ", queryPoints=[" + queryPoints.stream().map(p -> "[" + p.getX() + "," + p.getY() + "]").collect(Collectors.joining(", ")) + "]"
			+ ", duration=" + duration
			+ "]";
	}

	// Public static methods

	public static long getSecondsTillMidnight(final LocalDateTime dateTime) {
		// CHECKSTYLE:OFF MagicNumber
		return dateTime.getHour() * 3600 + dateTime.getMinute() * 60 + dateTime.getSecond();
		// CHECKSTYLE:ON MagicNumber
	}

}
