package it.unibz.inf.isochrone.config;

import com.tngtech.configbuilder.ConfigBuilder;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertyExtension;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertyLocations;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertySuffixes;
import com.tngtech.configbuilder.annotation.typetransformer.TypeTransformer;
import com.tngtech.configbuilder.annotation.typetransformer.TypeTransformers;
import com.tngtech.configbuilder.annotation.valueextractor.DefaultValue;
import com.tngtech.configbuilder.annotation.valueextractor.PropertyValue;
import com.tngtech.propertyloader.impl.interfaces.PropertyLoaderFilter;

import it.unibz.inf.isochrone.db.DbTable;
import it.unibz.inf.isochrone.db.DbTableType;
import it.unibz.inf.isochrone.db.DbUtils;
import it.unibz.inf.isochrone.util.ResourceHelper;

import java.io.IOException;
import java.lang.reflect.Field;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.sql.Connection;
import java.sql.SQLException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.validation.constraints.NotNull;

import org.apache.commons.jcs.JCS;
import org.apache.commons.jcs.access.CacheAccess;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * Configuration for individual datasets.
 *
 * @author <a href="mailto:nikolaus.krismer@uibk.ac.at">Nikolaus Krismer</a>
 */
@PropertyLocations(fromClassLoader = true)
@PropertyExtension("xml")
@PropertySuffixes(extraSuffixes = { }, hostNames = false)
public class ConfigDataset {
	public static class DateTimeTransformer extends TypeTransformer<String, LocalDateTime> {
		public static final DateTimeFormatter DATETIME_FORMATTER = DateTimeFormatter.ofPattern("MM/dd/yyyy HH:mm");

		@Override
		public LocalDateTime transform(final String str) {
			LocalDateTime targetTime = null;
			try {
				targetTime = LocalDateTime.parse(str, DATETIME_FORMATTER);
			} catch (final DateTimeParseException e) {
				targetTime = null;
				e.printStackTrace();
			}

			return targetTime;
		}
	}

	private static final Logger LOGGER = LogManager.getLogger(ConfigDataset.class);
	private static final Collection<String> ALL_DATASETS;
	private static final int DEFAULT_SRID = 3857;
	private static final String DS_CONFIG_EXTENSION = "xml";
	private static final String DS_CONFIG_PATTERN = ".*config\\_(.*)\\." + DS_CONFIG_EXTENSION;
	private static final String CACHE_IDENTIFIER = "datasets";
	private static final CacheAccess<String, ConfigDataset> CACHE_INSTANCE = JCS.getInstance(CACHE_IDENTIFIER);

	@PropertyValue("client.tArrival")
	@TypeTransformers(DateTimeTransformer.class)
	private LocalDateTime clientArrivalDateTime;

	@DefaultValue("200")
	@NotNull
	@PropertyValue("spatial.distance_tolerance")
	private Integer distanceTolerance;

	@DefaultValue("sidx_%datasetName%_iso_coverage")
	@NotNull
	@PropertyValue("idx.isoCoverage")
	private String indexIsoCoverage;

	@DefaultValue("sidx_%datasetName%_iso_edges")
	@NotNull
	@PropertyValue("idx.isoLinks")
	private String indexIsoLinks;

	@DefaultValue("sidx_%datasetName%_iso_nodes")
	@NotNull
	@PropertyValue("idx.isoNodes")
	private String indexIsoNodes;

	@DefaultValue("false")
	@NotNull
	@PropertyValue("synthetic")
	private boolean synthetic;

	@DefaultValue("%datasetName%_calendar")
	@NotNull
	@PropertyValue("tbl.calendar")
	private String tableCalendar;

	@DefaultValue("%datasetName%_edges")
	@NotNull
	@PropertyValue("tbl.links")
	private String tableEdges;

	@DefaultValue("%datasetName%_iso_coverage")
	@NotNull
	@PropertyValue("tbl.isoCoverage")
	private String tableIsoCoverage;

	@DefaultValue("%datasetName%_iso_edges")
	@NotNull
	@PropertyValue("tbl.isoLinks")
	private String tableIsoLinks;

	@DefaultValue("%datasetName%_iso_nodes")
	@NotNull
	@PropertyValue("tbl.isoNodes")
	private String tableIsoNodes;

	@DefaultValue("%datasetName%_iso_nann")
	@NotNull
	@PropertyValue("tbl.isoNodesAnnotations")
	private String tableIsoNodesAnnotations;

	@DefaultValue("%datasetName%_routes")
	@NotNull
	@PropertyValue("tbl.routes")
	private String tableRoutes;

	@DefaultValue("%datasetName%_schedules")
	@NotNull
	@PropertyValue("tbl.schedules")
	private String tableSchedules;

	@DefaultValue("%datasetName%_nodes")
	@NotNull
	@PropertyValue("tbl.nodes")
	private String tableNode;

	@DefaultValue("%datasetName%_nodes_density")
	@NotNull
	@PropertyValue("tbl.nodes.density")
	private String tableNodeDensity;

	private String datasetName;
	private int serverSRID = -1;
	protected DbTable coverageTable;
	protected DbTable edgeTable;
	protected DbTable nodeAnnotatedTable;
	protected DbTable nodeTable;

	static {
		ALL_DATASETS = getAvailableDatasets(true);
	}

	/**
	 * Gets a validated instance of a dataset configuration.
	 *
	 * @param datasetName the name of the dataset configuration to get
	 * @return the initialized dataset configuration
	 */
	public static ConfigDataset getInstance(final String datasetName) {
		if (!isAvailable(datasetName)) {
			return null;
		}

		ConfigDataset ds = CACHE_INSTANCE.get(datasetName);
		if (ds == null) {
			ds = createInstance(datasetName);
		}

		return ds;
	}

	/**
	 * Gets all valid dataset configurations.
	 *
	 * @return all initialized dataset configurations that are valid
	 */
	public static Collection<String> getAllDatasets() {
		return ALL_DATASETS;
	}

	/**
	 * Checks by name if a given dataset is available.
	 *
	 * @param datasetName the name to check for an existing dataset
	 * @return flag indicating if a dataset with the specified name is available
	 */
	public static boolean isAvailable(final String datasetName) {
		return ALL_DATASETS.contains(datasetName);
	}

	// Constructors

	protected ConfigDataset() { }

	protected ConfigDataset(final ConfigDataset o) {
		coverageTable = o.coverageTable;
		clientArrivalDateTime = o.getClientArrivalDateTime();
		datasetName = o.datasetName;
		distanceTolerance = o.distanceTolerance;
		edgeTable = o.edgeTable;
		indexIsoCoverage = o.indexIsoCoverage;
		indexIsoLinks = o.indexIsoLinks;
		indexIsoNodes = o.indexIsoNodes;
		nodeAnnotatedTable = o.nodeAnnotatedTable;
		nodeTable = o.nodeTable;
		serverSRID = o.serverSRID;
		synthetic = o.synthetic;
		tableCalendar = o.tableCalendar;
		tableEdges = o.tableEdges;
		tableIsoCoverage = o.tableIsoCoverage;
		tableIsoLinks = o.tableIsoLinks;
		tableIsoNodes = o.tableIsoNodes;
		tableIsoNodesAnnotations = o.tableIsoNodesAnnotations;
		tableRoutes = o.tableRoutes;
		tableSchedules = o.tableSchedules;
		tableNode = o.tableNode;
		tableNodeDensity = o.tableNodeDensity;
	}

	// (Public) Getter

	public LocalDateTime getClientArrivalDateTime() {
		return clientArrivalDateTime;
	}

	public String getDatasetName() {
		return datasetName;
	}

	public int getDistanceTolerance() {
		return distanceTolerance;
	}

	public DbTable getTableCoverage() {
		return coverageTable;
	}

	public DbTable getTableEdge() {
		return edgeTable;
	}

	public DbTable getTableNodeAnnotated() {
		return nodeAnnotatedTable;
	}

	public DbTable getTableNode() {
		return nodeTable;
	}

	public int getServerSRID() {
		return serverSRID;
	}

	public String getTablenameCalendar() {
		return tableCalendar;
	}

	public String getTablenameEdges() {
		return tableEdges;
	}

	public String getTablenameNode() {
		return tableNode;
	}

	public String getTablenameNodeDensity() {
		return tableNodeDensity;
	}

	public String getTablenameRoutes() {
		return tableRoutes;
	}

	public String getTablenameSchedule() {
		return tableSchedules;
	}

	// Public methods

	@Override
	public String toString() {
		return "ConfigDataset [datasetName=" + datasetName + ", distanceTolerance=" + distanceTolerance + ", serverSRID=" + serverSRID + "]";
	}

	// Protected static methods

	/**
	 * Gets the available datasets from the system.
	 * This returns a collection containing the names of the datasets only (not the whole path).
	 * So if there are configurations "config_bz.xml" and "config_sf.xml" are available (and both are valid),
	 * this returns: <br>
	 * <pre>Arrays.asList(new String[] {"bz", "sf"});</pre>
	 *
	 * @param filterInvalid flag indicating if only valid datasets will be returned (for more information see {@link #filterInvalidDatasets(Collection)}
	 * @return the available datasets as collection of strings
	 */
	protected static Collection<String> getAvailableDatasets(final boolean filterInvalid) {
		LOGGER.info("Getting available datasets");

		final ClassLoader cLoader = Thread.currentThread().getContextClassLoader();
		final Pattern dsConfigPattern = Pattern.compile(DS_CONFIG_PATTERN);

		final Collection<String> resources = new ArrayList<>();
		try {
			// find resources on disk and add to available ones
			URI r = null;
			try {
				final URL resource = cLoader.getResource("/");
				r = resource == null ? null : resource.toURI();
			} catch (final URISyntaxException e1) {
				LOGGER.warn("Datasets directory URI could not be created. There might be no configured datasets available!");
			}

			if (r != null) {
				final String resourceStr = r.toString();

				Path path = null;
				if (resourceStr.contains("!")) {
					final String[] array = resourceStr.split("!");
					final FileSystem fs = FileSystems.newFileSystem(URI.create(array[0]), new HashMap<>());
					path = (fs == null) ? null : fs.getPath(array[1]);
				} else {
					path = Paths.get(r);
				}

				if (path != null) {
					Files.walkFileTree(path, new SimpleFileVisitor<Path>() {
						@Override
						public FileVisitResult visitFile(final Path file, final BasicFileAttributes attrs) throws IOException {
							if (attrs.isRegularFile() && file.endsWith(DS_CONFIG_EXTENSION)) {
								final Path fPath = file.getFileName();
								if (fPath != null) {
									resources.add(fPath.toString());
								}
							}
							return FileVisitResult.CONTINUE;
						}
					});
				}
			}

			// find resources on classpath (in jar files...) and add to available ones
			resources.addAll(ResourceHelper.findFilesInClassPath(dsConfigPattern));

			// find resources on webapp resources (in tomcat lib folder, in WEB-INF...)  and add to available ones
			final Enumeration<URL> e = cLoader.getResources("../lib");
			final Collection<URL> webappResources = Collections.list(e);
			final Collection<String> files = ResourceHelper.findFilesInResources(webappResources, dsConfigPattern);
			resources.addAll(files);
		} catch (final IOException e) {
			LOGGER.warn("Datasets could not be read. There might be no configured datasets available!");
		}

		if (resources.size() <= 0) {
			LOGGER.debug("   no datasets available");
			return Collections.emptyList();
		}

		final List<String> datasets = new ArrayList<>(resources.size());
		for (final String fileName : resources) {
			final Matcher m = dsConfigPattern.matcher(fileName);
			if (m.matches()) {
				final String dataset = m.group(1);
				datasets.add(dataset);
				LOGGER.debug("   adding dataset: {}", dataset);
			}
		}

		Collections.sort(datasets);
		if (filterInvalid) {
			return filterInvalidDatasets(datasets);
		}

		return datasets;
	}

	// Private methods

	private String getIndexnameIsoCoverage() {
		return indexIsoCoverage;
	}

	private String getIndexnameIsoLinks() {
		return indexIsoLinks;
	}

	private String getIndexnameIsoNodes() {
		return indexIsoNodes;
	}

	private String getTablenameIsoCoverage() {
		return tableIsoCoverage;
	}

	private String getTablenameIsoLinks() {
		return tableIsoLinks;
	}

	private String getTablenameIsoNodes() {
		return tableIsoNodes;
	}

	private String getTablenameIsoNodesAnnotations() {
		return tableIsoNodesAnnotations;
	}

	private boolean hasValidDatabase(final Collection<String> allTables) {
		final Collection<String> configTables = new ArrayList<>();

		configTables.add(getTablenameEdges());
		configTables.add(getTablenameNode());
		if (!synthetic) {
			configTables.add(getTablenameCalendar());
			configTables.add(getTablenameRoutes());
			configTables.add(getTablenameSchedule());
			configTables.add(getTablenameNodeDensity());
		}

		return allTables.containsAll(configTables);
	};

	private void initServerSRID() {
		LOGGER.debug("   getting SRID from database table for dataset {}", datasetName);
		final ConfigIsochrone config = ConfigIsochrone.getInstance();

		Connection conn = null;
		try {
			conn = config.getConnection();
			serverSRID = DbUtils.getSRID(conn, tableNode);
		} catch (final SQLException e) {
			LOGGER.warn("   could not get SRID from database table {}... defaulting to \"EPSG:{}\"", tableNode, DEFAULT_SRID);
			serverSRID = DEFAULT_SRID;
		} finally {
			DbUtils.closeQuietly(conn);
		}
	}

	private void initTableEntries() {
		coverageTable = new DbTable(getTablenameIsoCoverage(), getIndexnameIsoCoverage(), DbTableType.COVERAGE);
		edgeTable = new DbTable(getTablenameIsoLinks(), getIndexnameIsoLinks(), DbTableType.EDGE);
		nodeAnnotatedTable = new DbTable(getTablenameIsoNodesAnnotations());
		nodeTable = new DbTable(getTablenameIsoNodes(), getIndexnameIsoNodes(), DbTableType.NODE);
	}

	private void replaceDefaultPlaceholders() {
		final Map<String, String> defaultValueReplacemens = new HashMap<>();
		defaultValueReplacemens.put("%datasetName%", datasetName);

		final Set<Entry<String, String>> replacements = defaultValueReplacemens.entrySet();
		final Field[] fieldArr = this.getClass().getDeclaredFields();
		for (final Field f : fieldArr) {
			if (f.isAnnotationPresent(DefaultValue.class)) {
				try {
					final Object o = f.get(this);
					if (o instanceof String && o.equals(f.getAnnotation(DefaultValue.class).value())) {
						String str = (String) o;
						for (final Entry<String, String> e : replacements) {
							str = str.replaceAll(e.getKey(), e.getValue());
						}

						if (!str.equals(o)) {
							LOGGER.trace("   replacing default value of {} with {} inconfiguration of dataset {}", o, str, datasetName);
							f.set(this, str);
						}
					}
				} catch (IllegalArgumentException | IllegalAccessException e) {
					LOGGER.warn("   could not replace default value for field {}", f);
				}
			}
		}
	}

	// Private static methods

	private static ConfigDataset createInstance(final String datasetName) {
		final List<String> baseNames = Arrays.asList(new String[]{"config_" + datasetName});
		final Class<? extends PropertyLoaderFilter> filter = SqlPropertyFilter.class;

		@SuppressWarnings("unchecked")
		final ConfigDataset config = new ConfigBuilder<>(ConfigDataset.class).withPropertyFilters(filter).overridePropertiesFiles(baseNames).build();
		config.datasetName = datasetName;
		config.replaceDefaultPlaceholders();
		config.initTableEntries();
		config.initServerSRID();

		CACHE_INSTANCE.put(datasetName, config);
		return config;
	}

	private static Collection<String> filterInvalidDatasets(final Collection<String> datasets) {
		LOGGER.info("Filtering invalid datasets (e.g. which have no database tables)");
		final ConfigIsochrone config = ConfigIsochrone.getInstance();

		final Collection<String> allTables;
		try {
			final Connection conn = config.getConnection();
			allTables = DbUtils.getAllTables(conn);
			DbUtils.closeQuietly(conn);
		} catch (final SQLException e) {
			LOGGER.warn("Could not get tables from database... datasets will not be filtered at all");
			return datasets;
		}

		final Collection<String> filteredList = new ArrayList<>(datasets.size());
		for (final String ds : datasets) {
			final ConfigDataset c = ConfigDataset.createInstance(ds);
			if (!c.hasValidDatabase(allTables)) {
				LOGGER.warn("   not using dataset configuration \"{}\" since configured tables are not available!", ds);
				continue;
			}

			LOGGER.debug("   dataset configuration \"{}\" will be available to clients", ds);
			filteredList.add(ds);
		}

		return filteredList;
	}

}
