package it.unibz.inf.isochrone.config;

import com.tngtech.configbuilder.ConfigBuilder;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertiesFiles;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertyExtension;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertyLocations;
import com.tngtech.configbuilder.annotation.propertyloaderconfiguration.PropertySuffixes;
import com.tngtech.configbuilder.annotation.valueextractor.DefaultValue;
import com.tngtech.configbuilder.annotation.valueextractor.PropertyValue;

import it.unibz.inf.isochrone.db.DbType;

import java.sql.Connection;
import java.sql.SQLException;

import javax.validation.constraints.NotNull;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * @author <a href="mailto:nikolaus.krismer@uibk.ac.at">Nikolaus Krismer</a>
 */
@PropertiesFiles({ "config", "config-test" })
@PropertyLocations(fromClassLoader = true)
@PropertyExtension("xml")
@PropertySuffixes(extraSuffixes = { }, hostNames = false)
public final class ConfigIsochrone {
	private static final Logger LOGGER = LogManager.getLogger(ConfigIsochrone.class);
	private static ConfigIsochrone instance = null;
	private ConnectionFactory factory;

	@NotNull
	@DefaultValue("4326")
	@PropertyValue("cfg.clientSRID")
	private int clientSRID;

	@NotNull
	@PropertyValue("db.database")
	private String database;

	@NotNull
	@DefaultValue("localhost")
	@PropertyValue("db.servername")
	private String host;

	@DefaultValue("10")
	@PropertyValue("db.maxConnections")
	private int maxConnections;

	@NotNull
	@PropertyValue("db.password")
	private String password;

	@DefaultValue("false")
	@PropertyValue("db.pooling")
	private boolean pooling;

	@NotNull
	@DefaultValue("5432")
	@PropertyValue("db.port")
	private int port;

	@NotNull
	@DefaultValue("POSTGRESQL")
	@PropertyValue("db.type")
	private String type;

	@NotNull
	@PropertyValue("db.username")
	private String user;

	public static synchronized ConfigIsochrone getInstance() {
		if (instance == null) {
			final ConfigIsochrone config = new ConfigBuilder<>(ConfigIsochrone.class).build();
			config.initConnectionFactory();

			instance = config;
		}

		return instance;
	}

	// Constructors

	private ConfigIsochrone() { }

	// Getter

	public int getClientSRID() {
		return clientSRID;
	}

	public Connection getConnection() {
		if (factory == null) {
			return null;
		}

		Connection conn = null;
		try {
			conn = factory.getConnection();
		} catch (final SQLException e) {
			conn = null;
		}

		return conn;
	}

	public String getDBName() {
		return database;
	}

	public String getDBPassword() {
		return password;
	}

	public int getDBPort() {
		return port;
	}

	public String getDBServerName() {
		return host;
	}

	public DbType getDBType() {
		return DbType.valueOf(type);
	}

	public String getDBUser() {
		return user;
	}

	public int getMaxConnections() {
		return maxConnections;
	}

	public boolean isPooling() {
		return pooling;
	}

	// Public methods

	@Override
	public String toString() {
		return "ConfigIsochrone [clientSRID=" + clientSRID + ", maxConnections=" + maxConnections + ", pooling=" + pooling + "]";
	}

	// Private methods

	private void initConnectionFactory() {
		LOGGER.debug("Initializing database connection factory");
		LOGGER.debug("   serverName: {}", host);
		LOGGER.debug("   dbName: {}", database);
		LOGGER.debug("   dbPort: {}", port);
		LOGGER.debug("   dbUser: {}", user);
		LOGGER.debug("   dbPassword: {}", (password == null ? "without password" : "with password"));
		LOGGER.debug("   isPooling: {}", pooling);

		try {
			factory = new ConnectionFactory(this);
		} catch (final SQLException e) {
			throw new RuntimeException("Unable to connect to the database: " + e.getMessage(), e);
		} catch (final ClassNotFoundException e) {
			throw new RuntimeException("Class not found: " + e.getMessage(), e);
		}
	}

}
