package it.unibz.inf.isochrone.config;

import com.jolbox.bonecp.BoneCP;
import com.jolbox.bonecp.BoneCPConfig;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import it.unibz.inf.isochrone.db.DbType;
import it.unibz.inf.isochrone.db.DbUtils;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.LinkedList;
import java.util.Properties;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.json.simple.JSONObject;
import org.sqlite.SQLiteConfig;

/**
 * Used for connecting to the database server and for getting connections
 * from a connection pool (if pooling is enabled in the config file).
 *
 * @author <a href="mailto:nikolaus.krismer@uibk.ac.at">Nikolaus Krismer</a>
 */
class ConnectionFactory {
	private static final int MIN_CONNECTIONS = 5;
	private static final String PATTERN_SQLITE_INMEMORY = ":memory:\\|(\\w|/|\\.)+\\|(\\w|/|\\.)+";
	private static final String PATTERN_SQLITE = "(\\w|/|\\.)+\\|(\\w|/|\\.)+";
	private Connection connection;
	private BoneCP connectionPool;
	private ConfigIsochrone config;
	private String jdbcUrl;
	private String spatialLiteLibrary;
	private Properties sqliteProperties;
	private Properties neo4jProperties;


	// Constructor

	/**
	 * Register the connection pool with the database.
	 *
	 * @param config the ConfigIsochrone object containing the values needed to create a connection (host, name of the database
	 * (for SQLite the path to the database and to the SpatiaLite lib are given, for SQLite In-Memory the path to the backup and to the
	 * SpatiaLite lib are given), user, password, port, type)
	 * @throws SQLException thrown if the connection (or connection pool) could not be generated
	 * @throws ClassNotFoundException thrown if the driver used for the database connection could not be found/loaded
	 */
	public ConnectionFactory(final ConfigIsochrone config) throws SQLException, ClassNotFoundException {
		this.config = config;

		String[] sqliteConfig = null;
		switch (config.getDBType()) {
		case SQLITE:
			sqliteConfig = parseSqliteName(config.getDBName());
			this.jdbcUrl = "jdbc:sqlite:" + sqliteConfig[0];
			spatialLiteLibrary = sqliteConfig[1];
			Class.forName("org.sqlite.JDBC"); // load the database driver

			// enabling dynamic extension loading, absolutely required by SpatiaLite
			final SQLiteConfig sqLiteConfig = new SQLiteConfig();
			sqLiteConfig.enableLoadExtension(true);
			sqliteProperties = sqLiteConfig.toProperties();

			break;
		case NEO4J:
			this.jdbcUrl = "jdbc:neo4j://" + config.getDBServerName() + ":" + config.getDBPort() + "/";
			Class.forName("org.neo4j.jdbc.Driver");
			neo4jProperties = new Properties();
			break;
		default:
			this.jdbcUrl = "jdbc:postgresql://" + config.getDBServerName() + ":" + config.getDBPort() + "/" + config.getDBName() + "?binaryTransfer=true";
			Class.forName("org.postgresql.Driver"); // load the database driver
		}

		if (config.isPooling()) {
			final int maxConnections = config.getMaxConnections();
			final BoneCPConfig poolConfig = new BoneCPConfig();
			//			poolConfig.setCloseConnectionWatch(true); // activate this only for debugging purposes!
			poolConfig.setDefaultAutoCommit(false);
			poolConfig.setJdbcUrl(jdbcUrl);
			poolConfig.setMinConnectionsPerPartition(Math.min(MIN_CONNECTIONS, maxConnections));
			poolConfig.setMaxConnectionsPerPartition(maxConnections);
			poolConfig.setPartitionCount(1);

			if (sqliteProperties != null) {
				poolConfig.setDriverProperties(sqliteProperties);
			} else {
				poolConfig.setUser(config.getDBUser());
				poolConfig.setPassword(config.getDBPassword());
			}

			connectionPool = new BoneCP(poolConfig);
		}

		// CHECKSTYLE:OFF MagicNumber
		if (sqliteConfig != null && sqliteConfig.length >= 3) {
			// restore database from file if SQLite should be used in-memory
			intoMemory(getConnection(), sqliteConfig[2]);
		}
		// CHECKSTYLE:ON MagicNumber
	}

	// Public methods

	/**
	 * Gets a connection from the connection pool.
	 *
	 * @return the connection from the pool
	 * @throws SQLException thrown if a connection could not be gotten from the connection pool (only thrown if a connection pool is used)
	 */
	public Connection getConnection() throws SQLException {
		if (config.isPooling()) {
			if (connectionPool == null) {
				throw new NullPointerException("ConnectionFactory has not registered any connectionPool configurations");
			}

			return initConnection();
		}

		connection = initConnection();
		return connection;
	}

	@Override
	public String toString() {
		return "ConnectionFactory [pooling=" + config.isPooling() + ", jdbcUrl=" + jdbcUrl + "]";
	}

	// Private methods

	private Connection createConnection() throws SQLException {
		if (config.isPooling()) {
			return connectionPool.getConnection();
		}

		final Connection c;
		if (sqliteProperties != null) {
			c = DriverManager.getConnection(jdbcUrl, sqliteProperties);
		} else if(neo4jProperties != null){
			c = DriverManager.getConnection(jdbcUrl, neo4jProperties);
		} else {
			c = DriverManager.getConnection(jdbcUrl, config.getDBUser(), config.getDBPassword());
		}

		c.setAutoCommit(false);

		return c;
	}

	private Connection initConnection() throws SQLException {
		if (config.getDBType() == DbType.POSTGRESQL && (config.getDBUser() == null || config.getDBPassword() == null)) {
			throw new NullPointerException("ConnectionFactory has not registered any connection configurations");
		}

		if (connection != null && !connection.isClosed()) {
			return connection;
		}

		final Connection c = createConnection();
		if (config.getDBType() == DbType.SQLITE) {
			loadSpatialLite(c, spatialLiteLibrary);
		}
		if(config.getDBType() == DbType.NEO4J){

			//--------------------------------------------------------
			String layerName = "spatial_layer";
			JSONObject jsonObject = new JSONObject();
			try{
				jsonObject.put("layer", layerName);
				HttpClient client = new DefaultHttpClient();
				HttpPost post = new HttpPost("http://localhost:7474/db/data/ext/SpatialPlugin/graphdb/getLayer");
				StringEntity input = new StringEntity(jsonObject.toString());
				input.setContentType("application/json");
				post.setEntity(input);
				HttpResponse response = client.execute(post);
				//if layer was not found create it
				if(response.getStatusLine().getStatusCode() == 500){
					createLayer(layerName, c);
				}
			}catch(Exception e){
				e.printStackTrace();
			}
		}


		return c;
	}

	private void createLayer(String layerName, Connection c){
		JSONObject jsonObject = new JSONObject();
		try{
			jsonObject.put("layer", layerName);
			jsonObject.put("format", "WKB");
			jsonObject.put("nodePropertyName", "geometry");
			HttpClient client = new DefaultHttpClient();
			HttpPost post = new HttpPost("http://localhost:7474/db/data/ext/SpatialPlugin/graphdb/addEditableLayer");
			StringEntity input = new StringEntity(jsonObject.toString());
			input.setContentType("application/json");
			post.setEntity(input);
			HttpResponse response = client.execute(post);
			if(response.getStatusLine().getStatusCode() == 500){
				throw new Exception("Spatial layer could not be created");
			}else{
				createIndex(layerName, c);
			}
		}catch(Exception e){
			e.printStackTrace();
		}
	}


	private void createIndex(String layerName, Connection c){
		JSONObject jsonObject = new JSONObject();
		try{
			jsonObject.put("name", layerName);
			JSONObject jsonConfig = new JSONObject();
			jsonConfig.put("provider", "spatial");
			jsonConfig.put("wkb", "geometry");
			jsonObject.put("config", jsonConfig);
			HttpClient client = new DefaultHttpClient();
			HttpPost post = new HttpPost("http://localhost:7474/db/data/index/node/");
			StringEntity input = new StringEntity(jsonObject.toString());
			input.setContentType("application/json");
			post.setEntity(input);
			HttpResponse response = client.execute(post);
			if(response.getStatusLine().getStatusCode() == 500){
				throw new Exception("Spatial index could not be created");
			}else{
				encodeNodes(layerName, c);
			}
		}catch(Exception e){
			e.printStackTrace();
		}
	}

	private void encodeNodes(String layerName, Connection c){
		JSONObject jsonObject = new JSONObject();
		try {
			jsonObject.put("label", "Node");
			jsonObject.put("layer", layerName);
			HttpClient client = new DefaultHttpClient();
			HttpPost post = new HttpPost("http://localhost:7474/db/data/ext/SpatialPlugin/graphdb/encodeGeometryNodesByLabel");
			StringEntity input = new StringEntity(jsonObject.toString());
			input.setContentType("application/json");
			post.setEntity(input);
			HttpResponse response = client.execute(post);
			if(response.getStatusLine().getStatusCode() == 500){
				throw new Exception("Node could not be encoded");
			}
			HttpEntity e = response.getEntity();
			EntityUtils.consume(response.getEntity());
			jsonObject.clear();
			jsonObject.put("label", "Edge");
			jsonObject.put("layer", layerName);
			post = new HttpPost("http://localhost:7474/db/data/ext/SpatialPlugin/graphdb/encodeGeometryNodesByLabel");
			input = new StringEntity(jsonObject.toString());
			input.setContentType("application/json");
			post.setEntity(input);
			response = client.execute(post);
			if(response.getStatusLine().getStatusCode() == 500){
				throw new Exception("Node could not be encoded");
			}
			
			addNodesToIndex(layerName, c);
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (ClientProtocolException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	private void addNodesToIndex(String layerName, Connection c){
		try{
			JSONObject jsonObject = new JSONObject();
			jsonObject.put("label", "Node");
			jsonObject.put("layer", layerName);
			HttpClient client = new DefaultHttpClient();
			HttpPost post = new HttpPost("http://localhost:7474/db/data/ext/SpatialPlugin/graphdb/addNodesByLabelToSpatialIndex");
			StringEntity input = new StringEntity(jsonObject.toString());
			input.setContentType("application/json");
			post.setEntity(input);
			HttpResponse response = client.execute(post);
			if(response.getStatusLine().getStatusCode() == 500){
				String body = "";
				HttpEntity e = response.getEntity();
				if(e != null){
					body = EntityUtils.toString(e);
				}
				throw new Exception("Node could not be added: " + body);

			}
			addEdgesToIndex(layerName, c);
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (ClientProtocolException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void addEdgesToIndex(String layerName, Connection c){
		try {
			JSONObject jsonObject = new JSONObject();
			jsonObject.put("label", "Edge");
			jsonObject.put("layer", layerName);
			HttpClient client = new DefaultHttpClient();
			HttpPost post = new HttpPost("http://localhost:7474/db/data/ext/SpatialPlugin/graphdb/addNodesByLabelToSpatialIndex");
			StringEntity input = new StringEntity(jsonObject.toString());
			input.setContentType("application/json");
			post.setEntity(input);
			HttpResponse response = client.execute(post);
			if(response.getStatusLine().getStatusCode() == 500){
				String body = "";
				HttpEntity e = response.getEntity();
				if(e != null){
					body = EntityUtils.toString(e);
				}
				throw new Exception("Node could not be added: " + body);

			}
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (ClientProtocolException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	// Private static methods

	private static void intoMemory(final Connection c, final String dbFile) throws SQLException {
		try (final Statement stmt = c.createStatement()) {
			stmt.executeUpdate("RESTORE FROM " + dbFile);
		}
	}

	@SuppressFBWarnings(
			value = {"SQL_NONCONSTANT_STRING_PASSED_TO_EXECUTE"},
			justification = "Since table names can not be filled with parameters, but we have variable table names we somewhere need to do this"
			)
	private static void loadSpatialLite(final Connection c, final String pathToLib) throws SQLException {
		// loading SpatiaLite
		try (final Statement stmt = c.createStatement()) {
			stmt.execute("SELECT load_extension('" + pathToLib + "')");
		}
	}

	private static String[] parseSqliteName(final String name) throws SQLException {
		if (!name.matches(PATTERN_SQLITE_INMEMORY) && !name.matches(PATTERN_SQLITE)) {
			throw new SQLException("Database configuration is invalid (name is unexpected)");
		}

		String connectionString = name.substring(0, name.indexOf('|'));
		final String pathToSpatiaLite = name.substring(name.lastIndexOf('|') + 1);

		if (name.matches(PATTERN_SQLITE_INMEMORY)) {
			connectionString = "file:" + connectionString + "?cache=shared;";
			final String pathToSqliteDb = name.substring(name.indexOf('|') + 1, name.lastIndexOf('|'));

			return new String[] {connectionString, pathToSpatiaLite, pathToSqliteDb};
		}

		return new String[] {connectionString, pathToSpatiaLite};
	}

}
