package it.unibz.inf.isochrone.db;

import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.WKBReader;

import java.awt.geom.Point2D;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.time.DayOfWeek;
import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.util.EntityUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import it.unibz.inf.isochrone.AlgorithmRunner;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.config.ConfigIsochrone;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Location;
import it.unibz.inf.isochrone.network.Node;
import it.unibz.inf.isochrone.network.NodeConnection;
import it.unibz.inf.isochrone.network.Seed;

/**
 * <p>
 * The database layer.<br>
 * Contains all database queries for the different algorithms.
 * All the prepared statements are cached for optimizing the access times, so the database doesn't have to recalculate the
 * Query plan.
 * Remember to NOT close statements retrieved by {@link #getPstmt(String)} for this to work.
 * The statements will be closed on execution of the {@link #close()} method.
 * </p>
 * <p>
 * Copyright: 2006 - 2009 <a href="http://www.inf.unibz.it/dis">Dis Research Group</a>
 * <br>
 * Domenikanerplatz - Bozen, Italy.
 * </p>
 *
 * @author Markus Innerebner
 * @version 2.2
 */
@SuppressFBWarnings(
		value = {"SQL_PREPARED_STATEMENT_GENERATED_FROM_NONCONSTANT_STRING", "SQL_NONCONSTANT_STRING_PASSED_TO_EXECUTE"},
		justification = "Since table names can not be filled with parameters, but we have variable table names we somewhere need to do this"
		)
public class Database implements AutoCloseable {
	private static final Logger LOGGER = LogManager.getLogger(AlgorithmRunner.class.getName());
	private static final String[] PRECACHED_PLACEHOLDERS;
	protected static final byte NW_MODE_CONTINUOUS = 0;
	protected static final byte NW_MODE_DISCRETE = 1;
	protected final ConfigDataset config;
	protected final boolean isMultimodal;
	protected final boolean isIncoming;
	protected final WKBReader wkbReader;
	private final AbstractDbQuery query;
	
	public List<Long> getCoordinateTime = new LinkedList<Long>();
	public List<Long> getCalendarTime = new LinkedList<Long>();
	public List<Long> getLinkTime = new LinkedList<Long>();
	public List<Long> getLinksTime = new LinkedList<Long>();
	public List<Long> getLinksInRangeTime = new LinkedList<Long>();
	public List<Long> getNodeTravellingCostsTime = new LinkedList<Long>();
	public List<Long> getRangeByNodeCountTime = new LinkedList<Long>();
	public List<Long> getProjectOnLinksTime = new LinkedList<Long>();
	public List<Long> readNetworkTime = new LinkedList<Long>();
	
	private Map<String, PreparedStatement> pstmtsCacheMap;
	/** The connection the the database. It is lazy-loaded by {@link #getPstmt(String)} method. */
	private Connection connection;

	// Static block

	static {

		final int preachingLength;
		switch (ConfigIsochrone.getInstance().getDBType()) {
		case NEO4J:
			preachingLength = 0;
			break;
		default:
			preachingLength = 10;
		}
		PRECACHED_PLACEHOLDERS = new String[preachingLength];
		for (int i = 0; i < preachingLength; ++i) {
			PRECACHED_PLACEHOLDERS[i] = calculatePlaceHolders(i, 0);
		}
	}

	// Constructor

	public Database(final ConfigDataset config, final Mode mode, final Direction direction) {
		this(config, mode, direction, false);
	}

	protected Database(final ConfigDataset config, final Mode mode, final Direction direction, final boolean fetchGeometries) {
		this.config = config;
		this.isIncoming = direction == Direction.INCOMING;
		this.isMultimodal = mode == Mode.MULTIMODAL;
		this.wkbReader = (fetchGeometries) ? new WKBReader() : null;
		switch (ConfigIsochrone.getInstance().getDBType()) {
		case SQLITE:
			this.query = new DbQuerySqlite(config, isIncoming, fetchGeometries);
			break;
		case NEO4J:
			this.query = new DbQueryNeo4j(config, isIncoming, fetchGeometries);
			break;
		default:
			this.query = new DbQueryPostgres(config, isIncoming, fetchGeometries);
		}
	}

	// Public methods

	public void activateStatistics() {
		query.activateStatistics();
	}

	/**
	 * Closes all the open prepared statements and the database connection.
	 * This should be called once you're done with the database.
	 */
	@Override
	public void close() {
		clearCacheMap();
		DbUtils.closeQuietly(connection);
	}

	/**
	 * Gets the dataset configuration that was used during initializing of the database queries.
	 *
	 * @return the dataset configuration initially used
	 */
	public ConfigDataset getConfigDataset() {
		return config;
	}

	/**
	 * Gets the coordinates of a node.
	 *
	 * @param nodeId the nodeId of the node for which we should find the coordinates
	 * @return a point with the coordinates of the node
	 */
	public Point2D getCoordinates(final int nodeId) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("getCoordinates: " + nodeId);
		final PreparedStatement stmt = getPstmt(query.getNodeCoordinates());
		Point2D gp = null;
		ResultSet rs = null;
		try {
			switch (ConfigIsochrone.getInstance().getDBType()) {
			case NEO4J:
				JSONObject jsonObject = new JSONObject();
				JSONParser parser = new JSONParser();
				final String queryStr = String.format(query.getNodeCoordinates(), "Node", nodeId);
				jsonObject.put("query", queryStr);
				HttpClient client = new DefaultHttpClient();
				HttpPost post = new HttpPost("http://localhost:7474/db/data/cypher");
				StringEntity input = new StringEntity(jsonObject.toString());
				input.setContentType("application/json");
				post.setEntity(input);
				HttpResponse response = client.execute(post);
				HttpEntity entity = response.getEntity();
				if(entity != null){
					String retSrc = EntityUtils.toString(entity); 
					// parsing JSON
					JSONObject objects = (JSONObject) parser.parse(retSrc);
					for(Object o : (JSONArray) objects.get("data")){
						final double lat = (double) ((JSONArray) o).get(0);
						final double lon = (double) ((JSONArray) o).get(1);
						LOGGER.debug("lat: " + lat + ", lon: " + lon);
						gp = new Point2D.Double(lat, lon);
					}
				}
				break;
			default:
				stmt.setInt(1, nodeId);
				rs = stmt.executeQuery();
				if (rs.next()) {
					LOGGER.debug("lat: " + rs.getDouble("X") + ", lon: " + rs.getDouble("Y"));
					gp = new Point2D.Double(rs.getDouble("X"), rs.getDouble("Y"));
				}
			}
		} catch (final SQLException e) {
			handleSqlException(e);
		} catch (org.json.simple.parser.ParseException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (ClientProtocolException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			DbUtils.closeQuietly(rs);
		}

		getCoordinateTime.add(System.currentTimeMillis()-l0);
		return gp;
	}

	/**
	 * Gets calendar entries for a given time span.
	 *
	 * @param startDate begin of the time span specified
	 * @param endDate end of the time span specified
	 * @return returns a set of service ids
	 */
	public Set<Integer> getCalendar(final LocalDate startDate, final LocalDate endDate) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("getCalendar: " + startDate + ", " + endDate);
		if (startDate == null || endDate == null) {
			return null;
		}

		final PreparedStatement stmt = getPstmt(query.getCalendar());
		final Set<Integer> codes = new LinkedHashSet<>();

		ResultSet rs = null;
		try {
			switch (ConfigIsochrone.getInstance().getDBType()) {
			case NEO4J:
				boolean[] weekDayMask = getWeekDayMask(startDate, endDate);
				stmt.setBoolean(1, weekDayMask[0]);
				stmt.setBoolean(2, weekDayMask[1]);
				stmt.setBoolean(3, weekDayMask[2]);
				stmt.setBoolean(4, weekDayMask[3]);
				stmt.setBoolean(5, weekDayMask[4]);
				stmt.setBoolean(6, weekDayMask[5]);
				stmt.setBoolean(7, weekDayMask[6]);
				stmt.setString(8, startDate.toString());
				stmt.setString(9, startDate.toString());
				break;
			default:
				// CHECKSTYLE:OFF MagicNumber
				stmt.setString(1, mask2String(getWeekDayMask(startDate, endDate)));
				stmt.setDate(2, Date.valueOf(startDate));
				stmt.setDate(3, Date.valueOf(endDate));
				// CHECKSTYLE:ON MagicNumber
			}


			rs = stmt.executeQuery();
			while (rs.next()) {
				LOGGER.debug("retireved service_id: " + rs.getInt("SERVICE_ID"));
				codes.add(rs.getInt("SERVICE_ID"));
			}
		} catch (final SQLException e) {
			handleSqlException(e);
		} finally {
			DbUtils.closeQuietly(rs);
		}

		getCalendarTime.add(System.currentTimeMillis() - l0);
		return codes;
	}

	/**
	 * Get a single link from the network and and add the source node to the nodes table if necessary.
	 *
	 * @param linkId the id of the link that should be returned
	 * @param nodes the nodes table to which the node should be added (can be null if the link is needed)
	 * @return the link with id linkId
	 */
	public Link getLink(final int linkId, final Map<Integer, Node> nodes) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("getLink: " + linkId + ", " + nodes.size());
		final PreparedStatement stmt = getPstmt(isMultimodal ? query.getLink() : query.getLinkContinuous());

		Link link = null;
		ResultSet rs = null;
		try {
			stmt.setInt(1, linkId);
			rs = stmt.executeQuery();
			if (rs.next()) {
				//				final int id = rs.getInt("ID");
				final int sourceId = rs.getInt("SOURCE");
				final int targetId = rs.getInt("TARGET");
				final short degree = rs.getShort("NODE_DEGREE");
				final double length = rs.getDouble("LENGTH");
				LOGGER.debug("source: "+ sourceId + ", target: "+ targetId + ", degree: " + degree + ", length: " + length);

				link = fetchGeometry(new Link(linkId, sourceId, targetId, length), rs);
				if (nodes != null) {
					final LineString geom = link.getGeometry();
					final Node node = updateNode(nodes, (isIncoming) ? sourceId : targetId, degree);
					if (geom != null) {
						node.setGeometry((isIncoming) ? geom.getStartPoint() : geom.getEndPoint());
					}
				}
			}
		} catch (final SQLException e) {
			handleSqlException(e);
		} finally {
			DbUtils.closeQuietly(rs);
		}

		getLinkTime.add(System.currentTimeMillis() - l0);
		return link;
	}

	/**
	 * Gets all links adjacent to a given node.
	 *
	 * @param node the node to which the links should be adjacent
	 * @param nodes the nodes HashMap to which eventual new nodes should be added
	 * @return a list of links adjacent to the given nodeId
	 */
	public Collection<Link> getLinks(final Node node, final Map<Integer, Node> nodes) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("getLinks: " + node + ", " + nodes.size());
		final int nodeId = node.getId();
		final PreparedStatement stmt = getPstmt(isMultimodal ? query.getLinks() : query.getLinksContinuous());
		final Collection<Link> links = new LinkedHashSet<>();

		ResultSet rs = null;
		try {
			stmt.setInt(1, nodeId);
			rs = stmt.executeQuery();
			while (rs.next()) {
				final int aNodeId = rs.getInt("NODE_ID");
				LOGGER.debug("node ID: " + aNodeId);
				int start, end;
				if (isIncoming) {
					start = aNodeId;
					end = nodeId;
				} else {
					start = nodeId;
					end = aNodeId;
				}

				final Link link = createLinkFromResultSet(rs, start, end);
				final LineString geom = link.getGeometry();
				links.add(link);

				final Node n = updateNode(nodes, aNodeId, rs.getShort("NODE_DEGREE"));
				if (geom != null) {
					n.setGeometry((isIncoming) ? geom.getStartPoint() : geom.getEndPoint());
				}
			}
		} catch (final SQLException e) {
			handleSqlException(e);
		} finally {
			DbUtils.closeQuietly(rs);
		}

		getLinksTime.add(System.currentTimeMillis()-l0);
		return links;
	}

	/**
	 * Loads the links in a circle from the database (for better explanation see the {@link it.unibz.inf.isochrone.algorithm.MineRX} algorithm).
	 *
	 * @param node the node for which the links will be loaded
	 * @param intersectionNodes if set then nodes will be loaded using ST_Difference  by using the nodes given
	 * @param nodes the nodes belonging to the links will be added to this map (if not already in there)
	 * @return the loaded links
	 */
	public Map<Integer, Collection<Link>> getLinksInRange(final Seed node, final Collection<Seed> intersectionNodes, final Map<Integer, Node> nodes) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("get links in range: " + node + ", " + intersectionNodes + ", "+ nodes.size());
		final Map<Integer, Collection<Link>> results = new HashMap<>();
		PreparedStatement stmt = null;
		ResultSet rs = null;
		final int srid = config.getServerSRID();
		final String pointString = DbUtils.toPointStr(node.getCenter());
		try {
			final Seed[] loadedNodes = intersectionNodes.toArray(new Seed[intersectionNodes.size()]);
			final HttpResponse response;
			final JSONParser parser = new JSONParser();
			switch(ConfigIsochrone.getInstance().getDBType()) {
			case NEO4J:
				final String queryStr; 
				if (intersectionNodes == null || intersectionNodes.isEmpty()) {
					JSONObject jsonObject = new JSONObject();
					queryStr = String.format(query.getLinksInRange(), "Node", pointString, "" + node.getRadius());
					jsonObject.put("query", queryStr);
					HttpClient client = new DefaultHttpClient();
					HttpPost post = new HttpPost("http://localhost:7474/db/data/cypher");
					StringEntity input = new StringEntity(jsonObject.toString());
					input.setContentType("application/json");
					post.setEntity(input);
					response = client.execute(post);
				}
				else{
					JSONObject jsonObject = new JSONObject();
					queryStr = String.format(query.getLinksInIntersection(), "Node", pointString, node.getRadius(), 
							"" + DbUtils.toPointStr(loadedNodes[loadedNodes.length - 1].getCenter()), "" + loadedNodes[loadedNodes.length-1].getRadius(), 
							loadedNodes.length > 1 ? "" + DbUtils.toPointStr(loadedNodes[loadedNodes.length - 2].getCenter()) : "", loadedNodes.length > 1 ? "" + loadedNodes[loadedNodes.length - 2].getRadius() : "");
					jsonObject.put("query", queryStr);
					HttpClient client = new DefaultHttpClient();
					HttpPost post = new HttpPost("http://localhost:7474/db/data/cypher");
					StringEntity input = new StringEntity(jsonObject.toString());
					input.setContentType("application/json");
					post.setEntity(input);
					response = client.execute(post);
				}
				HttpEntity entity = response.getEntity();
				if(entity != null){
					String retSrc = EntityUtils.toString(entity); 
					// parsing JSON
					JSONObject objects = (JSONObject) parser.parse(retSrc);
					for(Object o : (JSONArray) objects.get("data")){
						final int sourceId =  new BigDecimal((long) ((JSONArray) o).get(1)).intValueExact();
						final int targetId =  new BigDecimal((long) ((JSONArray) o).get(2)).intValueExact();
						final short degree = new BigDecimal((long) ((JSONArray) o).get(6)).shortValueExact();
						LOGGER.debug("source: " + sourceId + ", target: " + targetId + ", degree" + degree + ", "
								+ "id: " + ((JSONArray) o).get(0) + ", length: " + ((JSONArray) o).get(3) + ", edgeMode: " + ((JSONArray) o).get(4) + ", routeId: " + ((JSONArray) o).get(5));
						final Link link = createLinkFromJSONArray((JSONArray) o, sourceId, targetId);
						addLink(results, (isIncoming) ? link.getEndNode() : link.getStartNode(), link);
						updateNodes(nodes, link, degree);
					}
				}
				break;
			default:
				if (intersectionNodes == null || intersectionNodes.isEmpty()) {
					stmt = getPstmt(query.getLinksInRange());
					// CHECKSTYLE:OFF MagicNumber
					stmt.setString(1, pointString);
					stmt.setInt(2, srid);
					stmt.setDouble(3, node.getRadius());
					// CHECKSTYLE:ON MagicNumber
				} else {
					final String disjunctAreaString = "ST_Difference(ST_Buffer(ST_PointFromText('" + pointString + "',"
							+ srid + ")," + node.getRadius() + ")," + getSeedBufferSubQuery(loadedNodes, loadedNodes.length - 1, srid) + ")";
					stmt = getPstmt(String.format(query.getLinksInIntersection(), disjunctAreaString));
				}
				rs = stmt.executeQuery();
				while (rs.next()) {
					final short degree = rs.getShort("NODE_DEGREE");
					final int sourceId = rs.getInt("SOURCE");
					final int targetId = rs.getInt("TARGET");
					LOGGER.debug("source: " + sourceId + ", target: " + targetId + ", degree" + degree + ", "
							+ "id: " + rs.getInt("ID") + ", length: " + rs.getDouble("LENGTH") + ", edgeMode: " + rs.getByte("EDGE_MODE") + ", routeId: " + rs.getInt("ROUTE_ID"));
					final Link link = createLinkFromResultSet(rs, sourceId, targetId);
					addLink(results, (isIncoming) ? link.getEndNode() : link.getStartNode(), link);
					updateNodes(nodes, link, degree);
				}
			}
		} catch (final SQLException e) {
			System.err.println("Problems with node:" + node);
			handleSqlException(e);
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (ClientProtocolException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (org.json.simple.parser.ParseException e) {
			e.printStackTrace();
		} finally {
			DbUtils.closeQuietly(rs);
		}

		getLinksInRangeTime.add(System.currentTimeMillis()-l0);
		return results;
	}

	/**
	 * Gets the costs to go from one node to adjacent nodes.
	 *
	 * @param nConnection information about the the sourceNode, its links and routes
	 * @param serviceIds the days that should be taken into the account
	 * @param fromTime the earliest time for which we want to calculate the departure time
	 * @param toTime the latest time for which we want to calculate the departure time
	 * @return the cost of for traveling from the start node to adjacent nodes (map key == nodeId of adjacent node; value == travelling cost)
	 */
	public Map<Node, Long> getNodeTravellingCosts(final NodeConnection nConnection, final Collection<Integer> serviceIds, final long fromTime, final long toTime) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("getNodeTravellingCost: " + nConnection + ", " + serviceIds.size() + ", " + fromTime + ", " + toTime);
		final Object[] arrServiceId = (serviceIds == null) ? null : serviceIds.toArray();

		final Node node = nConnection.getSourceNode();
		final Set<Entry<Node, Collection<Link>>> entries = nConnection.getTargetEntries().entrySet();
		final Map<Node, Long> results = new LinkedHashMap<>(entries.size());
		for (final Entry<Node, Collection<Link>> entry : entries) {
			
			final Node targetNode = entry.getKey();
			final Collection<Integer> routes = NodeConnection.getTargetRoutes(entry.getValue());
			ResultSet rs = null;
			try {
				if (isIncoming) {
					rs = getLatestDepartureTime(targetNode.getId(), node.getId(), routes, arrServiceId, fromTime, Math.round(toTime - node.getDistance()));
					while (rs.next()) {
						final long departureTime = rs.getLong("TIME_D");
						final long arrivalTime = rs.getLong("TIME_A");
						final short routeId = rs.getShort("ROUTE_ID");
						LOGGER.debug("time_d: " + departureTime + ", arrivalTime: " + arrivalTime + ", route: " + routeId);
						
						targetNode.setDepartureTime(routeId, departureTime);
						node.setArrivalTime(routeId, arrivalTime);
						final long distance = toTime - departureTime;
						updateDistanceMap(results, routeId, targetNode, distance);
					}
				} else {
					rs = getEarliestArrivalTime(node.getId(), targetNode.getId(), routes, arrServiceId, Math.round(fromTime + node.getDistance()), toTime);
					while (rs.next()) {
						final long departureTime = rs.getLong("TIME_D");
						final long arrivalTime = rs.getLong("TIME_A");
						final short routeId = rs.getShort("ROUTE_ID");
						LOGGER.debug("time_d: " + departureTime + ", arrivalTime: " + arrivalTime + ", route: " + routeId);
						
						targetNode.setArrivalTime(routeId, arrivalTime);
						node.setDepartureTime(routeId, departureTime);
						final long distance = arrivalTime > 0 ? arrivalTime - fromTime : Long.MAX_VALUE;
						updateDistanceMap(results, routeId, targetNode, distance);
					}
				}
			} catch (final SQLException e) {
				handleSqlException(e);
			} finally {
				DbUtils.closeQuietly(rs);
			}
		}
		getNodeTravellingCostsTime.add(System.currentTimeMillis()-l0);
		return results;
	}

	/**
	 * Returns the maximal range for the given node with a number of specified vertices.
	 * The database query used finds the last two tuples having a memory size smaller than the specified memorySize and then interpolates the range.
	 *
	 * @param nodeId the node identifier
	 * @param nodeCount the number of vertices to load
	 * @return the available range as the euclidean distance
	 */
	public double getRangeByNodeCount(final int nodeId, final long nodeCount) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("getRangeByNodeCount: " + nodeId + ", " + nodeCount);
		final PreparedStatement stmt = getPstmt(query.getRangeByNodeCount());

		Integer densityMin = null;
		Integer densityMax = null;
		Double distanceMin = null;
		Double distanceMax = null;

		ResultSet rs = null;
		try {
			// CHECKSTYLE:OFF MagicNumber
			stmt.setInt(1, nodeId);
			stmt.setLong(2, nodeCount);
			stmt.setInt(3, nodeId);
			stmt.setLong(4, nodeCount);
			// CHECKSTYLE:ON MagicNumber
			rs = stmt.executeQuery();

			while (rs.next()) {
				final String mode = rs.getString("mode");
				if ("max".equalsIgnoreCase(mode)) {
					densityMax = rs.getInt("density");
					distanceMax = rs.getDouble("e_distance");
				} else if ("min".equalsIgnoreCase(mode)) {
					densityMin = rs.getInt("density");
					distanceMin = rs.getDouble("e_distance");
				}
			}
		} catch (final SQLException e) {
			handleSqlException(e);
		} finally {
			DbUtils.closeQuietly(rs);
		}
		// query did not return a single result. This should never happen!
		if (distanceMin == null && distanceMax == null) {
			getRangeByNodeCountTime.add(System.currentTimeMillis()-l0);
			return 0d;
		}
		
		// density was either below the smallest stored density or above maximal density
		// -> return distance without interpolation (we simply do not have enough information for such an interpolation)
		if (distanceMin == null) {
			getRangeByNodeCountTime.add(System.currentTimeMillis()-l0);
			return distanceMax.doubleValue();
		}
		if (distanceMax == null) {
			getRangeByNodeCountTime.add(System.currentTimeMillis()-l0);
			return distanceMin.doubleValue();
		}

		// values are identical (we exactly queried for a density that is stored in the DB)
		// -> return distance without interpolation
		if (distanceMin.compareTo(distanceMax) == 0) {
			getRangeByNodeCountTime.add(System.currentTimeMillis()-l0);
			return distanceMin;
		}
		LOGGER.debug("densityMin: "+ densityMin + ", distanceMin: " + distanceMin + ", densityMax: "+ densityMax + " distanceMax: " + distanceMax);
		getRangeByNodeCountTime.add(System.currentTimeMillis()-l0);
		return distanceMin + (distanceMax - distanceMin) / (densityMax - densityMin) * (nodeCount - densityMin);
	}
	
	public Map<String, Long> getStatistics() {
		return query.getStatistics();
	}

	/**
	 * Finds the nearest links to a specified point within the network with the configured default SRID.
	 *
	 * @param qNode the point to find the nearest links for
	 * @return list containing the links that the specified point was projected to.
	 *         This list is ordered by the distance between the link and the point (nearest links first)
	 */
	public List<Location> projectOnLinks(final Point2D qNode) {
		final int clientSRID = ConfigIsochrone.getInstance().getClientSRID();
		return projectOnLinks(qNode, clientSRID);
	}

	/**
	 * Finds the nearest links to a specified point within the network and the specified SRID.
	 *
	 * @param qNode the point to find the nearest links for
	 * @param clientSRID the SRID the given node is represented in
	 * @return list containing the links that the specified point was projected to.
	 *         This list is ordered by the distance between the link and the point (nearest links first)
	 */
	public List<Location> projectOnLinks(final Point2D qNode, final int clientSRID) {	
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("projectOnLinks: " + qNode + ", " + clientSRID);
		final List<Location> locations = new ArrayList<>();
		final int serverSRID = config.getServerSRID();
		final boolean isConversionSRID = clientSRID != serverSRID;
		final int tolerance = config.getDistanceTolerance();

		String pFromTxt = "ST_PointFromText(?, ?)";
		if (isConversionSRID) {
			pFromTxt = "ST_Transform(" + pFromTxt + ", ?)";
		}

		final String queryStr;
		switch(ConfigIsochrone.getInstance().getDBType()) {
		case NEO4J:
			queryStr = String.format(query.getProjectToLocation(), "Edge", DbUtils.toPointStr(qNode), clientSRID, NW_MODE_CONTINUOUS, tolerance);
			break;
		default:
			queryStr = String.format(query.getProjectToLocation(), config.getTablenameEdges(), pFromTxt, NW_MODE_CONTINUOUS);
		}

		Statement stmt = null;
		PreparedStatement statement	 = null;
		ResultSet rSet = null;
		try {
			switch(ConfigIsochrone.getInstance().getDBType()) {
			case NEO4J:
				JSONObject jsonObject = new JSONObject();
				JSONParser parser = new JSONParser();
				jsonObject.put("query", queryStr);
				HttpClient client = new DefaultHttpClient();
				HttpPost post = new HttpPost("http://localhost:7474/db/data/cypher");
				StringEntity input = new StringEntity(jsonObject.toString());
				input.setContentType("application/json");
				post.setEntity(input);
				HttpResponse response = client.execute(post);
				HttpEntity entity = response.getEntity();
				if(entity != null){
					String retSrc = EntityUtils.toString(entity); 
					// parsing JSON
					JSONObject objects = (JSONObject) parser.parse(retSrc);
					for(Object o : (JSONArray) objects.get("data")){
						final int link = new BigDecimal((long) ((JSONArray) o).get(0)).intValueExact();
						final double offset = (double) ((JSONArray) o).get(1);
						LOGGER.debug("link: " + link + ", offset: " + offset);
						locations.add(new Location(link, offset));
					}
				}
				break;
			default:
				statement = getPstmt(queryStr);
				statement.setString(1, DbUtils.toPointStr(qNode));
				statement.setInt(2, clientSRID);

				// CHECKSTYLE:OFF MagicNumber
				if (isConversionSRID) {
					statement.setInt(3, config.getServerSRID());
					statement.setInt(4, tolerance);
				} else {
					statement.setInt(3, tolerance);
				}
				// CHECKSTYLE:ON MagicNumber
				rSet = statement.executeQuery();
				while (rSet.next()) {
					final int link = rSet.getInt("ID");
					final double offset = rSet.getDouble("OFFSET");
					LOGGER.debug("link: " + link + ", offset: " + offset);
					locations.add(new Location(link, offset));
				}
			}

		} catch (final SQLException e) {
			handleSqlException(e);
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		} catch (ClientProtocolException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (org.json.simple.parser.ParseException e) {
			e.printStackTrace();
		} finally {
			DbUtils.closeQuietly(rSet);
		}
		getProjectOnLinksTime.add(System.currentTimeMillis()-l0);
		return locations;
	}

	/**
	 * Reads the whole network for the MDijkstra Algorithm.
	 * The given maps have to be initialized and will be filled up with
	 * the nodes and links in the network.
	 *
	 * @param nodes A map in which all the nodes should be stored.
	 * @param links A map in which all the links should be stored.
	 */
	public void readNetwork(final Map<Integer, Node> nodes, final Map<Integer, Link> links) {
		final long l0 = System.currentTimeMillis();
		LOGGER.debug("readNetwork: " + nodes.size() + ", " + links.size());
		final PreparedStatement stmt = getPstmt(isMultimodal ? query.getAllEdges() : query.getAllEdgesContinuous());

		ResultSet rs = null;
		try {
			rs = stmt.executeQuery();
			while (rs.next()) {
				final int startNodeId = rs.getInt("SOURCE");
				final int endNodeId = rs.getInt("TARGET");
				LOGGER.debug("source: " + startNodeId + ", target: " + endNodeId);
				// handle link
				final Link link = createLinkFromResultSet(rs, startNodeId, endNodeId);
				final LineString geom = link.getGeometry();
				links.put(link.getId(), link);

				// handle node
				int start, end;
				Point pStart, pEnd;
				if (isIncoming) {
					start = startNodeId;
					end = endNodeId;
					pStart = (geom == null) ? null : geom.getStartPoint();
					pEnd = (geom == null) ? null : geom.getEndPoint();
				} else {
					start = endNodeId;
					end = startNodeId;
					pStart = (geom == null) ? null : geom.getEndPoint();
					pEnd = (geom == null) ? null : geom.getStartPoint();
				}

				Node node = nodes.get(end);
				if (node == null) {
					node = new Node(end);
					node.setGeometry(pEnd);
					nodes.put(end, node);
				}

				if (node.getLinks() == null) {
					node.setLinks(new LinkedHashSet<Link>());
				}

				node.addLink(link);
				if (!nodes.containsKey(start)) {
					node = new Node(start);
					node.setGeometry(pStart);
					nodes.put(start, node);
				}
			}
		} catch (final SQLException e) {
			throw new RuntimeException("Error reading the Network: " + e);
		} finally {
			DbUtils.closeQuietly(rs);
		}
		readNetworkTime.add(System.currentTimeMillis()-l0);
	}

	// Protected methods

	/**
	 * Gets a prepared statement and caches it in a map. This
	 * is much more efficient than creating a new prepared statement
	 * each time a query needs to be called, and recalculating the
	 * queryplan each time.
	 *
	 * @param sql The prepared statement that is needed
	 * @return the cached prepared statement object
	 */
	// FIXME: Find a fix for prepared statements with parameters
	// (we should NOT set strings in preparedStatemens just because we want to cache something)
	protected PreparedStatement getPstmt(final String sql) {
		if (pstmtsCacheMap == null) {
			pstmtsCacheMap = new HashMap<>();
		}

		PreparedStatement pstmt = pstmtsCacheMap.get(sql);
		try {
			if (pstmt != null && !pstmt.isClosed()) {
				return pstmt;
			}

			pstmt = getConnection().prepareStatement(sql);
		} catch (final SQLException e) {
			e.printStackTrace();
			pstmt = null;
		}

		if (pstmt != null) {
			pstmtsCacheMap.put(sql, pstmt);
		}

		return pstmt;
	}

	protected Connection getConnection() {
		if (connection == null) {
			connection = ConfigIsochrone.getInstance().getConnection();
		} else {
			try {
				if (connection.isClosed()) {
					connection = ConfigIsochrone.getInstance().getConnection();
				}
			} catch (final SQLException e) {
				e.printStackTrace();
				connection = null;
			}
		}

		return connection;
	}

	protected void handleSqlException(final SQLException exception) {
		DbUtils.handleSqlException(getConnection(), exception);
	}

	// Package private methods

	void clearCacheMap() {
		if (pstmtsCacheMap == null) {
			return;
		}

		for (final PreparedStatement pstmt : pstmtsCacheMap.values()) {
			DbUtils.closeQuietly(pstmt);
		}

		pstmtsCacheMap = null;
	}

	AbstractDbQuery getQuery() {
		return query;
	}

	// Static package private methods

	static String mask2String(final boolean[] mask) {
		final StringBuilder b = new StringBuilder();
		for (final boolean m : mask) {
			b.append(m ? "1" : "0");
		}

		return b.toString();
	}

	/**
	 * Transforms a given date range into a machine-readable bitmask of week days.
	 *
	 * @param from starting date
	 * @param to ending date
	 * @return the bit-sequence (first bit represents monday) containing 1 for every day between the given dates, 0 otherwise
	 */
	static boolean[] getWeekDayMask(final LocalDate from, final LocalDate to) {
		final int nrWeekDays = 7;
		final int offsetFirstWeekday = DayOfWeek.MONDAY.getValue();
		final boolean[] resultArr = new boolean[nrWeekDays];

		final boolean switchDates = from.isAfter(to);
		final LocalDate tCal = (switchDates) ? from : to;
		LocalDate fCal = (switchDates) ? to : from;

		int remainingWeekDays = nrWeekDays;
		do {
			final int weekDayIndex = fCal.getDayOfWeek().getValue() - offsetFirstWeekday;
			final int arrIndex = (weekDayIndex < 0) ? (nrWeekDays - (Math.abs(weekDayIndex) % nrWeekDays)) : (weekDayIndex % nrWeekDays);
			resultArr[arrIndex] = true;

			fCal = fCal.plus(1, ChronoUnit.DAYS);
		} while (--remainingWeekDays > 0 && fCal.isBefore(tCal));

		return resultArr;
	}

	// Private methods

	private void addLink(final Map<Integer, Collection<Link>> linkList, final int nodeId, final Link link) {
		if (!linkList.containsKey(nodeId)) {
			linkList.put(nodeId, new LinkedHashSet<>());
		}

		linkList.get(nodeId).add(link);
	}

	private Link createLinkFromJSONArray(final JSONArray arr, final int startNodeId, final int endNodeId) throws SQLException {
		final Link link;
		if (Byte.parseByte(arr.get(4).toString()) == NW_MODE_CONTINUOUS) {
			link = new Link(new BigDecimal((long) arr.get(0)).intValueExact(), startNodeId, endNodeId, (double) arr.get(3));
		} else {
			link = new Link(new BigDecimal((long) arr.get(0)).intValueExact(), startNodeId, endNodeId, arr.get(5) != null ? new BigDecimal((long) arr.get(5)).shortValueExact() : 0);
		}

		return fetchGeometry(link, arr);
	}
	
	private Link createLinkFromResultSet(final ResultSet rs, final int startNodeId, final int endNodeId) throws SQLException {
		final Link link;
		if (rs.getByte("EDGE_MODE") == NW_MODE_CONTINUOUS) {
			link = new Link(rs.getInt("ID"), startNodeId, endNodeId, rs.getDouble("LENGTH"));
		} else {
			link = new Link(rs.getInt("ID"), startNodeId, endNodeId, rs.getObject("ROUTE_ID") != null ? rs.getShort("ROUTE_ID") : 0);
		}

		return fetchGeometry(link, rs);
	}

	private Link fetchGeometry(final Link link, final JSONArray arr) throws SQLException {
		if (link == null || wkbReader == null) {
			return link;
		}

		try {
			final byte[] wkb = (byte[]) arr.get(9);
			if (wkb != null) {
				link.setGeometry((LineString) wkbReader.read(wkb));
			}
		} catch (final ParseException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return link;
	}
	
	private Link fetchGeometry(final Link link, final ResultSet rs) throws SQLException {
		if (link == null || wkbReader == null) {
			return link;
		}

		try {
			final byte[] wkb = rs.getBytes("GEOMETRY");
			if (wkb != null) {
				link.setGeometry((LineString) wkbReader.read(wkb));
			}
		} catch (final ParseException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return link;
	}

	/**
	 * Gets the earliest arrival time at the target when traveled from the source.
	 *
	 * @param sId the id of the source node
	 * @param tId the id of the target node
	 * @param rIds the identifiers of the routes that should be considered
	 * @param arrServiceId the date codes for which the times are calculated
	 * @param from the earliest time for which we want to calculate the departure time
	 * @param to the latest time for which we want to calculate the departure time
	 * @throws SQLException thrown is the DB update statement can not be executed (or the prepared statement parameters could not be set)
	 * @return a resultSet with all the times that come into consideration
	 */
	private ResultSet getEarliestArrivalTime(final int sId, final int tId, final Collection<Integer> rIds, final Object[] arrServiceId, final long from, final long to) throws SQLException {
		final Object[] arrRoutes = rIds.toArray();
		final boolean isServiceIdSet = arrServiceId != null && arrServiceId.length > 0;
		int idx = 1;
		final String sql;
		switch(ConfigIsochrone.getInstance().getDBType()){
		case NEO4J:
			if (isServiceIdSet) {
				sql = String.format(query.getEarliestArrivalTimeHomo(false), preparePlaceHolders(arrRoutes.length, 4), 
						String.format(" AND n.time_a >= {"+(arrRoutes.length+4)+"} AND n.time_a <= {"+(arrRoutes.length+5)+"} AND c.service_id IN [%s]", preparePlaceHolders(arrServiceId.length, 6+arrRoutes.length)));
			} else {
				sql = String.format(query.getEarliestArrivalTimeHomo(false), preparePlaceHolders(arrRoutes.length, 4));
			}
			break;
		default:
			if (isServiceIdSet) {
				sql = String.format(query.getEarliestArrivalTimeHomo(true), preparePlaceHolders(arrRoutes.length, 0), preparePlaceHolders(arrServiceId.length, 0));
			} else {
				sql = String.format(query.getEarliestArrivalTimeHomo(false), preparePlaceHolders(arrRoutes.length, 0));
			}
		}

		final PreparedStatement statement = getPstmt(sql);
		statement.setInt(idx++, sId);
		statement.setInt(idx++, tId);
		idx = setValues(idx, statement, arrRoutes);
		statement.setLong(idx++, to);
		if (isServiceIdSet) {
			statement.setLong(idx++, from);
			statement.setLong(idx++, to);
			setValues(idx, statement, arrServiceId);
		}

		return statement.executeQuery();
	}

	/**
	 * Gets the latest departure time to start from source, to arrive at the target in time.
	 *
	 * @param sId the id of the source node
	 * @param tId the id of the target node
	 * @param rIds the identifiers of the routes that should be considered
	 * @param arrServiceId the date codes for which the times are calculated
	 * @param from the earliest time for which we want to calculate the departure time
	 * @param to the latest time for which we want to calculate the departure time
	 * @throws SQLException thrown is the DB update statement can not be executed (or the prepared statement parameters could not be set)
	 * @return a resultSet with all the times that come into consideration
	 */
	private ResultSet getLatestDepartureTime(final int sId, final int tId, final Collection<Integer> rIds, final Object[] arrServiceId, final long from, final long to) throws SQLException {
		final Object[] arrRoutes = rIds.toArray();
		final boolean isServiceIdSet = arrServiceId != null && arrServiceId.length > 0;
		int idx = 1;

		String sql;
		final PreparedStatement statement;
		switch(ConfigIsochrone.getInstance().getDBType()){
		case NEO4J:
			if (isServiceIdSet) {
				sql = String.format(query.getLatestDepartureTimeHomo(false), preparePlaceHolders(arrRoutes.length, 4), 
						String.format(" AND n.time_a >= {"+(arrRoutes.length+4)+"} AND n.time_a <= {"+(arrRoutes.length+5)+"} AND c.service_id IN [%s]", preparePlaceHolders(arrServiceId.length, 6+arrRoutes.length)));
			} else {
				sql = String.format(query.getLatestDepartureTimeHomo(false), preparePlaceHolders(arrRoutes.length, 4));
			}
			statement = getPstmt(sql);
			statement.setInt(idx++, sId);
			statement.setInt(idx++, tId);
			statement.setLong(idx++, from);
			idx = setValues(idx, statement, arrRoutes);
			break;
		default:
			if (isServiceIdSet) {
				sql = String.format(query.getLatestDepartureTimeHomo(true), preparePlaceHolders(arrRoutes.length, 0), preparePlaceHolders(arrServiceId.length, 0));
			} else {
				sql = String.format(query.getLatestDepartureTimeHomo(false), preparePlaceHolders(arrRoutes.length, 0));
			}
			statement = getPstmt(sql);
			statement.setInt(idx++, sId);
			statement.setInt(idx++, tId);
			idx = setValues(idx, statement, arrRoutes);
			statement.setLong(idx++, from);
		}

		if (isServiceIdSet) {
			statement.setLong(idx++, from);
			statement.setLong(idx++, to);
			setValues(idx, statement, arrServiceId);
		}
		return statement.executeQuery();
	}

	private Node updateNode(final Map<Integer, Node> nodes, final int nodeId, final Short nrLinks) {
		Node node = nodes.get(nodeId);
		// add node if not already in map (with nrLinks given)
		if (node == null) {
			node = new Node(nodeId, nrLinks);
			nodes.put(nodeId, node);
		}

		// update node (with nrLinks given)
		if (nrLinks != null && node.getUnvisitedLinkCount() == -1) {
			node.setNrLinks(nrLinks);
		}

		return node;
	}

	private void updateNodes(final Map<Integer, Node> nodes, final Link link, final short nrLinks) {
		final LineString geom = link.getGeometry();

		int start, end;
		Point pStart, pEnd;
		if (isIncoming) {
			start = link.getStartNode();
			end = link.getEndNode();
			pStart = (geom == null) ? null : geom.getStartPoint();
			pEnd = (geom == null) ? null : geom.getEndPoint();
		} else {
			start = link.getEndNode();
			end = link.getStartNode();
			pStart = (geom == null) ? null : geom.getEndPoint();
			pEnd = (geom == null) ? null : geom.getStartPoint();
		}

		updateNode(nodes, start, nrLinks).setGeometry(pStart);
		updateNode(nodes, end, null).setGeometry(pEnd);
	}

	// Protected static methods

	/**
	 * Prepares a number of place holders for a prepared statement.
	 * In difference to @see {@link #calculatePlaceHolders(int, int)} this uses a caching approach for small lengths
	 *
	 * @param length The number of placeholders that is needed
	 * @return a string with the placeholders plugged in to a statement.
	 */
	protected static String preparePlaceHolders(final int length, final int parameterNumber) {
		if (length <= 0) {
			return "";
		}

		if (length < PRECACHED_PLACEHOLDERS.length) {
			return PRECACHED_PLACEHOLDERS[length];
		}

		return calculatePlaceHolders(length, parameterNumber);
	}

	/**
	 * Calculates a number of place holders for a prepared statement.
	 * In difference to @see {@link #preparePlaceHolders(int, int)} this does not use any caching approach at all.
	 *
	 * @param length The number of placeholders that is needed
	 * @return a string with the placeholders plugged in to a statement.
	 */
	static String calculatePlaceHolders(final int length, int parameterNumber) {
		final StringBuilder builder;
		final int max;
		switch(ConfigIsochrone.getInstance().getDBType()){
		case NEO4J:
			if (length <= 0) {
				return "";
			}
			final int digitLength;
			final int biggerInts;
			if(parameterNumber < 10){
				digitLength = length - parameterNumber;
				biggerInts = length -digitLength;
			}
			else{
				digitLength = 0;
				biggerInts = length;
			}
			builder = new StringBuilder((4 * digitLength) + (5 * biggerInts) - 1);
			builder.append("{" + (parameterNumber++) + "}");

			max = length - 1;
			for (int i = 0; i < max; ++i) {
				builder.append(",{"+(parameterNumber++) +"}");
			}
			break;
		default:
			if (length <= 0) {
				return "";
			}

			builder = new StringBuilder((2 * length) - 1);
			builder.append("?");

			max = length - 1;
			for (int i = 0; i < max; ++i) {
				builder.append(",?");
			}
		}

		return builder.toString();
	}

	/**
	 * Sets the values in the placeholders in a prepared statement, generated by
	 * the {@link #preparePlaceHolders(int, int)} method.
	 *
	 * @see #preparePlaceHolders(int, int)
	 * @param startIdx the start index where the first object should be placed in the prepared statement
	 * @param preparedStatement the prepared statement in which the values should be inserted.
	 * @param values the values that should be inserted into the prepared statement
	 * @throws SQLException thrown if a parameter could not be set into the prepared statement
	 * @return the index of the first position that was being left unfilled by the method
	 */
	protected static final int setValues(final int startIdx, final PreparedStatement preparedStatement, final Object... values) throws SQLException {
		for (int i = 0; i < values.length; ++i) {
			preparedStatement.setObject(startIdx + i, values[i]);
		}

		return startIdx + values.length;
	}

	// Package private static methods

	static String getSeedBufferSubQuery(final Seed[] loaded, final int idx, final int srid) {
		final String pointStr = DbUtils.toPointStr(loaded[idx].getCenter());
		final String bufferStr = "ST_Buffer(ST_PointFromText('" + pointStr  + "'," + srid + ")," + loaded[idx].getRadius() + ")";
		if (idx <= 0) {
			return bufferStr;
		}

		return "ST_Union(" + bufferStr + "," + getSeedBufferSubQuery(loaded, idx - 1, srid) + ")";
	}

	// Private static methods

	private static void updateDistanceMap(final Map<Node, Long> results, final int routeId, final Node node, final long distance) {
		Long minDistance = results.get(node);
		if (minDistance == null) {
			minDistance = Long.MAX_VALUE;
		}

		if (distance < minDistance) {
			results.put(node, distance);
			node.setCheapestReachedRouteId(routeId);
		}
	}

	private String convertStreamToString(InputStream is) {

		BufferedReader reader = new BufferedReader(new InputStreamReader(is));
		StringBuilder sb = new StringBuilder();

		String line = null;
		try {
			while ((line = reader.readLine()) != null) {
				sb.append(line + "\n");
			}
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				is.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return sb.toString();
	}
}
