package it.unibz.inf.isochrone.db;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;

import java.awt.geom.Point2D;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;

/**
 * A collection of JDBC helper methods. This class is thread safe.
 */
@SuppressFBWarnings(
	value = {"SQL_PREPARED_STATEMENT_GENERATED_FROM_NONCONSTANT_STRING", "SQL_NONCONSTANT_STRING_PASSED_TO_EXECUTE"},
	justification = "Since table names can not be filled with parameters, but we have variable table names we somewhere need to do this"
)
public final class DbUtils {
	private static final String QUERY_ALL_TABLES = "SELECT tablename FROM pg_tables WHERE schemaname='public'";
	private static final String QUERY_GET_SRID = "SELECT DISTINCT ST_SRID(geometry) AS srid FROM %s WHERE geometry IS NOT NULL";

	// CHECKSTYLE:OFF EmptyBlock

	// Constructor

	private DbUtils() { }

	// Public static methods

	/**
	 * Close a <code>Connection</code>, avoid closing if null.
	 *
	 * @param conn Connection to close.
	 * @throws SQLException if a database access error occurs
	 */
	public static void close(final Connection conn) throws SQLException {
		if (conn != null) {
			conn.close();
		}
	}

	/**
	 * Close a <code>ResultSet</code>, avoid closing if null.
	 *
	 * @param rs ResultSet to close.
	 * @throws SQLException if a database access error occurs
	 */
	public static void close(final ResultSet rs) throws SQLException {
		if (rs != null) {
			rs.close();
		}
	}

	/**
	 * Close a <code>Statement</code>, avoid closing if null.
	 *
	 * @param stmt Statement to close.
	 * @throws SQLException if a database access error occurs
	 */
	public static void close(final Statement stmt) throws SQLException {
		if (stmt != null) {
			stmt.close();
		}
	}

	/**
	 * Close a <code>Connection</code>, avoid closing if null and hide
	 * any SQLExceptions that occur.
	 *
	 * @param conn Connection to close.
	 */
	public static void closeQuietly(final Connection conn) {
		try {
			close(conn);
		} catch (final SQLException e) {
			// quiet
		}
	}

	/**
	 * Close a <code>Connection</code>, <code>Statement</code> and <code>ResultSet</code>. Avoid closing if null and hide any
	 * SQLExceptions that occur.
	 *
	 * @param conn Connection to close.
	 * @param stmt Statement to close.
	 * @param rs ResultSet to close.
	 */
	public static void closeQuietly(final Connection conn, final Statement stmt, final ResultSet rs) {
		try {
			closeQuietly(rs);
		} finally {
			try {
				closeQuietly(stmt);
			} finally {
				closeQuietly(conn);
			}
		}
	}

	/**
	 * Close a <code>ResultSet</code>, avoid closing if null and hide any
	 * SQLExceptions that occur.
	 *
	 * @param rs ResultSet to close.
	 */
	public static void closeQuietly(final ResultSet rs) {
		try {
			close(rs);
		} catch (final SQLException e) {
			// quiet
		}
	}

	/**
	 * Close a <code>Statement</code>, avoid closing if null and hide
	 * any SQLExceptions that occur.
	 *
	 * @param stmt Statement to close.
	 */
	public static void closeQuietly(final Statement stmt) {
		try {
			close(stmt);
		} catch (final SQLException e) {
			// quiet
		}
	}

	// CHECKSTYLE:ON EmptyBlock

	public static Collection<String> getAllTables(final Connection connection) throws SQLException {
		if (connection == null) {
			throw new NullPointerException("Connection to database could not be initialized");
		}

		final Collection<String> result = new ArrayList<>();
		try (
			final Statement stmt = connection.createStatement();
			final ResultSet rs = stmt.executeQuery(QUERY_ALL_TABLES)
		) {
			while (rs.next()) {
				result.add(rs.getString("tablename"));
			}
		}

		return result;
	}

	public static int getSRID(final Connection connection, final String tableName) throws SQLException {
		if (connection == null) {
			throw new NullPointerException("Connection to database could not be initialized");
		}

		final String queryStr = String.format(QUERY_GET_SRID, tableName);
		int numFound = 0;
		int srid = -1;
		try (
			final Statement stmt = connection.createStatement();
			final ResultSet rs = stmt.executeQuery(queryStr)
		) {
			if (rs.next()) {
				++numFound;
				srid = rs.getInt("srid");
			}
		}

		if (numFound != 1) {
			return -1;
		}

		return srid;
	}

	public static void handleSqlException(final Connection connection, final SQLException exception) {
		exception.printStackTrace();

		try {
			connection.rollback();
		} catch (final SQLException e) {
			System.err.println("Could not rollback transaction. This will most likely cause subsequent errors!");
		}
	}

	public static String toPointStr(final Point2D p) {
		return "POINT(" + p.getX() + " " + p.getY() + ")";
	}

}
