package it.unibz.inf.isochrone.network;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

public abstract class AbstractOutput implements IOutput {
	private static final Logger LOGGER = LogManager.getLogger(AbstractOutput.class);

	// Factory method

	/**
	 * Factory method using the default constructor of an abstractOutput class.
	 *
	 * @param <T> subclass of AbstractOutput. In this type the result will be returned.
	 * @param clazz the type of the instance to create.
	 * @param parameters the parameters to pass to the constructor that is used for creation
	 * @return instance of type clazz initialized with the constructor matching the given parameters
	 */
	public static <T extends AbstractOutput> T createInstance(final Class<T> clazz, final Object... parameters) {
		final Class<?>[] parameterTypes = new Class<?>[parameters.length];
		for (int i = 0; i < parameters.length; ++i) {
			parameterTypes[i] = parameters[i].getClass();
		}

		if (LOGGER.isTraceEnabled()) {
			LOGGER.trace("Creating new output object");
			LOGGER.trace("   object class: {}", clazz.getSimpleName());
			LOGGER.trace("   using constructor with {} parameters", parameters.length);
		}

		Constructor<T> oConstructor = null;
		try {
			oConstructor = clazz.getConstructor(parameterTypes);
		} catch (NoSuchMethodException | SecurityException e) {
			// could not find requested constructor
			oConstructor = null;
		}

		if (oConstructor == null) {
			return null;
		}

		T output;
		try {
			output = oConstructor.newInstance(parameters);
		} catch (InvocationTargetException | InstantiationException | IllegalAccessException e) {
			// could not generate requested class
			return null;
		}

		return output;
	}

	// Public methods

	public boolean addLinks(final Collection<Link> links) {
		boolean changed = false;
		for (final Link link : links) {
			if (addLink(link)) {
				changed = true;
			}
		}

		return changed;
	}

	public boolean addNodes(final Collection<Node> nodes) {
		boolean changed = false;
		for (final Node node : nodes) {
			if (addNode(node)) {
				changed = true;
			}
		}

		return changed;
	}

}
