package it.unibz.inf.isochrone.network;

import com.vividsolutions.jts.geom.LineString;

/**
 * Represents a link in a network, defined by a startnode, an endnode
 * and either a linkCost or a routeId depending on whether the link is
 * continuous or discrete.
 */
public class Link extends NetworkEntity {
	private static final long serialVersionUID = 1448549673878126309L;
	private final int startNode;
	private final int endNode;
	private final boolean continuous;
	private final double length;
	private final int route;
	private double endOffset = Double.MIN_VALUE;
	private double startOffset = Double.MIN_VALUE;
	private LineString geometry;

	/**
	 * Creates a continuous link.
	 *
	 * @param id the id of the created link
	 * @param startNodeId the id of the node from which the link starts
	 * @param endNodeId the id of the node to which the link connects
	 * @param length the costs of the link (also called length)
	 */
	public Link(final int id, final int startNodeId, final int endNodeId, final double length) {
		super(id);
		this.startNode = startNodeId;
		this.endNode = endNodeId;
		this.route = 0;
		this.length = length;

		continuous = true;
	}

	/**
	 * Creates a discrete link.
	 *
	 * @param id the id of the created link
	 * @param startNodeId the id of the node from which the link starts
	 * @param endNodeId the id of the node to which the link connects
	 * @param routeId the id of the route this link belongs to
	 */
	public Link(final int id, final int startNodeId, final int endNodeId, final int routeId) {
		super(id);
		this.startNode = startNodeId;
		this.endNode = endNodeId;
		this.route = routeId;
		this.length = 0;

		continuous = false;
	}

	// Getter

	public int getEndNode() {
		return endNode;
	}

	public double getEndOffset() {
		return endOffset;
	}

	public LineString getGeometry() {
		return geometry;
	}

	public double getLength() {
		return length;
	}

	/**
	 * Returns the "opposite" node of the current Link object.
	 *
	 * @param node relative Node reference
	 * @return opposite Node of this link (or null)
	 */
	public int getOppositeOf(final Node node) {
		if (getStartNode() == node.getId()) {
			return endNode;
		}

		return startNode;
	}

	public int getRoute() {
		if (!continuous) {
			return route;
		}

		throw new RuntimeException("There is no route for a continuous link");
	}

	public int getStartNode() {
		return startNode;
	}

	public double getStartOffset() {
		return startOffset;
	}

	public boolean isContinuous() {
		return continuous;
	}

	// Setter

	public void setEndOffset(final double endOffset) {
		this.endOffset = endOffset;
	}

	public void setGeometry(final LineString geometry) {
		this.geometry = geometry;
	}

	public void setStartOffset(final double startOffset) {
		this.startOffset = startOffset;
	}

	// Public methods

	@Override
	public boolean equals(final Object other) {
		return super.equals(other);
	}

	@Override
	public int hashCode() {
		return super.hashCode();
	}

	@Override
	public String toString() {
		return "Link [id=" + getId() + ", continuous=" + continuous + ", length=" + length + "]";
	}

}
