package it.unibz.inf.isochrone.network;

import com.vividsolutions.jts.geom.Point;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Represent a node in a network.
 * The number of adjacent links is used to track how many links have already been explored.
 */
public class Node extends NetworkEntity {
	private static final long serialVersionUID = 844268548673937865L;
	private static final int MEMORY_SIZE = 32; // to be exact: 241 bits (ceiled to 32 bytes)
	private final Map<Integer, Schedule> routes;
	private Collection<Link> links;
	private int cheapestReachedRouteId = Integer.MIN_VALUE;
	private boolean closed;
	private double distance;
	private short nrUnvisitedLinks;
	private Point geometry;

	// Constructors

	public Node(final int nodeId) {
		this(nodeId, null);
	}

	public Node(final int nodeId, final Short nrLinks) {
		super(nodeId);

		closed = false;
		distance = Double.POSITIVE_INFINITY;
		nrUnvisitedLinks = (nrLinks == null) ? -1 : nrLinks;
		routes = new LinkedHashMap<>();
	}

	// Getter

	public Collection<Link> getLinks() {
		return links;
	}

	public int getCheapestReachedRouteId() {
		return cheapestReachedRouteId;
	}

	public double getDistance() {
		return distance;
	}

	@Override
	public Point getGeometry() {
		return geometry;
	}

	public short getUnvisitedLinkCount() {
		return nrUnvisitedLinks;
	}

	public Map<Integer, Schedule> getRouteSchedules() {
		return new Hashtable<>(routes);
	}

	public String getState() {
		if (isExpired()) {
			return "EXPIRED";
		}
		if (isClosed()) {
			return "CLOSED";
		}

		return "OPEN";
	}


	public boolean isClosed() {
		return closed;
	}

	public boolean isExpired() {
		return getUnvisitedLinkCount() == 0;
	}

	// Setter

	public void setLinks(final Collection<Link> links) {
		this.links = links;
	}

	public void setArrivalTime(final int routeId, final long arrivalTime) {
		getRoute(routeId).setArrivalTime(arrivalTime);
	}

	public void setCheapestReachedRouteId(final int cheapestReachedRouteId) {
		this.cheapestReachedRouteId = cheapestReachedRouteId;
	}

	public void setDepartureTime(final int routeId, final long departureTime) {
		getRoute(routeId).setDepartureTime(departureTime);
	}

	public void setDistance(final double distance) {
		this.distance = distance;
	}

	public void setGeometry(final Point geometry) {
		this.geometry = geometry;
	}

	public void setNrLinks(final short nrLinks) {
		nrUnvisitedLinks = nrLinks;
	}

	// Public methods

	public void addLink(final Link link) {
		if (links == null) {
			links = new ArrayList<>();
		}

		links.add(link);
	}

	public void addLinks(final Collection<Link> linkCollection) {
		if (links == null) {
			links = new ArrayList<>();
		}

		links.addAll(linkCollection);
	}

	public void close() {
		closed = true;
	}

	@Override
	public int compareTo(final NetworkEntity other) {
		if (other == null) {
			return -1;
		}
		if (!(other instanceof Node)) {
			return super.compareTo(other);
		}

		final Node n = (Node) other;
		if (distance < n.distance) {
			return -1;
		}
		if (distance > n.distance) {
			return 1;
		}

		return getId() - n.getId();
	}

	public boolean containsRoutes() {
		return !routes.isEmpty();
	}

	@Override
	public boolean equals(final Object obj) {
		return super.equals(obj);
	}

	@Override
	public int hashCode() {
		return super.hashCode();
	}

	@Override
	public String toString() {
		return "Node [id=" + getId() + ", distance=" + distance + "]";
	}

	public void visitLink() {
		nrUnvisitedLinks--;
	}

	// Public static methods

	/**
	 * The size of an object belonging to this class needs in memory (in bytes).
	 *
	 * @return size needed in memory (in bytes)
	 */
	public static int sizeOf() {
		return MEMORY_SIZE;
	}

	// Private methods

	private Schedule getRoute(final int routeId) {
		Schedule s = routes.get(routeId);
		if (s == null) {
			s = new Schedule();
			routes.put(routeId, s);
		}

		return s;
	}

}
