package it.unibz.inf.isochrone.util;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.regex.Pattern;

public final class ResourceHelper {
	private static final String EXTENSION_JAR = ".jar";
	private static final String EXTENSION_WAR = ".war";
	private static final String PROPERTY_CP = "java.class.path";

	private ResourceHelper() { }

	/**
	 * Gets all resources matching the given pattern from the path elements given.
	 *
	 * @param classPathElements the resource elements to search for files in
	 * @param fileNamePattern the pattern to match against
	 * @return the resources in the order they are found
	 * @throws IOException if an I/O error occurs
	 */
	public static Set<String> findFilesInResources(final Collection<URL> classPathElements, final Pattern fileNamePattern) throws IOException {
		final String[] retval = new String[classPathElements.size()];

		int i = 0;
		for (final URL element : classPathElements) {
			retval[i++] = element.getFile();
		}

		return findFilesInResources(retval, fileNamePattern);
	}

	public static Set<String> findFilesInResources(final String[] pathElements, final Pattern fileNamePattern) throws IOException {
		final Set<String> result = new HashSet<String>();
		for (final String element : pathElements) {
			final File newFile = new File(element);
			if (newFile.isDirectory()) {
				result.addAll(findResourceInDirectory(newFile, fileNamePattern));
			} else {
				result.addAll(findResourceInFile(newFile, fileNamePattern));
			}
		}

		return result;
	}

	public static Set<String> findFilesInClassPath(final Pattern fileNamePattern) throws IOException {
		final String classPath = System.getProperty(PROPERTY_CP);
		final String[] pathElements = classPath.split(File.pathSeparator);

		return findFilesInResources(pathElements, fileNamePattern);
	}

	private static Collection<String> findResourceInFile(final File resourceFile, final Pattern fileNamePattern) throws IOException {
		final Set<String> result = new HashSet<String>();
		final String absPath = resourceFile.getAbsolutePath();
		if (resourceFile.canRead() && (absPath.endsWith(EXTENSION_JAR) || absPath.endsWith(EXTENSION_WAR))) {
			try (final JarFile jarFile = new JarFile(resourceFile)) {
				final Enumeration<JarEntry> entries = jarFile.entries();
				while (entries.hasMoreElements()) {
					final JarEntry singleEntry = entries.nextElement();
					if (fileNamePattern.matcher(singleEntry.getName()).matches()) {
						result.add(jarFile.getName() + File.separatorChar + singleEntry.getName());
					}
				}
			}
		}

		return result;
	}

	private static Set<String> findResourceInDirectory(final File directory, final Pattern fileNamePattern) throws IOException {
		final Set<String> result = new HashSet<String>();
		final File[] files = directory.listFiles();
		if (files == null) {
			return result;
		}

		for (final File currentFile : files) {
			if (fileNamePattern.matcher(currentFile.getAbsolutePath()).matches()) {
				result.add(currentFile.getAbsolutePath());
			} else if (currentFile.isDirectory()) {
				result.addAll(findResourceInDirectory(currentFile, fileNamePattern));
			} else {
				result.addAll(findResourceInFile(currentFile, fileNamePattern));
			}
		}

		return result;
	}
}
