package it.unibz.inf.isochrone;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.config.ConfigIsochrone;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Location;
import it.unibz.inf.isochrone.network.Node;

import java.awt.geom.Point2D;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public final class GeoHelper {
	private static final String ALL_DENSITIES = "SELECT DISTINCT density"
		+ " FROM %s ORDER BY density";
//	private static final String POINT_BY_NODE_ID = "SELECT"
//		+ " ST_X(ST_TRANSFORM(geometry, 4326)), ST_Y(ST_TRANSFORM(geometry, 4326)) FROM %s WHERE id = ?";

	// constructor

	private GeoHelper() { }

	// Public static methods

	/**
	 * Gets the link which is nearest to the given point.
	 *
	 * @param p the point to which the nearest links are to be found
	 * @return list of link ids starting in the specified node (ordered by length descending).
	 */
	public static int findNearestLinkId(final ConfigDataset ds, final Direction direction, final Point2D p) {
		final List<Integer> ids = findNearestLinkIds(ds, direction, p);
		return (ids.isEmpty()) ? -1 : ids.get(0);
	}

	/**
	 * Gets the links which are nearest to the given point.
	 *
	 * @param p the point to which the nearest links are to be found
	 * @return list of link ids starting in the specified node (ordered by length descending).
	 */
	public static List<Integer> findNearestLinkIds(final ConfigDataset ds, final Direction direction, final Point2D p) {
		final List<Integer> result = new ArrayList<Integer>();

		try (final Database db = new Database(ds, Mode.MULTIMODAL, direction)) {
			final List<Location> locations = db.projectOnLinks(p);
			if (!locations.isEmpty()) {
				for (final Location l : locations) {
					result.add(l.getLinkId());
				}
			}
		}

		return result;
	}

	/**
	 * Gets the id of the node which is nearest to the given point.
	 *
	 * @param p the point to find the nearest node if for (in EPSG:4326 -> lat/lng)
	 * @return id of the node nearest to p (or -1 if no point could be found)
	 */
	public static int findNearestNodeId(final ConfigDataset db, final Direction direction, final Point2D p) {
		final List<Integer> ids = findNearestNodeIds(db, direction, p);
		return (ids.isEmpty()) ? -1 : ids.get(0);
	}

	/**
	 * Gets the ids of the nodes which are nearest to the given point.
	 *
	 * @param p the point to find the nearest node if for (in EPSG:4326 -> lat/lng)
	 * @return id of the node nearest to p (or -1 if no point could be found)
	 */
	public static List<Integer> findNearestNodeIds(final ConfigDataset ds, final Direction direction, final Point2D p) {
		final boolean isIncoming = direction == Direction.INCOMING;
		final List<Integer> result;
		final Map<Integer, Node> nodes;
		try (final Database db = new Database(ds, Mode.MULTIMODAL, direction)) {
			final List<Location> locations = db.projectOnLinks(p);
			final int size = locations.size() * 2;
			nodes = new HashMap<>(size);
			result = new ArrayList<>(size);

			for (final Location location : locations) {
				final Link l = db.getLink(location.getLinkId(), nodes);
				if (l != null) {
					final int id = isIncoming ? l.getStartNode() : l.getEndNode();
					nodes.get(id).setDistance(isIncoming ? location.getOffset() : l.getLength() - location.getOffset());
					result.add(id);
				}
			}
		}

		if (!nodes.isEmpty()) {
			result.sort((i, j) -> {
				final Node n0 = nodes.get(i);
				if (n0 == null) {
					return 1;
				}
				final Node n1 = nodes.get(j);
				if (n1 == null) {
					return -1;
				}

				return Double.compare(n0.getDistance(), n1.getDistance());
			});
		}

		return result;
	}

	@SuppressFBWarnings("SQL_PREPARED_STATEMENT_GENERATED_FROM_NONCONSTANT_STRING")
	public static Collection<Integer> getDensities(final ConfigDataset ds) {
		final String tableName = ds.getTablenameNodeDensity();
		final String query = String.format(ALL_DENSITIES, tableName);
		final ConfigIsochrone config = ConfigIsochrone.getInstance();

		// CHECKSTYLE:OFF MagicNumber
		Collection<Integer> result = new ArrayList<>(5000);
		// CHECKSTYLE:ON MagicNumber

		try (final Connection conn = config.getConnection();
			final PreparedStatement stmt = conn.prepareStatement(query)) {
			try (final ResultSet rs = stmt.executeQuery()) {
				while (rs.next()) {
					result.add(rs.getInt("density"));
				}
			}
		} catch (final SQLException e) {
			result = Collections.emptyList();
		}

		return result;
	}

}
