package it.unibz.inf.isochrone;

import info.monitorenter.gui.chart.Chart2D;
import info.monitorenter.gui.chart.IAxis;
import info.monitorenter.gui.chart.IAxis.AxisTitle;
import info.monitorenter.gui.chart.ITrace2D;
import info.monitorenter.gui.chart.traces.Trace2DSimple;
import it.unibz.inf.isochrone.config.ConfigIsochrone;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Stroke;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;

import javax.imageio.ImageIO;

public class SamplePlotter {
	public enum Aggregator {
		AVERAGE(MathHelper::average, "average"),
		FIRST(MathHelper::first, null),
		MEDIAN(MathHelper::median, "median"),
		NONE(null, "raw");

		private transient Function<List<Double>, Double> aggregateFn;
		private String filenamePart;

		private Aggregator(final Function<List<Double>, Double> aggregateFn, final String filenamePart) {
			this.aggregateFn = aggregateFn;
			this.filenamePart = filenamePart;
		}

		public Double apply(final List<Double> list) {
			return aggregateFn.apply(list);
		}

		public String getFilenamePart() {
			return (filenamePart == null) ? "" : "_" + filenamePart;
		}
	}
	private static final String CSV_HEADER_DURATION = "Duration";
	private static final String CSV_SEPARATOR = ",";
	private static final String FILE_ENCODING = "UTF-8";

	private final Aggregator aggregator;
	private final String fileName;
	private final Map<String, Map<SampleProperty, SampleInformation>> samples;
	private final SampleProperty sampleProperty;

	// Constructors

	public SamplePlotter(final Aggregator aggregator, final String fileName, final SampleRunner sampleRunner, final SampleProperty sampleProperty) {
		this.aggregator = aggregator;
		this.fileName = fileName;
		this.sampleProperty = sampleProperty;

		this.samples = new HashMap<>(2);
		samples.put("duration", sampleRunner.getDurationSamples());
		samples.put("nodes", sampleRunner.getNodeSamples());
	}

	public SamplePlotter(final Aggregator aggregator, final String fileName, final SampleInformation sampleDuration, final SampleProperty sampleProperty) {
		this.aggregator = aggregator;
		this.fileName = fileName;
		this.sampleProperty = sampleProperty;

		final Map<SampleProperty, SampleInformation> durationSamples = new HashMap<>(1);
		durationSamples.put(sampleProperty, sampleDuration);

		this.samples = new HashMap<>(2);
		samples.put("duration", durationSamples);
	}

	// Public methods

	public void write2Csv() {
		final Set<Entry<String, Map<SampleProperty, SampleInformation>>> sampleSet = samples.entrySet();
		for (final Entry<String, Map<SampleProperty, SampleInformation>> sampleEntry : sampleSet) {
			final String sampleType = sampleEntry.getKey();
			final Map<SampleProperty, SampleInformation> sampleMap = sampleEntry.getValue();
			if (sampleMap == null || sampleMap.isEmpty()) {
				continue;
			}

			final Set<Entry<String, Map<Long, Map<String, List<Double>>>>> datasetSet = sampleMap.get(sampleProperty).getInformation().entrySet();
			for (final Entry<String, Map<Long, Map<String, List<Double>>>> dsEntry : datasetSet) {
				final String datasetName = dsEntry.getKey();
				final Map<Long, Map<String, List<Double>>> m = dsEntry.getValue();

				if (m.size() <= 1 && m.containsKey(null)) {
					write2CsvWithoutDuration(datasetName, sampleType, m);
				} else {
					write2CsvWithDuration(datasetName, sampleType, m.entrySet());
				}
			}
		}
	}

	public void write2Png() {
		if (aggregator == Aggregator.NONE) {
			// png can only be written for aggregated values
			return;
		}

		final Set<Entry<String, Map<SampleProperty, SampleInformation>>> sampleSet = samples.entrySet();
		for (final Entry<String, Map<SampleProperty, SampleInformation>> sampleEntry : sampleSet) {
			final String sampleType = sampleEntry.getKey();
			final Map<SampleProperty, SampleInformation> sampleMap = sampleEntry.getValue();
			if (sampleMap == null || sampleMap.isEmpty()) {
				continue;
			}

			final Set<Entry<String, Map<Long, Map<String, List<Double>>>>> datasetSet = sampleMap.get(sampleProperty).getInformation().entrySet();
			for (final Entry<String, Map<Long, Map<String, List<Double>>>> dsEntry : datasetSet) {
				final String datasetName = dsEntry.getKey();
				final Set<Entry<Long, Map<String, List<Double>>>> runtimeSet = dsEntry.getValue().entrySet();
				final int nrAlgorithms = runtimeSet.iterator().next().getValue().size();

				final Map<String, ITrace2D> traceMap = new HashMap<>(nrAlgorithms);
				final Stroke stroke = new BasicStroke(3);
				final Chart2D chart = new Chart2D();

				final IAxis<?> axisX = chart.getAxisX();
				axisX.setAxisTitle(new AxisTitle(sampleTypeToTitle(sampleType)));

				final IAxis<?> axisY = chart.getAxisY();
				final String yLabel = sampleProperty.getUnit();
				if (yLabel != null) {
					axisY.setAxisTitle(new AxisTitle(yLabel));
				}
				axisY.setPaintGrid(true);

				// Create colors
				final Color[] colorArr = new Color[nrAlgorithms];
				for (int i = 0; i < nrAlgorithms; ++i) {
					// CHECKSTYLE:OFF MagicNumber
					colorArr[i] = Color.getHSBColor((float) i / (float) nrAlgorithms, 0.85f, 1.0f);
					// CHECKSTYLE:ON MagicNumber
				}

				int nrAlgorithm = 0;
				for (final Entry<Long, Map<String, List<Double>>> entryDuration : runtimeSet) {
					final Long duration = entryDuration.getKey();
					if (duration == null || duration < 0) {
						break;
					}

					final Set<Entry<String, List<Double>>> resultEntries = entryDuration.getValue().entrySet();
					for (final Entry<String, List<Double>> resultEntry : resultEntries) {
						final String algorithmName = resultEntry.getKey();
						ITrace2D trace = traceMap.get(algorithmName);
						if (trace == null) {
							trace = new Trace2DSimple(algorithmName);
							trace.setColor(colorArr[nrAlgorithm++]);
							trace.setStroke(stroke);
							traceMap.put(algorithmName, trace);

							chart.addTrace(trace);
						}

						trace.addPoint(duration, aggregator.apply(resultEntry.getValue()));
					}
				}

				try {
					ImageIO.write(chart.snapShot(), "png", getFile(datasetName, sampleType, "png"));
				} catch (final IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

	public void write2Xml() {
		final Set<Entry<String, Map<SampleProperty, SampleInformation>>> sampleSet = samples.entrySet();
		for (final Entry<String, Map<SampleProperty, SampleInformation>> sampleEntry : sampleSet) {
			final String sampleType = sampleEntry.getKey();
			final Map<SampleProperty, SampleInformation> sampleMap = sampleEntry.getValue();
			if (sampleMap == null || sampleMap.isEmpty()) {
				continue;
			}

			final Set<Entry<String, Map<Long, Map<String, List<Double>>>>> datasetSet = sampleMap.get(sampleProperty).getInformation().entrySet();
			for (final Entry<String, Map<Long, Map<String, List<Double>>>> dsEntry : datasetSet) {
				final String datasetName = dsEntry.getKey();
				final Set<Entry<Long, Map<String, List<Double>>>> runtimeSet = dsEntry.getValue().entrySet();

				final File file = getFile(datasetName, sampleType, "xml");
				try (final BufferedWriter fw = new BufferedWriter(new PrintWriter(file, FILE_ENCODING))) {
					fw.write("<?xml version=\"1.0\" encoding=\"" + FILE_ENCODING + "\"?>");
					fw.newLine();
					fw.write("<results>");
					fw.newLine();

					for (final Entry<Long, Map<String, List<Double>>> entryDuration : runtimeSet) {
						final Long duration = entryDuration.getKey();
						final String durationKey = (duration == null || duration < 0) ? "all" : String.valueOf(duration);
						fw.write("\t<time_" + durationKey + ">");
						fw.newLine();

						final Set<Entry<String, List<Double>>> resultSet = entryDuration.getValue().entrySet();
						for (final Entry<String, List<Double>> resultEntry : resultSet) {
							final String algorithmName = resultEntry.getKey();
							final List<Double> results = getAggregatedResults(resultEntry.getValue());
							for (final Double runtime : results) {
								fw.write("\t\t<" + algorithmName + ">" + runtime + "</" + algorithmName + ">");
								fw.newLine();
							}
						}

						fw.write("\t</time_" + durationKey + ">");
						fw.newLine();
					}
					fw.write("</results>");
					fw.newLine();
					fw.flush();
				} catch (final IOException e) {
					e.printStackTrace();
				}
			}
		}
	}

	public void writeAll() {
		write2Csv();
		write2Png();
		write2Xml();
	}

	// Private methods

	private List<Double> getAggregatedResults(final List<Double> results) {
		if (aggregator == Aggregator.NONE) {
			return results;
		}

		final List<Double> resultList = new ArrayList<>(1);
		resultList.add(aggregator.apply(results));

		return resultList;
	}

	private File getFile(final String datasetName, final String sampleType, final String extension) {
		final String dbType = ConfigIsochrone.getInstance().getDBType().getAbbreviation() + "_";
		final String aggregatorName = aggregator.getFilenamePart() + "_";
		final String property = sampleProperty.getFilenameSuffix();

		return new File(fileName + dbType + aggregatorName + datasetName + "_" + sampleType + property + "." + extension);
	}

	private void write2CsvWithDuration(final String datasetName, final String sampleType, final Set<Entry<Long, Map<String, List<Double>>>> runtimeSet) {
		final File file = getFile(datasetName, sampleType, "csv");

		boolean isFirstLine = true;
		try (final BufferedWriter fw = new BufferedWriter(new PrintWriter(file, FILE_ENCODING))) {
			for (final Entry<Long, Map<String, List<Double>>> entryDuration : runtimeSet) {
				final Long duration = entryDuration.getKey();
				final Map<String, List<Double>> resultMap = entryDuration.getValue();

				if (isFirstLine) {
					// if algorithms have not been set we are in first loop iteration -> write csv header
					fw.write(CSV_HEADER_DURATION + CSV_SEPARATOR);
					final Iterator<String> algorithms = resultMap.keySet().iterator();
					while (algorithms.hasNext()) {
						fw.write(algorithms.next());
						if (algorithms.hasNext()) {
							fw.write(CSV_SEPARATOR);
						}
					}

					fw.newLine();
					isFirstLine = false;
				}

				// pivoting results -> so they can be easily printed to csv format
				final Collection<List<Double>> l = resultMap.values();
				final int numSamples = (aggregator == Aggregator.NONE) ? l.iterator().next().size() : 1;
				final double[][] durationRuntimeArr = new double[numSamples][l.size()];
				int i = 0;
				for (final List<Double> allResults : l) {
					final List<Double> results = getAggregatedResults(allResults);
					int j = 0;
					for (final Double d : results) {
						durationRuntimeArr[j][i] = d;
						++j;
					}
					++i;
				}

				for (final double[] arrRuntimes : durationRuntimeArr) {
					fw.write(duration + CSV_SEPARATOR);
					for (int k = 0; k < arrRuntimes.length; ++k) {
						fw.write(String.valueOf(arrRuntimes[k]));
						if (k + 1 < arrRuntimes.length) {
							fw.write(CSV_SEPARATOR);
						}
					}
					fw.newLine();
				}
			}

			fw.flush();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	private void write2CsvWithoutDuration(final String datasetName, final String sampleType, final Map<Long, Map<String, List<Double>>> runtimeMap) {
		final Map<String, List<Double>> resultMap = (runtimeMap == null) ? null : runtimeMap.get(null);
		final Set<Entry<String, List<Double>>> resultSet = (resultMap == null) ? null : resultMap.entrySet();
		if (resultSet == null) {
			return;
		}

		final File file = getFile(datasetName, sampleType, "csv");
		try (final BufferedWriter fw = new BufferedWriter(new PrintWriter(file, FILE_ENCODING))) {
			fw.write("Raw values: ");
			fw.newLine();
			for (final Entry<String, List<Double>> runtimeEntry : resultSet) {
				final String queryName = runtimeEntry.getKey();
				final List<Double> results = runtimeEntry.getValue();
				fw.write(queryName + CSV_SEPARATOR);
				final Iterator<Double> iter = results.iterator();
				while (iter.hasNext()) {
					fw.write(String.valueOf(iter.next().doubleValue()));
					if (iter.hasNext()) {
						fw.write(CSV_SEPARATOR);
					}
				}
				fw.newLine();
			}

			fw.newLine();
			fw.write("Average:");
			fw.newLine();
			for (final Entry<String, List<Double>> runtimeEntry : resultSet) {
				fw.write(runtimeEntry.getKey() + CSV_SEPARATOR);
				fw.write(String.valueOf(MathHelper.average(runtimeEntry.getValue())));
				fw.newLine();
			}

			fw.newLine();
			fw.write("Median:");
			fw.newLine();
			for (final Entry<String, List<Double>> runtimeEntry : resultSet) {
				fw.write(runtimeEntry.getKey() + CSV_SEPARATOR);
				fw.write(String.valueOf(MathHelper.median(runtimeEntry.getValue())));
				fw.newLine();
			}

			fw.flush();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	// Private static methods

	private static String sampleTypeToTitle(final String sampleType) {
		final String title;
		switch (sampleType) {
			case "nodes":
				title = "|nodes (isochrone)|";
				break;
			case "duration":
				title = "dmax [seconds]";
				break;
			default:
				title = "unknown";
		}

		return title;
	}

}
