package it.unibz.inf.isochrone;

import it.unibz.inf.isochrone.TestHelper.AlgorithmResult;
import it.unibz.inf.isochrone.algorithm.AbstractAlgorithm;
import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics;
import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics.Statistic;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.MemoryOutput;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.Function;

import org.apache.logging.log4j.Logger;

public final class SampleRunner {
	private static final int[] RUNNER_STEP_OFFSETS = {
		60, // 1 minute (in sec)
		3600, // 1 hour (in sec)
		86400 // 1 day (in sec)
	};
	private static final int[] RUNNER_STEPS_PER_OFFSET = {
		5, // every 5 minutes
		2, // every 2 hours
		1 // every day
	};

	private final Map<String, Collection<ConfigAlgorithm>> dsConfigurations;
	private final Map<SampleProperty, SampleInformation> durationSamples;
	private final Map<SampleProperty, SampleInformation> nodeSamples;
	private final Collection<SampleProperty> sampleProperties;
	private Set<Integer> durationSteps;
	private Set<Integer> nodeSteps;

	// Builder class

	public static class SampleRunnerBuilder {
		private Map<String, ConfigAlgorithmBuilder> datasets = new HashMap<>();
		private final Collection<SampleProperty> sampleProperties;
		private Integer maxDuration;
		private Integer maxNodes;
		private boolean useCache = false;

		// Constructors

		public SampleRunnerBuilder(final Collection<SampleProperty> sampleProperties) {
			this.sampleProperties = sampleProperties;
		}

		public SampleRunnerBuilder(final SampleProperty sampleProperty, final SampleProperty... anotherProperty) {
			this.sampleProperties = new ArrayList<>((anotherProperty == null) ? 1 : (anotherProperty.length + 1));
			this.sampleProperties.add(sampleProperty);
			if (anotherProperty != null) {
				this.sampleProperties.addAll(Arrays.asList(anotherProperty));
			}
		}

		// Public methods

		public SampleRunnerBuilder addDataset(final String datasetName, final ConfigAlgorithmBuilder params) {
			datasets.put(datasetName, params);
			return this;
		}

		public SampleRunner build() {
			if (maxDuration == null && maxNodes == null) {
				throw new IllegalStateException("Either maxDuration or maxNodes (or both) has to be set!");
			}

			final SampleRunner r = new SampleRunner(this);
			if (datasets.isEmpty()) {
				final Map<String, ConfigAlgorithmBuilder> builderMap = new HashMap<>();
				// CHECKSTYLE:OFF MagicNumber
				builderMap.put("bz", new ConfigAlgorithmBuilder(TestConfiguration.DEFAULT_QUERYPOINT).dateTime(TestConfiguration.DEFAULT_DATETIME));
				// builderMap.put("it", new ConfigAlgorithmBuilder(21).dateTime(LocalDateTime.of(2011, Month.OCTOBER, 20, 15, 0, 0)));
				// builderMap.put("sf", new ConfigAlgorithmBuilder(21355).dateTime(LocalDateTime.of(2010, Month.OCTOBER, 6, 9, 50, 0)));
				// builderMap.put("st", new ConfigAlgorithmBuilder(2008706).dateTime(LocalDateTime.of(2010, Month.JULY, 14, 10, 30, 0)));
				// CHECKSTYLE:ON MagicNumber

				for (final Entry<String, ConfigAlgorithmBuilder> e : builderMap.entrySet()) {
					final String datasetName = e.getKey();
					final ConfigAlgorithmBuilder dsConfig = e.getValue();
					if (useCache) {
						dsConfig.useCache();
					}
					addDataset(datasetName, dsConfig);
				}
			}

			final Set<Entry<String, ConfigAlgorithmBuilder>> dsSet = datasets.entrySet();
			for (final Entry<String, ConfigAlgorithmBuilder> e : dsSet) {
				r.addDataset(e.getKey(), e.getValue());
			}

			return r;
		}

		// CHECKSTYLE:OFF HiddenField
		public SampleRunnerBuilder addMaxDuration(final int maxDuration) {
			this.maxDuration = maxDuration;
			return this;
		}

		public SampleRunnerBuilder addMaxNodes(final int maxNodes) {
			this.maxNodes = maxNodes;
			return this;
		}
		// CHECKSTYLE:ON HiddenField
	}

	// Constructor

	/**
	 * Initializes the time- and datasets that can be used for sampling.
	 * The information needed for plotting the samples can be retrieved from the method {@link #sample(Function, Collection, Logger)}
	 */
	private SampleRunner(final SampleRunnerBuilder builder) {
		this.dsConfigurations = new LinkedHashMap<>();
		this.durationSamples = new EnumMap<>(SampleProperty.class);
		this.nodeSamples = new EnumMap<>(SampleProperty.class);
		this.sampleProperties = builder.sampleProperties;

		if (builder.maxDuration != null) {
			initDurationSteps(builder.maxDuration);
			for (final SampleProperty p : sampleProperties) {
				durationSamples.put(p, new SampleInformation());
			}
		}

		if (builder.maxNodes != null) {
			initNodeSteps(builder.maxNodes);
			for (final SampleProperty p : sampleProperties) {
				nodeSamples.put(p, new SampleInformation());
			}
		}
	}

	// Public methods

	public Map<SampleProperty, SampleInformation> getDurationSamples() {
		return durationSamples;
	}

	public Map<SampleProperty, SampleInformation> getNodeSamples() {
		return nodeSamples;
	}

	public <T extends MemoryOutput> void sample(final Function<TestConfiguration, AlgorithmResult<T>> fn, final Collection<Class<? extends AbstractAlgorithm>> classes, final Logger logger) {
		final Set<Entry<String, Collection<ConfigAlgorithm>>> datasets = dsConfigurations.entrySet();
		for (final Class<? extends AbstractAlgorithm> clazz : classes) {
			for (final Entry<String, Collection<ConfigAlgorithm>> dataset : datasets) {
				final String dsName = dataset.getKey();
				final Collection<ConfigAlgorithm> algorithmConfigurations = dataset.getValue();

				logger.info("Executing \"{}\" for dataset \"{}\"", clazz.getSimpleName(), dsName);
				final Map<SampleProperty, Long> propertySums = new EnumMap<>(SampleProperty.class);
				for (final SampleProperty p : sampleProperties) {
					propertySums.put(p, 0L);
				}

				for (final ConfigAlgorithm config : algorithmConfigurations) {
					final long startTime = System.currentTimeMillis();
					final AlgorithmResult<T> result = fn.apply(new TestConfiguration(clazz, dsName, config));

					final AlgorithmStatistics stats = result.getStatistics();
					final long runtime = System.currentTimeMillis() - startTime;
					if (logger.isTraceEnabled()) {
						logger.trace("   runtime for duration \"{}\": {}msec", config.getDuration(), runtime);
						logger.trace("   nodes (in memory): {}", stats.getNumericStat(Statistic.LOADED_NODES_CURRENT).longValue());
						logger.trace("   nodes (in output): {}", result.getOutput().getNodes().size());
						logger.trace("   links (in output): {}", result.getOutput().getLinks().size());
					}

					for (final SampleProperty p : sampleProperties) {
						final Long pValue = (p == SampleProperty.RUNTIME) ? runtime : p.getSamplingInformation(stats);
						if (config.getDuration() != null) {
							durationSamples.get(p).add(config.getDuration(), clazz.getSimpleName(), dsName, pValue.doubleValue());
							propertySums.put(p, propertySums.get(p) + pValue);
						}
						if (config.getNodes() != null) {
							nodeSamples.get(p).add(config.getNodes(), clazz.getSimpleName(), dsName, pValue.doubleValue());
						}
					}
				}

				for (final SampleProperty p : sampleProperties) {
					durationSamples.get(p).add(-1L, clazz.getSimpleName(), dsName, propertySums.get(p).doubleValue());
				}
				logger.info("-----------------------------");
			}
		}
	}

	// Private methods

	private void addDataset(final String datasetName, final ConfigAlgorithmBuilder builder) {
		if (builder == null) {
			return;
		}
		if (dsConfigurations.containsKey(datasetName)) {
			return;
		}

		final int durationSize = (durationSteps == null) ? 0 : durationSteps.size();
		final int nodeSize = (nodeSteps == null) ? 0 : nodeSteps.size();
		final int sumSize = durationSize + nodeSize;
		if (sumSize <= 0) {
			return;
		}

		final List<ConfigAlgorithm> dsParams = new ArrayList<>(sumSize);
		if (durationSize > 0) {
			for (final Integer duration : durationSteps) {
				try {
					final ConfigAlgorithmBuilder bClones = builder.clone();
					dsParams.add(bClones.duration(duration).build());
				} catch (final CloneNotSupportedException e) {
					// in case of an exception we simly do not add the builder to the dsConfigurations
				}
			}
		}

		if (nodeSize > 0) {
			for (final Integer nodes : nodeSteps) {
				try {
					final ConfigAlgorithmBuilder bClones = builder.clone();
					dsParams.add(bClones.nodes(nodes).build());
				} catch (final CloneNotSupportedException e) {
					// in case of an exception we simly do not add the builder to the dsConfigurations
				}
			}
		}

		dsConfigurations.put(datasetName, dsParams);

		// initialize dataset here (to save some time later while testing time is measured)
		ConfigDataset.getInstance(datasetName);
	}

	private void initDurationSteps(final int maxDuration) {
		if (durationSteps != null) {
			return;
		}

		durationSteps = new TreeSet<>();
		for (int i = 0; i < RUNNER_STEP_OFFSETS.length; ++i) {
			final int offset = RUNNER_STEP_OFFSETS[i];
			final int stepWidth = RUNNER_STEPS_PER_OFFSET[i];
			for (int j = 0; j < offset; j += stepWidth) {
				final int duration = Math.max(offset, offset * j);
				if (duration > maxDuration) {
					return;
				}

				durationSteps.add(duration);
			}
		}
	}

	private void initNodeSteps(final int maxNodes) {
		if (nodeSteps != null) {
			return;
		}

		// TODO: Check if offsets here make sense (can we use time offsets for node counts?)
		nodeSteps = new TreeSet<>();
		for (int i = 0; i < RUNNER_STEP_OFFSETS.length; ++i) {
			final int offset = RUNNER_STEP_OFFSETS[i];
			final int stepWidth = RUNNER_STEPS_PER_OFFSET[i];
			for (int j = 0; j < offset; j += stepWidth) {
				final int duration = Math.max(offset, offset * j);
				if (duration > maxNodes) {
					return;
				}

				nodeSteps.add(duration);
			}
		}
	}

}
