package it.unibz.inf.isochrone;

import it.unibz.inf.isochrone.algorithm.AbstractAlgorithm;
import it.unibz.inf.isochrone.algorithm.AlgorithmException;
import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics;
import it.unibz.inf.isochrone.algorithm.MDijkstra;
import it.unibz.inf.isochrone.algorithm.Mine;
import it.unibz.inf.isochrone.algorithm.MineR;
import it.unibz.inf.isochrone.algorithm.MineRX;
import it.unibz.inf.isochrone.algorithm.MineX;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.config.ConfigIsochrone;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.db.DbUtils;
import it.unibz.inf.isochrone.network.MemoryOutput;
import it.unibz.inf.isochrone.network.NetworkEntity;

import java.lang.reflect.InvocationTargetException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.testng.Assert;

public final class TestHelper {
	private static final Comparator<Class<?>> COMPARATOR_CLASSNAME = new Comparator<Class<?>>() {
		@Override
		public int compare(final Class<?> o1, final Class<?> o2) {
			if (o1 == null) {
				return -1;
			}
			if (o2 == null) {
				return 1;
			}

			return o1.getCanonicalName().compareTo(o2.getCanonicalName());
		}
	};
	public static final class AlgorithmResult<T extends MemoryOutput> {
		private final AbstractAlgorithm algorithm;
		private final T output;

		private AlgorithmResult(final Class<T> clazz, final AbstractAlgorithm algorithm) {
			super();
			this.algorithm = algorithm;
			this.output = algorithm.getResult(clazz);
		}

		public AbstractAlgorithm getAlgorithm() {
			return algorithm;
		}

		public T getOutput() {
			return output;
		}

		public AlgorithmStatistics getStatistics() {
			return algorithm.getStatistics();
		}
	}

	// Constructor

	private TestHelper() { }

	// Public static methods

	public static boolean clearDatabaseCache() {
		final ConfigIsochrone config = ConfigIsochrone.getInstance();
		final String jdbcUrl = "jdbc:postgresql://" + config.getDBServerName() + ":" + config.getDBPort() + "/" + config.getDBName() + "?binaryTransfer=true";

		Connection conn = null;
		Statement stmt = null;
		try {
			conn = DriverManager.getConnection(jdbcUrl, config.getDBUser(), config.getDBPassword());
			// set autoCommit to true, so no transaction block is created (discard all can not run inside a transaction block)
			conn.setAutoCommit(true);

			stmt = conn.createStatement();
			stmt.execute("DISCARD ALL");
		} catch (final SQLException e) {
			return false;
		} finally {
			DbUtils.closeQuietly(stmt);
			DbUtils.closeQuietly(conn);
		}

		return true;
	}

	public static AlgorithmResult<MemoryOutput> runQuietly(final AbstractAlgorithm algorithm) {
		AlgorithmResult<MemoryOutput> result = null;
		try {
			result = run(algorithm);
		} catch (final AlgorithmException e) {
			Assert.fail(e.getClass().getSimpleName(), e);
		}

		return result;
	}

	public static AlgorithmResult<MemoryOutput> runQuietly(final TestConfiguration config) {
		AlgorithmResult<MemoryOutput> result = null;
		try {
			result = run(config);
		} catch (final AlgorithmException e) {
			Assert.fail(e.getClass().getSimpleName(), e);
		}

		return result;
	}

	public static AlgorithmResult<MemoryOutput> run(final AbstractAlgorithm algorithm) throws AlgorithmException {
		return new AlgorithmResult<>(MemoryOutput.class, algorithm);
	}

	public static AlgorithmResult<MemoryOutput> run(final TestConfiguration config) throws AlgorithmException {
		final ConfigDataset c = ConfigDataset.getInstance(config.getDataset());
		if (c == null) {
			Assert.fail("Could not create dataset from configuration");
		}

		final ConfigAlgorithm configAlgorithm = config.getAlgorithmConfiguration();
		final Database database = new Database(c, configAlgorithm.getMode(), configAlgorithm.getDirection());

		AbstractAlgorithm algorithm = null;
		try {
			algorithm = config.getAlgorithmClass().getConstructor(Database.class, ConfigAlgorithm.class).newInstance(database, configAlgorithm);
			algorithm.activateStatistics();
		} catch (final InvocationTargetException e) {
			final Throwable inException = e.getTargetException();
			if (inException instanceof AlgorithmException) {
				throw (AlgorithmException) inException;
			}

			Assert.fail(e.getMessage());
		} catch (final InstantiationException | IllegalAccessException | IllegalArgumentException | NoSuchMethodException | SecurityException e) {
			Assert.fail(e.getMessage());
		}

		final AlgorithmResult<MemoryOutput> result = new AlgorithmResult<>(MemoryOutput.class, algorithm);
		database.close();

		return result;
	}

	public static Collection<Class<? extends AbstractAlgorithm>> getAllAlgorithms() {
		return getAllAlgorithms(null);
	}

	public static Collection<Class<? extends AbstractAlgorithm>> getAllAlgorithms(final Boolean expireNodes) {
		final Set<Class<? extends AbstractAlgorithm>> algorithms = new TreeSet<>(COMPARATOR_CLASSNAME);

		algorithms.add(MDijkstra.class);
		if (expireNodes == null || !expireNodes) {
			algorithms.add(Mine.class);
			algorithms.add(MineR.class);
		}
		if (expireNodes == null || expireNodes) {
			algorithms.add(MineX.class);
			algorithms.add(MineRX.class);
		}

		return algorithms;
	}

	public static <T extends NetworkEntity> Collection<T> getSortedCollection(final Collection<T> collection) {
		final List<T> l1 = new ArrayList<>(collection);
		Collections.sort(l1);

		return l1;
	}

}
