package it.unibz.inf.isochrone.algorithm;

import it.unibz.inf.isochrone.TestConfiguration;
import it.unibz.inf.isochrone.TestHelper;
import it.unibz.inf.isochrone.TestHelper.AlgorithmResult;
import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics.Statistic;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.db.Database;
import it.unibz.inf.isochrone.network.MemoryOutput;

import java.awt.geom.Point2D;
import java.time.LocalDateTime;
import java.time.Month;
import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

public class AlgorithmTest {
	private static final Logger LOGGER = LogManager.getLogger(AlgorithmTest.class);
	private static final LocalDateTime DATETIME = LocalDateTime.of(2011, Month.JULY, 1, 12, 0, 0);
	private static final LocalDateTime DATETIME_INVALID = LocalDateTime.of(2000, Month.FEBRUARY, 1, 0, 0, 0);
	private static final int DEFAULT_DURATION = TestConfiguration.DEFAULT_MAX_DURATION;
	private static final int DURATION = 50;
	private static final Point2D.Double QUERYPOINT = TestConfiguration.DEFAULT_QUERYPOINT;
	private static final Point2D.Double QUERYPOINT_BUS = TestConfiguration.DEFAULT_QUERYPOINT_BUS;
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT = new ConfigAlgorithmBuilder(QUERYPOINT).duration(DEFAULT_DURATION).useCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT_OUTGOING = new ConfigAlgorithmBuilder(QUERYPOINT).direction(Direction.OUTGOING).duration(DEFAULT_DURATION).useCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT_UNIMODAL = new ConfigAlgorithmBuilder(QUERYPOINT).duration(DEFAULT_DURATION).mode(Mode.UNIMODAL).useCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_LIMITED_MEMORY = new ConfigAlgorithmBuilder(QUERYPOINT).duration(DEFAULT_DURATION).build();
	private static final ConfigAlgorithm TEST_CONFIG_LIMITED_MEMORY_BUS_STATION = new ConfigAlgorithmBuilder(QUERYPOINT_BUS).duration(DEFAULT_DURATION).build();
	private static final ConfigAlgorithm TEST_CONFIG_LOCAL = new ConfigAlgorithmBuilder(QUERYPOINT).dateTime(DATETIME).duration(DURATION).useCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_LOCAL_OUTGOING = new ConfigAlgorithmBuilder(QUERYPOINT).dateTime(DATETIME).direction(Direction.OUTGOING).duration(DURATION).useCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_LOCAL_UNIMODAL = new ConfigAlgorithmBuilder(QUERYPOINT).dateTime(DATETIME).duration(DURATION).mode(Mode.UNIMODAL).useCache().build();

	// Public methods

	@BeforeClass
	public void setup() {
		LOGGER.info("Starting test(s) in class \"{}\"", AlgorithmTest.class.getSimpleName());
		// loads dataset configurations (we do this before testing, so timings are correct)
		ConfigDataset.getInstance(TestConfiguration.DEFAULT_DATASET);
		LOGGER.info("-----------------------------");
	}

	// Test methods

	@Test
	public void testWithDefaultParameters() {
		LOGGER.info("Running tests with default test parameters");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			runQuietly(a, TEST_CONFIG_DEFAULT);
		}
	}

	@Test
	public void testWithDefaultOutgoingParameters() {
		LOGGER.info("Running tests with default test parameters (and outgoing direction)");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			runQuietly(a, TEST_CONFIG_DEFAULT_OUTGOING);
		}
	}

	@Test
	public void testWithDefaultUnimodalParameters() {
		LOGGER.info("Running tests with default test parameters (in unimodal mode)");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			runQuietly(a, TEST_CONFIG_DEFAULT_UNIMODAL);
		}
	}

	// FIXME: Only discrete nodes (bus stations, ...) are part of bz_nodes_density... so what if first range query happens at continuous link?
	@Test(enabled = false)
	public void testWithLimitedMemory() throws AlgorithmException {
		final long maxMemory = 20000L;

		LOGGER.info("Running tests with limited memory (MineR(X); default configuration)");
		LOGGER.info("-----------------------------");

		final long startTimeMillis = System.currentTimeMillis();
		final ConfigAlgorithm algorithmConfig = TEST_CONFIG_LIMITED_MEMORY;
		final ConfigDataset c = ConfigDataset.getInstance(TestConfiguration.DEFAULT_DATASET);
		if (c == null) {
			Assert.fail("Could not create dataset from configuration");
		}

		final Database database = new Database(c, algorithmConfig.getMode(), algorithmConfig.getDirection());
		final AlgorithmResult<MemoryOutput> resultMineR = TestHelper.run(new MineR(database, algorithmConfig, maxMemory));
		logResult(resultMineR, algorithmConfig, startTimeMillis);
		final AlgorithmResult<MemoryOutput> resultMineRX = TestHelper.run(new MineRX(database, algorithmConfig, maxMemory));
		logResult(resultMineRX, algorithmConfig, startTimeMillis);

		database.close();
	}

	@Test(enabled = false)
	public void testWithLimitedMemoryBusStation() throws AlgorithmException {
		final long maxMemory = 20000L;

		LOGGER.info("Running tests with limited memory (MineR(X); default configuration)");
		LOGGER.info("-----------------------------");

		final long startTimeMillis = System.currentTimeMillis();
		final ConfigAlgorithm algorithmConfig = TEST_CONFIG_LIMITED_MEMORY_BUS_STATION;
		final ConfigDataset c = ConfigDataset.getInstance(TestConfiguration.DEFAULT_DATASET);
		if (c == null) {
			Assert.fail("Could not create dataset from configuration");
		}

		final Database database = new Database(c, algorithmConfig.getMode(), algorithmConfig.getDirection());
		final AlgorithmResult<MemoryOutput> resultMineR = TestHelper.run(new MineR(database, algorithmConfig, maxMemory));
		logResult(resultMineR, algorithmConfig, startTimeMillis);
		final AlgorithmResult<MemoryOutput> resultMineRX = TestHelper.run(new MineRX(database, algorithmConfig, maxMemory));
		logResult(resultMineRX, algorithmConfig, startTimeMillis);

		database.close();
	}

	@Test
	public void testWithLocalParameters() {
		LOGGER.info("Running tests with local test parameters");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			runQuietly(a, TEST_CONFIG_LOCAL);
		}
	}

	@Test
	public void testWithLocalOutgoingParameters() {
		LOGGER.info("Running tests with local test parameters (and outgoing direction)");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			runQuietly(a, TEST_CONFIG_LOCAL_OUTGOING);
		}
	}

	@Test
	public void testWithLocalUnimodalParameters() {
		LOGGER.info("Running tests with local test parameters (in unimodal mode)");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			runQuietly(a, TEST_CONFIG_LOCAL_UNIMODAL);
		}
	}

	@Test(expectedExceptions = AlgorithmException.class)
	public void testInvalidConfiguration() throws AlgorithmException {
		LOGGER.info("Running test regaring AlgorithmException");
		LOGGER.info("-----------------------------");

		final ConfigAlgorithm invalidConfiguration = new ConfigAlgorithmBuilder(QUERYPOINT).dateTime(DATETIME_INVALID).duration(TestConfiguration.DEFAULT_MAX_DURATION).build();
		run(MDijkstra.class, invalidConfiguration);
	}

	@Test(expectedExceptions = IllegalStateException.class)
	public void testInvalidConfigurationBuilder() throws AlgorithmException {
		LOGGER.info("Running test regaring IllegalStateException");
		LOGGER.info("-----------------------------");

		final ConfigAlgorithm invalidConfiguration = new ConfigAlgorithmBuilder(QUERYPOINT).build();
		run(MDijkstra.class, invalidConfiguration);
	}

	@Test(expectedExceptions = AlgorithmException.class)
	public void testInvalidConfigurationNullDate() throws AlgorithmException {
		LOGGER.info("Running test regaring AlgorithmException");
		LOGGER.info("-----------------------------");

		final ConfigAlgorithm invalidConfiguration = new ConfigAlgorithmBuilder(QUERYPOINT).dateTime(null).duration(TestConfiguration.DEFAULT_MAX_DURATION).build();
		run(MDijkstra.class, invalidConfiguration);
	}

	// Private static methods

	private static void runQuietly(final Class<? extends AbstractAlgorithm> algorithmClass, final ConfigAlgorithm configAlgorithm) {
		try {
			run(algorithmClass, configAlgorithm);
		} catch (final AlgorithmException e) {
			Assert.fail(e.getClass().getSimpleName(), e);
		}
	}

	private static void run(final Class<? extends AbstractAlgorithm> algorithmClass, final ConfigAlgorithm algorithmConfig) throws AlgorithmException {
		LOGGER.info("Executing \"{}\"", algorithmClass.getSimpleName());
		final long startTimeMillis = System.currentTimeMillis();
		final TestConfiguration configTest = new TestConfiguration(algorithmClass, algorithmConfig);
		final AlgorithmResult<MemoryOutput> result = TestHelper.run(configTest);

		logResult(result, algorithmConfig, startTimeMillis);
	}

	private static void logResult(final AlgorithmResult<MemoryOutput> result, final ConfigAlgorithm algorithmConfig, final long startTimeMillis) {
		if (LOGGER.isInfoEnabled()) {
			final long runTime = System.currentTimeMillis() - startTimeMillis;
			LOGGER.info("   runtime for duration \"{}\": {}msec", algorithmConfig.getDuration(), runTime);
			LOGGER.info("   nodes (in memory): {}", result.getStatistics().getNumericStat(Statistic.LOADED_NODES_CURRENT).longValue());
			LOGGER.info("   nodes (in output): {}", result.getOutput().getNodes().size());
			LOGGER.info("   links (in output): {}", result.getOutput().getLinks().size());
			LOGGER.info("-----------------------------");
		}
	}

}
