package it.unibz.inf.isochrone.algorithm;

import it.unibz.inf.isochrone.TestConfiguration;
import it.unibz.inf.isochrone.TestHelper;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.MemoryOutput;
import it.unibz.inf.isochrone.network.Node;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

/**
 * Sophisticated performance test that executes each algorithm
 * and computes different isochrones multiple times.
 */
public final class ComparisonTest {
	// General properties
	private static final Logger LOGGER = LogManager.getLogger(ComparisonTest.class);
	private static final int MAX_RUNTIME = 18000; // 5 hours;
	private static final int[] RUNTIME_OFFSETS = {
		60,		// 1 minute (in sec)
		3600,	// 1 hour (in sec)
		86400	// 1 day (in sec)
	};
	private static final int[] RUNTIME_POINTS_PER_OFFSET = {
		5,		// every 5 minutes
		2, 		// every 2 hours
		1		// every day
	};
	private static Map<String, Collection<ConfigAlgorithm>> dsConfigurations = new LinkedHashMap<>();
	private static Set<Integer> durations;

	// Public methods

	@BeforeClass(alwaysRun = true)
	public void initTimeAndParamSet() {
		LOGGER.info("Starting test(s) in class \"{}\"", ComparisonTest.class.getSimpleName());
		initTimeSet();
		initDatasets();
		LOGGER.info("-----------------------------");
	}

	// Test methods

	@Test
	public void testAllAlgorithms() {
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a1 : algorithms) {
			for (final Class<? extends AbstractAlgorithm> a2 : algorithms) {
				if (a1.equals(a2)) {
					continue;
				}

				runForAllDatasets(a1, a2);
			}
		}
	}

	// Private static methods

	private static synchronized void addDataset(final String datasetName, final ConfigAlgorithmBuilder builder) {
		if (dsConfigurations.containsKey(datasetName) || builder == null) {
			return;
		}

		final List<ConfigAlgorithm> dsParams = new ArrayList<>(durations.size());
		for (final Integer duration : durations) {
			dsParams.add(builder.duration(duration).useCache().build());
		}

		dsConfigurations.put(datasetName, dsParams);

		// initialize dataset here (to save some time later while testing time is measured)
		ConfigDataset.getInstance(datasetName);
	}

	private static synchronized void initDatasets() {
		// CHECKSTYLE:OFF MagicNumber
		addDataset("bz", new ConfigAlgorithmBuilder(TestConfiguration.DEFAULT_QUERYPOINT).dateTime(TestConfiguration.DEFAULT_DATETIME));
		// addDataset("it", new ConfigAlgorithmBuilder(21).dateTime(LocalDateTime.of(2011, Month.OCTOBER, 20, 15, 0, 0)));
		// addDataset("sf", new ConfigAlgorithmBuilder(21355).dateTime(LocalDateTime.of(2010, Month.OCTOBER, 6, 9, 50, 0)));
		// addDataset("st", new ConfigAlgorithmBuilder(2008706).dateTime(LocalDateTime.of(2010, Month.JULY, 14, 10, 30, 0)));
		// CHECKSTYLE:ON MagicNumber
	}

	private static synchronized void initTimeSet() {
		if (durations != null) {
			return;
		}

		durations = new TreeSet<>();
		for (int i = 0; i < RUNTIME_OFFSETS.length; ++i) {
			final int offset = RUNTIME_OFFSETS[i];
			final int stepWidth = RUNTIME_POINTS_PER_OFFSET[i];
			for (int j = 0; j < offset; j += stepWidth) {
				final int runtime = Math.max(offset, offset * j);
				if (runtime > MAX_RUNTIME) {
					return;
				}

				durations.add(runtime);
			}
		}
	}

	private static void runForAllDatasets(final Class<? extends AbstractAlgorithm> class1, final Class<? extends AbstractAlgorithm> class2) {
		final String c1 = class1.getSimpleName();
		final String c2 = class2.getSimpleName();
		final Set<Entry<String, Collection<ConfigAlgorithm>>> datasets = dsConfigurations.entrySet();
		for (final Entry<String, Collection<ConfigAlgorithm>> dsEntry : datasets) {
			final String dsName = dsEntry.getKey();
			final Collection<ConfigAlgorithm> algorithmConfigurations = dsEntry.getValue();

			LOGGER.info("Comparing results for \"{}\" and \"{}\" on dataset \"{}\"", c1, c2, dsName);
			for (final ConfigAlgorithm config : algorithmConfigurations) {
				LOGGER.debug("   comparing for isochrone duration: {}", config.getDuration());
				final MemoryOutput o1 = TestHelper.runQuietly(new TestConfiguration(class1, dsName, config)).getOutput();
				final MemoryOutput o2 = TestHelper.runQuietly(new TestConfiguration(class2, dsName, config)).getOutput();
				final Collection<Node> n1 = TestHelper.getSortedCollection(o1.getNodes());
				final Collection<Node> n2 = TestHelper.getSortedCollection(o2.getNodes());
				final Collection<Link> l1 = TestHelper.getSortedCollection(o1.getLinks());
				final Collection<Link> l2 = TestHelper.getSortedCollection(o2.getLinks());
				Assert.assertEquals(n1, n2,  "Nodes from algorithm \"" + c1 + "\" are not equal to nodes from algorithm \"" + c2 + "\"");
				Assert.assertEquals(l1, l2,  "Links from algorithm \"" + c1 + "\" are not equal to links from algorithm \"" + c2 + "\"");
			}
			LOGGER.info("-----------------------------");
		}
	}

}

