package it.unibz.inf.isochrone.algorithm;

import it.unibz.inf.isochrone.TestConfiguration;
import it.unibz.inf.isochrone.TestHelper;
import it.unibz.inf.isochrone.TestHelper.AlgorithmResult;
import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics.Statistic;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.MemoryOutput;

import java.awt.geom.Point2D;
import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

public class IncrementalTest {
	private static final Logger LOGGER = LogManager.getLogger(IncrementalTest.class);
	private static final int DURATION_10 = 600;
	private static final int DURATION_15 = 900;
	private static final int DURATION_30 = 1800;
	private static final Point2D.Double QUERYPOINT = TestConfiguration.DEFAULT_QUERYPOINT;
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT_10 = new ConfigAlgorithmBuilder(QUERYPOINT).duration(DURATION_10).useCache().useIncrementalCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT_15 = new ConfigAlgorithmBuilder(QUERYPOINT).duration(DURATION_15).useCache().useIncrementalCache().build();
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT_30 = new ConfigAlgorithmBuilder(QUERYPOINT).duration(DURATION_30).useCache().useIncrementalCache().build();

	// Public methods

	@BeforeClass
	public void setup() {
		LOGGER.info("Starting test(s) in class \"{}\"", IncrementalTest.class.getSimpleName());
		// loads dataset configurations (we do this before testing, so timings are correct)
		ConfigDataset.getInstance(TestConfiguration.DEFAULT_DATASET);
		LOGGER.info("-----------------------------");
	}

	// Test methods

	@Test
	public void testDurationGrowing() {
		LOGGER.info("Running tests with growing duration");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			AbstractAlgorithm.clearCache();

			runQuietly(a, TEST_CONFIG_DEFAULT_15);
			runQuietly(a, TEST_CONFIG_DEFAULT_30);
		}
	}

	@Test
	public void testDurationShrinking() {
		LOGGER.info("Running tests with shrinking duration");
		LOGGER.info("-----------------------------");
		final Collection<Class<? extends AbstractAlgorithm>> algorithms = TestHelper.getAllAlgorithms();
		for (final Class<? extends AbstractAlgorithm> a : algorithms) {
			AbstractAlgorithm.clearCache();

			runQuietly(a, TEST_CONFIG_DEFAULT_15);
			runQuietly(a, TEST_CONFIG_DEFAULT_10);
		}
	}

	// Private static methods

	private static void runQuietly(final Class<? extends AbstractAlgorithm> algorithmClass, final ConfigAlgorithm configAlgorithm) {
		try {
			run(algorithmClass, configAlgorithm);
		} catch (final AlgorithmException e) {
			Assert.fail(e.getClass().getSimpleName(), e);
		}
	}

	private static void run(final Class<? extends AbstractAlgorithm> algorithmClass, final ConfigAlgorithm algorithmConfig) throws AlgorithmException {
		LOGGER.info("Executing \"{}\"", algorithmClass.getSimpleName());
		final long startTimeMillis = System.currentTimeMillis();
		final TestConfiguration configTest = new TestConfiguration(algorithmClass, algorithmConfig);
		final AlgorithmResult<MemoryOutput> result = TestHelper.run(configTest);

		logResult(result, algorithmConfig, startTimeMillis);
	}

	private static void logResult(final AlgorithmResult<MemoryOutput> result, final ConfigAlgorithm algorithmConfig, final long startTimeMillis) {
		if (LOGGER.isInfoEnabled()) {
			final long runTime = System.currentTimeMillis() - startTimeMillis;
			LOGGER.info("   runtime for duration \"{}\": {}msec", algorithmConfig.getDuration(), runTime);
			LOGGER.info("   nodes (in memory): {}", result.getStatistics().getNumericStat(Statistic.LOADED_NODES_CURRENT).longValue());
			LOGGER.info("   nodes (in output): {}", result.getOutput().getNodes().size());
			LOGGER.info("   links (in output): {}", result.getOutput().getLinks().size());
			LOGGER.info("-----------------------------");
		}
	}

}
