package it.unibz.inf.isochrone.algorithm;

import it.unibz.inf.isochrone.TestConfiguration;
import it.unibz.inf.isochrone.TestHelper;
import it.unibz.inf.isochrone.TestHelper.AlgorithmResult;
import it.unibz.inf.isochrone.algorithm.AlgorithmStatistics.Statistic;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.MemoryOutput;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

/**
 * Start this with jvm arguments can be used to get some more information about the isochrone algorithms (heap, cpu samples, ...).
 * This can be done by using JVM arguments while starting. Most useful arguments are:<ul>
 *   <li>-agentlib:hprof=heap=dump,format=b,file=build/reports/heapdump.hprof</li>
 *   <li>-agentlib:hprof=cpu=samples,file=build/reports/cpusamples.txt</li>
 * </ul>
 * The algorithms under test can be changes by using the TEST_ALGORITHM class member.
 * <b>Do not forget</b> to <b>TEMPORARILY</b> enable the test's enabled boolean flag (do not check this in into git repository).
 *
 * @author nikolaus.krismer@uibk.ac.at
 */
@Test(enabled = false)
public class ProfilingTest {
	private static final Logger LOGGER = LogManager.getLogger(ProfilingTest.class);
	private static final Collection<Class<? extends AbstractAlgorithm>> ALGORITHMS;
	private static final Point2D.Double QUERYPOINT = TestConfiguration.DEFAULT_QUERYPOINT;
	private static final boolean USE_CLEARING = true;
	private static final ConfigAlgorithm TEST_CONFIG_DEFAULT = new ConfigAlgorithmBuilder(QUERYPOINT).duration(TestConfiguration.DEFAULT_MAX_DURATION).build();

	// Static block

	static {
//		TEST_ALGORITHMS = TestHelper.getAllAlgorithms();
		ALGORITHMS = new ArrayList<>(1);
		ALGORITHMS.add(MineX.class);
	}

	// Public methods

	@BeforeClass
	public void setup() {
		LOGGER.info("Starting test(s) in class \"{}\"", ProfilingTest.class.getSimpleName());
		// loads dataset configurations (we do this before testing, so timings are correct)
		ConfigDataset.getInstance(TestConfiguration.DEFAULT_DATASET);
		LOGGER.info("-----------------------------");
	}

	// Test methods

	public void testWithDefaultParameters() {
		LOGGER.info("Running tests with default test parameters");
		LOGGER.info("-----------------------------");

		if (ALGORITHMS == null || ALGORITHMS.size() <= 0) {
			throw new IllegalArgumentException("No algorithms to test set");
		}

		for (final Class<? extends AbstractAlgorithm> a : ALGORITHMS) {
			run(a, TEST_CONFIG_DEFAULT);
		}

		System.gc();
	}

	// Private static methods

	private static void run(final Class<? extends AbstractAlgorithm> algorithmClass, final ConfigAlgorithm configAlgorithm) {
		LOGGER.info("Executing \"{}\"", algorithmClass.getSimpleName());
		final long startTime = System.currentTimeMillis();
		final TestConfiguration configTest = new TestConfiguration(algorithmClass, configAlgorithm);
		final AlgorithmResult<MemoryOutput> result = TestHelper.runQuietly(configTest);

		if (LOGGER.isInfoEnabled()) {
			final long runTime = System.currentTimeMillis() - startTime;
			LOGGER.info("   runtime for duration \"{}\": {}msec", configAlgorithm.getDuration(), runTime);
			LOGGER.info("   nodes (in memory): {}", result.getStatistics().getNumericStat(Statistic.LOADED_NODES_CURRENT).longValue());

			if (USE_CLEARING) {
				// free memory of algorithm to get heap dump
				result.getAlgorithm().clear();
			}

			final MemoryOutput output = result.getOutput();
			LOGGER.info("   nodes (in output): {}", output.getNodes().size());
			LOGGER.info("   links (in output): {}", output.getLinks().size());
			if (USE_CLEARING) {
				output.clear();
			}

			LOGGER.info("-----------------------------");
		}
	}
}
