package it.unibz.inf.isochrone.algorithm;

import it.unibz.inf.isochrone.GeoHelper;
import it.unibz.inf.isochrone.TestConfiguration;
import it.unibz.inf.isochrone.TestHelper;
import it.unibz.inf.isochrone.TestHelper.AlgorithmResult;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.ConfigAlgorithmBuilder;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.MemoryOutput;
import it.unibz.inf.isochrone.network.Node;

import java.awt.geom.Point2D;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.time.Month;
import java.util.Collection;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

/**
 * This checks if some well known isochrones are calculated the right way. Note
 * that there is the possibility that one of the test fails due to changes in
 * the network. So one should check for such changes (e.g. by visual isochrone
 * comparison using a GUI), before fixing a broken test.
 */
public class ReachabilityTest {
	private static final Logger LOGGER = LogManager.getLogger(ReachabilityTest.class);
	private static final LocalDateTime DATETIME = LocalDateTime.of(2011, Month.OCTOBER, 14, 9, 0, 0);
	private static final int DURATION = 900;
	private static final Point2D.Double QUERYPOINT = TestConfiguration.DEFAULT_QUERYPOINT;
	private static final BigDecimal WALKING_SPEED = BigDecimal.valueOf(1.8d);
	private static final ConfigAlgorithm TEST_CONFIG = new ConfigAlgorithmBuilder(QUERYPOINT).dateTime(DATETIME).duration(DURATION).walkingSpeed(WALKING_SPEED).useCache().build();
	private int nodeBusStationFagenstr;
	private int nodeLoretoBridge;

	// Public methods

	@BeforeClass
	public void setup() {
		LOGGER.info("Starting test(s) in class \"{}\"", ReachabilityTest.class.getSimpleName());
		// loads dataset configurations (we do this before testing, so timings are correct)
		final ConfigDataset ds = ConfigDataset.getInstance(TestConfiguration.DEFAULT_DATASET);
		initNodeIds(ds);
	}

	// Test methods

	@Test
	public void testBzBusStationLink15() {
		LOGGER.info("Checking nodes from FUB (15min incoming isochrone)");
		final boolean reachable = isReachableLink(nodeBusStationFagenstr, new TestConfiguration(MineX.class, TEST_CONFIG));
		Assert.assertTrue(reachable, "Expected link to node \"" + nodeBusStationFagenstr + "\" not in 15min incoming isochrone to FUB");
	}

	@Test
	public void testBzBusStationNode15() {
		LOGGER.info("Checking nodes from FUB (15min incoming isochrone)");
		final boolean reachable = isReachableNode(nodeBusStationFagenstr, new TestConfiguration(MineX.class, TEST_CONFIG));
		Assert.assertTrue(reachable, "Expected node \"" + nodeBusStationFagenstr + "\" not in 15min incoming isochrone to FUB");
	}

	@Test
	public void testBzWalkingLink15() {
		LOGGER.info("Checking nodes from FUB (15min incoming isochrone)");
		final boolean reachable = isReachableLink(nodeLoretoBridge, new TestConfiguration(MineX.class, TEST_CONFIG));
		Assert.assertTrue(reachable, "Expected link to node \"" + nodeLoretoBridge + "\" not in 15min incoming isochrone to FUB");
	}

	@Test
	public void testBzWalkingNode15() {
		LOGGER.info("Checking nodes from FUB (15min incoming isochrone)");
		final boolean reachable = isReachableNode(nodeLoretoBridge, new TestConfiguration(MineX.class, TEST_CONFIG));
		Assert.assertTrue(reachable, "Expected node \"" + nodeLoretoBridge + "\" not in 15min incoming isochrone to FUB");
	}

	// Private methods

	private void initNodeIds(final ConfigDataset ds) {
		// located at Loretobrücke (side of eisack river: near Josef-Mayr-Nusser-Straße)
		final Point2D pointLoretoBridge = new Point2D.Double(11.3539219377604d, 46.4945680394173d);
		// bus station in Fagenstraße (near Defregger Str.; reachable by bus route 1)
		final Point2D pointDefreggerStation = new Point2D.Double(11.3389320669492d, 46.5067191466022d);

		nodeLoretoBridge = GeoHelper.findNearestNodeId(ds, TEST_CONFIG.getDirection(), pointLoretoBridge);
		nodeBusStationFagenstr = GeoHelper.findNearestNodeId(ds, TEST_CONFIG.getDirection(), pointDefreggerStation);
	}

	private static boolean isReachableNode(final int nodeId, final TestConfiguration config) {
		final AlgorithmResult<MemoryOutput> result = TestHelper.runQuietly(config);
		final Collection<Node> resultNodes = result.getOutput().getNodes();
		LOGGER.info("   found {} reachable nodes", resultNodes.size());

		return resultNodes.stream().anyMatch(node -> node.getId() == nodeId);
	}

	private static boolean isReachableLink(final int endNodeId, final TestConfiguration config) {
		final AlgorithmResult<MemoryOutput> result = TestHelper.runQuietly(config);
		final Collection<Link> resultLinks = result.getOutput().getLinks();
		LOGGER.info("   found {} reachable links", resultLinks.size());

		return resultLinks.stream().anyMatch(node -> node.getEndNode() == endNodeId);
	}
}
