package it.unibz.inf.isochrone.db;

import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Node;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

public class DatabaseTest {
	private static final Logger LOGGER = LogManager.getLogger(DatabaseTest.class);
	private static final int EXPECTED_LINK_COUNT = 7287;
	private static final int EXPECTED_NODE_COUNT = 2687;
	private static final int LINK_ID = 1846;
	private static final int NODE_ID = 5013;
	private static final String TEST_DATASET = "bz";
	/**
	 * Private inner class that reads the network on initialization and encapsulated the database object for easier use.
	 */
	private static class InitializedDatabase {
		private final Map<Integer, Node> nodes;
		private final Map<Integer, Link> links;
		private final Database db;

		public InitializedDatabase(final Database db) {
			this.db = db;

			nodes = new HashMap<>();
			links = new LinkedHashMap<>();

			db.readNetwork(nodes, links);
		}

		public void close() {
			db.close();
		}
	}

	// Public methods

	@BeforeClass
	public void setup() {
		LOGGER.info("Starting test(s) in class \"{}\"", DatabaseTest.class.getSimpleName());
		LOGGER.info("-----------------------------");
	}

	// Test methods

	@Test
	public void testCreateDatabaseObject() {
		final ConfigDataset config = ConfigDataset.getInstance(TEST_DATASET);
		final Database db = new Database(config, Mode.MULTIMODAL, Direction.INCOMING);
		Assert.assertNotNull(db, "Database object could not be created!");
		db.close();
	}

	@Test
	public void testReadNetworkLink() {
		final Map<Integer, Node> compareNodes = new HashMap<>();
		final Map<Integer, Link> compareLinks = new HashMap<>();
		fillHashMaps(compareNodes, compareLinks);

		final InitializedDatabase db = initDb();
		Assert.assertEquals(db.links.get(LINK_ID), compareLinks.get(LINK_ID), "Links comparison");
		db.close();
	}

	@Test
	public void testReadNetworkNode() {
		final Map<Integer, Node> compareNodes = new HashMap<>();
		final Map<Integer, Link> compareLinks = new HashMap<>();
		fillHashMaps(compareNodes, compareLinks);

		final InitializedDatabase db = initDb();
		Assert.assertEquals(db.nodes.get(NODE_ID), compareNodes.get(NODE_ID), "Nodes comparison");
		db.close();
	}

	@Test
	public void testCountLinkEdges() {
		final InitializedDatabase db = initDb();
		Assert.assertEquals(db.links.size(), EXPECTED_LINK_COUNT, "Link count was unexpected");
		db.close();
	}

	@Test
	public void testCountNodeEdges() {
		final InitializedDatabase db = initDb();

		// 2687 Doesn't correspond to the number of nodes in
		// the nodes table, but the number of distinct
		// target/source node in the edges table
		Assert.assertEquals(db.nodes.size(), EXPECTED_NODE_COUNT, "Node count was unexpected");
		db.close();
	}

	// Private static methods

	private static InitializedDatabase initDb() {
		final ConfigDataset config = ConfigDataset.getInstance(TEST_DATASET);
		return new InitializedDatabase(new Database(config, Mode.MULTIMODAL, Direction.INCOMING));
	}

	private static void fillHashMaps(final Map<Integer, Node> nodeHashMap, final Map<Integer, Link> linkHashMap) {
		final Link l1 = new Link(LINK_ID, 5172, 885, 15);
		final Node n1 = new Node(NODE_ID);
		n1.addLink(l1);

		nodeHashMap.put(NODE_ID, n1);
		linkHashMap.put(LINK_ID, l1);
	}

}
