package it.unibz.inf.isochrone.db;

import it.unibz.inf.isochrone.SampleInformation;
import it.unibz.inf.isochrone.SamplePlotter;
import it.unibz.inf.isochrone.SamplePlotter.Aggregator;
import it.unibz.inf.isochrone.SampleProperty;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

public class LogParserPostgresql {
	private static final Logger LOGGER = LogManager.getLogger(LogParserPostgresql.class);
	private static final Charset CHARSET = Charset.forName("UTF-8");
	private static enum LanguageKey { DURATION, TIMEUNIT, PARSE, BIND, EXECUTE }
	private enum ParseState { NONE, BIND, PARSE, EXECUTE, UNKNOWN }
	private static final Map<String, Map<LanguageKey, String>> REGEXP_LNG;
	private final ConfigDataset dataset;
	private final File file;
	private final Long startTimeMillis;
	private LogTimings dbTimes;
	private String logLanguage = null;

	// Static block

	static {
		final EnumMap<LanguageKey, String> deMap = new EnumMap<>(LanguageKey.class);
		deMap.put(LanguageKey.DURATION, "Dauer");
		deMap.put(LanguageKey.TIMEUNIT, "ms");
		deMap.put(LanguageKey.PARSE, ".*Parsen.*");
		deMap.put(LanguageKey.BIND, ".*Binden.*");
		deMap.put(LanguageKey.EXECUTE, ".*Ausführen.*");

		final EnumMap<LanguageKey, String> enMap = new EnumMap<>(LanguageKey.class);
		enMap.put(LanguageKey.DURATION, "duration");
		enMap.put(LanguageKey.TIMEUNIT, "ms");
		enMap.put(LanguageKey.PARSE, ".*parse.*");
		enMap.put(LanguageKey.BIND, ".*bind.*");
		enMap.put(LanguageKey.EXECUTE, ".*execute.*");

		REGEXP_LNG = new HashMap<>(2);
		REGEXP_LNG.put("de", deMap);
		REGEXP_LNG.put("en", enMap);
	}

	// Constructor

	public LogParserPostgresql(final File file, final Long startTimeMillis, final ConfigDataset dataset) {
		this.dataset = dataset;
		this.file = file;
		this.startTimeMillis = startTimeMillis;
	}

	// Public methods

	public Map<String, Map<String, List<LogTimings>>> getTimings() throws IOException, SQLException {
		final Map<String, Map<String, List<LogTimings>>> datasetTimings = new HashMap<String, Map<String, List<LogTimings>>>();
		try (final BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file), CHARSET))) {
			Map<String, String> queryMap = null;
			try (final Database dbIncoming = new Database(dataset, Mode.MULTIMODAL, Direction.INCOMING)) {
				final AbstractDbQuery dbQuery = dbIncoming.getQuery();
				final BeanInfo bInfo = Introspector.getBeanInfo(AbstractDbQuery.class, Object.class);
				final PropertyDescriptor[] pDescArr = bInfo.getPropertyDescriptors();

				queryMap = new HashMap<>(pDescArr.length);
				for (final PropertyDescriptor pDesc : pDescArr) {
					final Method mGetter = pDesc.getReadMethod();
					if (mGetter.getParameterCount() == 0 && String.class.isAssignableFrom(mGetter.getReturnType())) {
						queryMap.put(pDesc.getName(), (String) mGetter.invoke(dbQuery));
					}
				}
			} catch (final IntrospectionException | InvocationTargetException | IllegalAccessException e) {
				queryMap = Collections.emptyMap();
			}

			final Map<String, List<LogTimings>> timeMap = new HashMap<String, List<LogTimings>>();
			for (final String propertyName : queryMap.keySet()) {
				timeMap.put(propertyName, new ArrayList<LogTimings>());
			}

			String line;
			while ((line = br.readLine()) != null) {
				if (isOutdatedLine(line)) {
					continue;
				}

				for (final Entry<String, String> e : queryMap.entrySet()) {
					if (ParseState.EXECUTE == parseTimes(line, convertQueryToPattern(e.getValue()))) {
						timeMap.get(e.getKey()).add(dbTimes);
						dbTimes = null;
					}
				}
			}

			datasetTimings.put(dataset.getDatasetName(), timeMap);
		}

		return datasetTimings;
	}

	// Public static methods

	public static SamplePlotter getPlotter(final File logfile, final String resultFile, final Long startTimeMillis, final ConfigDataset dataset) throws SQLException, IOException {
		final LogParserPostgresql time = new LogParserPostgresql(logfile, startTimeMillis, dataset);
		final Map<String, Map<String, List<LogTimings>>> times = time.getTimings();

		final SampleInformation sampleInformation = new SampleInformation();
		final Set<Entry<String, Map<String, List<LogTimings>>>> entries = times.entrySet();
		for (final Entry<String, Map<String, List<LogTimings>>> e : entries) {
			final String datasetName = e.getKey();
			LOGGER.info("Execution times of dataset \"{}\":", datasetName);

			final Map<String, List<LogTimings>> value = e.getValue();
			final Set<Entry<String, List<LogTimings>>> valueEntries = value.entrySet();
			for (final Entry<String, List<LogTimings>> ve: valueEntries) {
				final String queryName = ve.getKey();
				final List<LogTimings> dbTimesList = ve.getValue();
				LOGGER.info(" > query \"{}\"", queryName);

				double sum = 0;
				for (final LogTimings t : dbTimesList) {
					sampleInformation.add(queryName + "ParseTime", datasetName, t.getParseTime());
					sampleInformation.add(queryName + "BindTime", datasetName, t.getBindTime());
					sampleInformation.add(queryName + "ExecutionTime", datasetName, t.getExecutionTime());
					sampleInformation.add(queryName + "Overall", datasetName, t.getOverallTime());

					sum += t.getOverallTime();
					LOGGER.trace("   {}", t.toString());
				}

				if (dbTimesList.isEmpty()) {
					LOGGER.info("   no log entries found!");
				} else {
					LOGGER.info("   average execution time: {}ms", sum / dbTimesList.size());
				}
			}
		}

		return new SamplePlotter(Aggregator.NONE, resultFile, sampleInformation, SampleProperty.RUNTIME);
	}

	// Private methods

	private String convertQueryToPattern(final String str) {
		String query = str;

		final StringBuilder ret = new StringBuilder();
		int index0 = query.indexOf('?');
		int index1 = query.indexOf("%s");
		while (index0 >= 0 || index1 >= 0) {
			if (index0 >= 0 && (index1 < 0 || index0 < index1)) {
				ret.append(Pattern.quote(query.substring(0, index0)) + ".*");
				query = query.substring(index0 + 1, query.length());
			} else if (index1 >= 0 && (index0 < 0 || index1 < index0)) {
				ret.append(Pattern.quote(query.substring(0, index1)) + ".*");
				query = query.substring(index1 + 2, query.length());
			}

			index0 = query.indexOf('?');
			index1 = query.indexOf("%s");
		}

		ret.append(Pattern.quote(query));
		return ret.toString();
	}

	private boolean initLogLanguage(final String line, final String query) {
		for (final Entry<String, Map<LanguageKey, String>> entry : REGEXP_LNG.entrySet()) {
			final String language = entry.getKey();
			final Map<LanguageKey, String> m = entry.getValue();

			final String prefix = m.get(LanguageKey.DURATION) + ":";
			final String suffix = m.get(LanguageKey.TIMEUNIT);
			if (line.matches(".*" + prefix + ".*" + suffix + ".*" + query)) {
				LOGGER.debug("Language of log file seems to be \"{}\"", language);
				logLanguage = language;
				return true;
			}
		}

		return false;
	}

	private boolean isOutdatedLine(final String line) {
		final Pattern p = Pattern.compile("\\<(.*)\\>(.*)");
		if (startTimeMillis != null) {
			final Matcher m = p.matcher(line);
			if (m.matches()) {
				final String dateStr = m.group(1).trim();
				final SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss.SSS");
				try {
					final Date d = format.parse(dateStr);
					if (d.getTime() <= startTimeMillis) {
						return true;
					}
				} catch (final ParseException e) {
					return false;
				}
			}
		}

		return false;
	}

	private ParseState parseTimes(final String line, final String query) {
		if (logLanguage == null && !initLogLanguage(line, query)) {
			return ParseState.NONE;
		}

		final Map<LanguageKey, String> m = REGEXP_LNG.get(logLanguage);
		final String prefix = m.get(LanguageKey.DURATION) + ":";
		final String suffix = m.get(LanguageKey.TIMEUNIT);
		if (!line.matches(".*" + prefix + ".*" + suffix + ".*" + query)) {
			return ParseState.NONE;
		}

		final double t = Double.valueOf(line.substring(line.indexOf(prefix) + prefix.length(), line.indexOf(suffix) - 1));
		if (dbTimes == null) {
			dbTimes = new LogTimings();
		}

		if (line.matches(m.get(LanguageKey.PARSE))) {
			dbTimes.setParseTime(t);
			return ParseState.PARSE;
		}
		if (line.matches(m.get(LanguageKey.BIND))) {
			dbTimes.setBindTime(t);
			return ParseState.BIND;
		}
		if (line.matches(m.get(LanguageKey.EXECUTE))) {
			dbTimes.setExecutionTime(t);
			return ParseState.EXECUTE;
		}

		return ParseState.UNKNOWN;
	}

}
