package it.unibz.inf.isochrone.db;

import edu.umd.cs.findbugs.annotations.SuppressFBWarnings;
import it.unibz.inf.isochrone.GeoHelper;
import it.unibz.inf.isochrone.MathHelper;
import it.unibz.inf.isochrone.SampleInformation;
import it.unibz.inf.isochrone.SamplePlotter;
import it.unibz.inf.isochrone.SamplePlotter.Aggregator;
import it.unibz.inf.isochrone.SampleProperty;
import it.unibz.inf.isochrone.TestHelper;
import it.unibz.inf.isochrone.algorithm.AbstractAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Direction;
import it.unibz.inf.isochrone.config.ConfigAlgorithm.Mode;
import it.unibz.inf.isochrone.config.ConfigDataset;
import it.unibz.inf.isochrone.network.Link;
import it.unibz.inf.isochrone.network.Node;
import it.unibz.inf.isochrone.network.Schedule;
import it.unibz.inf.isochrone.network.Seed;

import java.awt.geom.Point2D;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.DateTimeException;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.TextStyle;
import java.time.temporal.ChronoUnit;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * You need to make some changes to your postgresql.conf for this class to work.
 * Set the following properties accordingly:
 * <ul>
 * <li>log_duration = on</li>
 * <li>log_directory= '/var/log/postgresql'</li>
 * <li>log_file_mode = 0644</li>
 * <li>log_line_prefix = '< %m >'</li>
 * <li>log_min_duration_statement = 0</li>
 * </ul>
 * Also be sure that you logging files can be accessed by the executing user
 */
@SuppressFBWarnings(
	value = {"SQL_PREPARED_STATEMENT_GENERATED_FROM_NONCONSTANT_STRING"},
	justification = "Since table names can not be filled with parameters, but we have variable table names we somewhere need to do this"
)
public class QueryRuntimeTest {
	private static final long DURATION = 3660L;
	private static final Logger LOGGER = LogManager.getLogger(QueryRuntimeTest.class);
	private static final String[] POSTGRESQL_LOGFOLDERS = {
		"/home/niko/workspace/isochrone-vm/log/postgresql",
		"/home/jenkins/jobs/isochrone-vm/workspace/log/postgresql",
		"/var/log/postgresql"
	};
	private static final Random RANDOM_GENERATOR = new Random();
	private static final int REPEAT_COUNT = 10;
	private static final String RESULTFILE_NAME = "build/reports/tests/query-runtimes";
	private static final String RESULTFILE_NAME_DBLOG = "build/reports/tests/query-runtimes-dblog";
	private static final String TEST_DATASET = "bz";
	private static final boolean USE_CACHE_CLEAR = false;
	private static final boolean USE_ONLY_QUERIES_WITH_RESULTS = true;
	private static final boolean USE_RANDOM = false;
	/** If this is set to a value greater than 0, then the first x queries will not be used in the runtime calculation result. */
	private static final int USE_VALUES_AFTER = 0;
	private static final double WALKING_SPEED = 4.5d;
	private final Map<Integer, Link> links = new LinkedHashMap<Integer, Link>();
	private final Map<Integer, Node> nodes = new HashMap<Integer, Node>();
	private final Set<Integer> routeCodes = new HashSet<Integer>();
	private final Set<Schedule> schedules = new HashSet<Schedule>();
	private final Set<Integer> serviceCodes = new HashSet<Integer>();
	private Integer[] arrLinkKeys;
	private Integer[] arrNodeKeys;
	private String currentMethodName;
	private Database dbIncoming;
	private Database dbOutgoing;
	private int linkSchlossweg2BoznerStr;
	private int nodeBusStationDominikanerplatz;
	private int nodeBusStationLongonstr;
	private int nodeLorisMusyStr2Einsteinstr;
	private int nodeReginaPacisDalmatienstr;
	private String[] placeholdersPrecached;
	private SampleInformation sampleInformation = new SampleInformation();
	private long startTimeOverall;
	private long startTimeMillis;
	private ConfigDataset ds;

	// Public methods

	@BeforeClass
	public void setup() throws SQLException {
		LOGGER.info("Starting test(s) in class \"{}\"", QueryRuntimeTest.class.getSimpleName());
		LOGGER.info("-----------------------------");
		startTimeMillis = System.currentTimeMillis();
		initDatasetAndNodeIds();

		dbOutgoing = new Database(ds, Mode.MULTIMODAL, Direction.OUTGOING);
		dbIncoming = new Database(ds, Mode.MULTIMODAL, Direction.INCOMING);
		dbIncoming.readNetwork(nodes, links);
		initSchedules();

		arrLinkKeys = links.keySet().toArray(new Integer[links.size()]);
		arrNodeKeys = nodes.keySet().toArray(new Integer[nodes.size()]);

		final int precachingLength = 10;
		placeholdersPrecached = new String[precachingLength];
		for (int i = 0; i < placeholdersPrecached.length; ++i) {
			placeholdersPrecached[i] = Database.calculatePlaceHolders(i, 0);
		}
	}

	@BeforeMethod
	public void startOverallRuntimeTest(final Method method) {
		startTimeOverall = System.nanoTime();
		currentMethodName = method.getName();
		LOGGER.info("-----------------------------");
		LOGGER.info("Executing test {}", currentMethodName);
		LOGGER.info("   dataset \"{}\"", TEST_DATASET);
		LOGGER.info("   repeating test {} times", REPEAT_COUNT);
	}

	@AfterMethod
	public void storeOverallRuntime() {
		printStatistics();
		currentMethodName = null;
		LOGGER.info("-----------------------------");
	}

	@AfterClass
	public void writeRuntimesCloseConnections() throws IOException, SQLException {
		LOGGER.debug("Tests are all finished... writing results:");
		LOGGER.debug(" - {}", RESULTFILE_NAME);
		new SamplePlotter(Aggregator.NONE, RESULTFILE_NAME, sampleInformation, SampleProperty.RUNTIME).writeAll();

		// find the log file that traced the tests (we do this by last modified date of the log files configured in POSTGRESQL_LOGFOLDERS)
		File logFile = null;
		final String logFileName = "postgresql-" + LocalDate.now().getDayOfWeek().getDisplayName(TextStyle.SHORT_STANDALONE, Locale.ENGLISH) + ".log";
		for (final String logFolderName : POSTGRESQL_LOGFOLDERS) {
			final File logFileTmp = new File(logFolderName  + File.separatorChar + logFileName);
			LOGGER.trace("Checking logFile for existance/access:");
			LOGGER.trace(" - {}", logFileTmp.getAbsoluteFile());
			if (logFileTmp.exists() && logFileTmp.canRead()) {
				if (logFile == null || logFileTmp.lastModified() > logFile.lastModified()) {
					logFile = logFileTmp;
				}
			}
		}

		if (logFile != null) {
			LOGGER.debug(" - {}", RESULTFILE_NAME_DBLOG);
			LogParserPostgresql.getPlotter(logFile, RESULTFILE_NAME_DBLOG, startTimeMillis, ds).writeAll();
		}

		dbIncoming.close();
		dbOutgoing.close();
	}

	// Test methods

	@SuppressFBWarnings("RV_RETURN_VALUE_IGNORED")
	//@Test(dependsOnMethods = "testCreateDatabaseObject", invocationCount = REPEAT_COUNT)
	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testAllEdges() throws SQLException {
		final String query = dbIncoming.getQuery().getAllEdges();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			try (final PreparedStatement stmt = dbIncoming.getPstmt(query)) {
				stmt.executeQuery();
			}
			maxCount += endTiming(startTime, i);
		}
	}

	@SuppressFBWarnings("RV_RETURN_VALUE_IGNORED")
	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testAllEdgesContinuous() throws SQLException {
		final String query = dbIncoming.getQuery().getAllEdgesContinuous();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			try (final PreparedStatement stmt = dbIncoming.getPstmt(query)) {
				stmt.executeQuery();
			}
			maxCount += endTiming(startTime, i);
		}
	}

	//TODO get a random Date from the DB entries, randomize also the year
	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testCalendar() throws SQLException {
		final String query = dbIncoming.getQuery().getCalendar();
		final LocalDate defaultDate = dbIncoming.getConfigDataset().getClientArrivalDateTime().toLocalDate();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			clearCache(dbIncoming);

			final LocalDate date = getRandomDate(defaultDate);
			final String dateMaskStr = Database.mask2String(Database.getWeekDayMask(date, date));
			final Date sqlDate = Date.valueOf(date);

			final long startTime = beginTiming(i);
//			System.out.println(query);
//			System.out.println(dateMaskStr);
//			System.out.println(sqlDate);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			// CHECKSTYLE:OFF MagicNumber
			stmt.setString(1, dateMaskStr);
			stmt.setString(2, sqlDate.toString());
			stmt.setString(3, sqlDate.toString());
			// CHECKSTYLE:ON MagicNumber
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test
	public void testCreateDatabaseObject() {
		Assert.assertNotNull(dbIncoming, "Database object could not be created!");
		Assert.assertNotNull(dbIncoming.getQuery(), "Query object could not be initialized!");
		Assert.assertNotNull(dbOutgoing, "Database object could not be created!");
		Assert.assertNotNull(dbOutgoing.getQuery(), "Query object could not be initialized!");
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testEarliestArrivalTimeHomo() throws SQLException {
		final String query = String.format(dbOutgoing.getQuery().getEarliestArrivalTimeHomo(true), Database.preparePlaceHolders(routeCodes.size(), 0), Database.preparePlaceHolders(serviceCodes.size(), 0));

		final LocalDateTime defaultFrom = dbOutgoing.getConfigDataset().getClientArrivalDateTime();
		final LocalDateTime defaultTo = defaultFrom.plus(2, ChronoUnit.HOURS);

		final Object[] routeCodeArr = this.routeCodes.toArray();
		final Object[] serviceCodeArr = this.serviceCodes.toArray();
		final long from = ConfigAlgorithm.getSecondsTillMidnight(defaultFrom);
		final long to = ConfigAlgorithm.getSecondsTillMidnight(defaultTo);
		final Schedule s = new Schedule(from, to, nodeBusStationDominikanerplatz, nodeBusStationLongonstr);

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final Schedule randomSchedule = getRandomValuesFromExistentSchedule(s);

			clearCache(dbOutgoing);

			final long startTime = beginTiming(i);

			int idx = 1;
			final PreparedStatement stmt = dbOutgoing.getPstmt(query);
			stmt.setInt(idx++, randomSchedule.getSource());
			stmt.setInt(idx++, randomSchedule.getTarget());
			idx = Database.setValues(idx, stmt, routeCodeArr);
			stmt.setLong(idx++, randomSchedule.getArrivalTime());
			stmt.setLong(idx++, randomSchedule.getDepartureTime());
			stmt.setLong(idx++, randomSchedule.getArrivalTime());
			Database.setValues(idx, stmt, serviceCodeArr);

			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testGetLink() throws SQLException {
		final String query = dbIncoming.getQuery().getLink();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int linkID = getRandomId(arrLinkKeys, linkSchlossweg2BoznerStr);
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			stmt.setInt(1, linkID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testGetLinkContinous() throws SQLException {
		final String query = dbIncoming.getQuery().getLinkContinuous();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int linkID = getRandomId(arrLinkKeys, linkSchlossweg2BoznerStr);
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			stmt.setInt(1, linkID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testGetLinksInIntersetion() throws SQLException  {
		final String query = dbIncoming.getQuery().getLinksInIntersection();
		final int srid = dbIncoming.getConfigDataset().getServerSRID();
		final int numOfIntersectionNodes = 5;

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			// intersectionNodes seed
			final Seed[] intersectionNodes = new Seed[numOfIntersectionNodes];
			for (int j = 0; j < numOfIntersectionNodes; ++j) {
				final Node randomNode = nodes.get(getRandomId(arrNodeKeys, -1));
				final Point2D seedCoordinate = dbIncoming.getCoordinates(getRandomId(arrNodeKeys, -1));
				final Point2D randomNodeCoordinate = dbIncoming.getCoordinates(randomNode.getId());
				randomNode.setDistance(seedCoordinate.distance(randomNodeCoordinate));
				intersectionNodes[j] = new Seed(seedCoordinate, getLoadableRadius(randomNode));
			}

			// node seed
			final Node randomNode = nodes.get(getRandomId(arrNodeKeys, -1));
			final Point2D seedCoordinate = dbIncoming.getCoordinates(getRandomId(arrNodeKeys, -1));
			final Point2D randomNodeCoordinate = dbIncoming.getCoordinates(randomNode.getId());
			randomNode.setDistance(seedCoordinate.distance(randomNodeCoordinate));
			final Seed seed = new Seed(seedCoordinate, getLoadableRadius(randomNode));

			final String pointString = DbUtils.toPointStr(seedCoordinate);
			final String disjunctAreaString = "st_difference(st_buffer(st_pointfromtext('" + pointString + "',"
					+ srid + ")," + seed.getRadius() + ")," + Database.getSeedBufferSubQuery(intersectionNodes, intersectionNodes.length - 1, srid) + ")";

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(String.format(query, disjunctAreaString));
			maxCount += endTiming(startTime, i, stmt.executeQuery(), false);
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testGetLinksInRange() throws SQLException  {
		final String query = dbIncoming.getQuery().getLinksInRange();
		final int srid = dbIncoming.getConfigDataset().getServerSRID();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			// node seed
			final Node randomNode = nodes.get(getRandomId(arrNodeKeys, -1));
			final Point2D seedCoordinate = dbIncoming.getCoordinates(getRandomId(arrNodeKeys, -1));
			final Point2D randomNodeCoordinate = dbIncoming.getCoordinates(randomNode.getId());
			randomNode.setDistance(seedCoordinate.distance(randomNodeCoordinate));

			final Seed seed = new Seed(seedCoordinate, getLoadableRadius(randomNode));

			LOGGER.debug("Radius to load links in is: {}", seed.getRadius());
			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			// CHECKSTYLE:OFF MagicNumber
			stmt.setString(1, DbUtils.toPointStr(seed.getCenter()));
			stmt.setInt(2, srid);
			stmt.setDouble(3, seed.getRadius());
			// CHECKSTYLE:ON MagicNumber
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testLatestDepartureTimeHomo() throws SQLException {
		final String query = String.format(dbIncoming.getQuery().getLatestDepartureTimeHomo(true), Database.preparePlaceHolders(routeCodes.size(), 0), Database.preparePlaceHolders(serviceCodes.size(), 0));

		final LocalDateTime defaultFrom = dbIncoming.getConfigDataset().getClientArrivalDateTime();
		final LocalDateTime defaultTo = defaultFrom.plus(2, ChronoUnit.HOURS);

		final Object[] routeCodeArr = this.routeCodes.toArray();
		final Object[] serviceCodeArr = this.serviceCodes.toArray();
		final long from = ConfigAlgorithm.getSecondsTillMidnight(defaultFrom);
		final long to = ConfigAlgorithm.getSecondsTillMidnight(defaultTo);
		final Schedule s = new Schedule(from, to, nodeBusStationDominikanerplatz, nodeBusStationLongonstr);

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final Schedule randomSchedule = getRandomValuesFromExistentSchedule(s);

			clearCache(dbIncoming);

			final long startTime = beginTiming(i);

			int idx = 1;
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			stmt.setInt(idx++, randomSchedule.getSource());
			stmt.setInt(idx++, randomSchedule.getTarget());
			idx = Database.setValues(idx, stmt, routeCodeArr);
			stmt.setLong(idx++, randomSchedule.getDepartureTime());
			stmt.setLong(idx++, randomSchedule.getDepartureTime());
			stmt.setLong(idx++, randomSchedule.getArrivalTime());
			Database.setValues(idx, stmt, serviceCodeArr);
			// CHECKSTYLE:ON MagicNumber

			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testLinksIncoming() throws SQLException {
		final String query = dbIncoming.getQuery().getLinks();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int nodeID = getRandomId(arrNodeKeys, nodeLorisMusyStr2Einsteinstr);
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			stmt.setInt(1, nodeID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testLinksIncomingContinuous() throws SQLException {
		final String query = dbIncoming.getQuery().getLinksContinuous();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int nodeID = getRandomId(arrNodeKeys, nodeLorisMusyStr2Einsteinstr);
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			stmt.setInt(1, nodeID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testLinksOutgoing() throws SQLException {
		final String query = dbOutgoing.getQuery().getLinks();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int nodeID = getRandomId(arrNodeKeys, nodeLorisMusyStr2Einsteinstr);
			clearCache(dbOutgoing);

			final long startTime = beginTiming(i);

			final PreparedStatement stmt = dbOutgoing.getPstmt(query);
			stmt.setInt(1, nodeID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testLinksOutgoingContinuous() throws SQLException {
		final String query = dbOutgoing.getQuery().getLinksContinuous();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int nodeID = getRandomId(arrNodeKeys, nodeLorisMusyStr2Einsteinstr);
			clearCache(dbOutgoing);

			final long startTime = beginTiming(i);

			final PreparedStatement stmt = dbOutgoing.getPstmt(query);
			stmt.setInt(1, nodeID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testNodeCoordinates() throws SQLException {
		final String query = dbIncoming.getQuery().getNodeCoordinates();

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int nodeID = getRandomId(arrNodeKeys, nodeReginaPacisDalmatienstr);
			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			stmt.setInt(1, nodeID);
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	// TODO: calculate densities with DensityGenerator in insochroen-tools or use only node fot with the densities are stored in the DB
	@Test(dependsOnMethods = "testCreateDatabaseObject")
	public void testRangeByNodeCount() throws SQLException {
		final Collection<Integer> allDensities = GeoHelper.getDensities(dbIncoming.getConfigDataset());
		final Integer[] densityArr = allDensities.toArray(new Integer[allDensities.size()]);
		final String query = dbIncoming.getQuery().getRangeByNodeCount();
		int density = densityArr[densityArr.length / 2];

		int maxCount = REPEAT_COUNT;
		for (int i = 0; i < maxCount; ++i) {
			final int nodeID = getRandomId(arrNodeKeys, nodeReginaPacisDalmatienstr);
			if (USE_RANDOM) {
				density = densityArr[getRandom(0, densityArr.length - 1)];
			}

			clearCache(dbIncoming);

			final long startTime = beginTiming(i);
			final PreparedStatement stmt = dbIncoming.getPstmt(query);
			// CHECKSTYLE:OFF MagicNumber
			stmt.setInt(1, nodeID);
			stmt.setInt(2, density);
			stmt.setInt(3, nodeID);
			stmt.setInt(4, density);
			// CHECKSTYLE:ON MagicNumber
			maxCount += endTiming(startTime, i, stmt.executeQuery());
		}
	}

	// Private methods

	private void addRuntimeEntry(final double runtime) {
		sampleInformation.add(currentMethodName, TEST_DATASET, runtime);
	}

	private long beginTiming(final int i) {
		LOGGER.trace("   execution #{}", i);
		return System.nanoTime();
	}

	private int endTiming(final long startTime, final int i) throws SQLException {
		return endTiming(startTime, i, null);
	}

	private int endTiming(final long startTime, final int i, final ResultSet r) throws SQLException {
		return endTiming(startTime, i, r, true);
	}

	private int endTiming(final long startTime, final int i, final ResultSet r, final boolean checkLoop) throws SQLException {
		final long runtime = System.nanoTime() - startTime;
		if (!shouldRuntimeBeAdded(i, r)) {
			if (checkLoop && !USE_RANDOM && i >= USE_VALUES_AFTER) {
				Assert.fail("Could not find a query result for default parameters -> this would lead to an inifinte loop!");
			}

			return 1;
		}

		LOGGER.trace("   time for run #{} (in nanosec): {}", i + 1, runtime);
		addRuntimeEntry(runtime);

		return 0;
	}

	// TODO: set also IDs for other cities => use dataProvider classes (see TestNG documentation)
	private void initDatasetAndNodeIds() throws SQLException {
		ds = ConfigDataset.getInstance(TEST_DATASET);

		if ("bz".equalsIgnoreCase(ds.getDatasetName())) {
			final Point2D pointBusStationDominikanerplatz = new Point2D.Double(11.3517095437001d, 46.4977876034576d);
			final Point2D pointBusStationLongonstr = new Point2D.Double(11.3423179140855d, 46.5012370448826d);
			final Point2D pointLorisMusyStr = new Point2D.Double(11.3289947917919d, 46.468132605146d);
			final Point2D pointReginaPacisKircheDalmatienstr = new Point2D.Double(11.3366096301632d, 46.4916534411482d);
			final Point2D pointSchlossweg = new Point2D.Double(11.3025856173558d, 46.4823344251495d);

			linkSchlossweg2BoznerStr = GeoHelper.findNearestLinkId(ds, Direction.INCOMING, pointSchlossweg);
			nodeBusStationDominikanerplatz = GeoHelper.findNearestNodeId(ds, Direction.INCOMING, pointBusStationDominikanerplatz);
			nodeBusStationLongonstr = GeoHelper.findNearestNodeId(ds, Direction.INCOMING, pointBusStationLongonstr);
			nodeLorisMusyStr2Einsteinstr = GeoHelper.findNearestNodeId(ds, Direction.INCOMING, pointLorisMusyStr);
			nodeReginaPacisDalmatienstr = GeoHelper.findNearestNodeId(ds, Direction.INCOMING, pointReginaPacisKircheDalmatienstr);
		}
	}

	/**
	 * Reads the whole Schedules form the DB.
	 * The given sets have to be initialized and will be filled up with
	 * the schedules, serviceIDs and RouteIDs.
	 *
	 * @throws SQLException if the schedules could not be read from the database
	 */
	private void initSchedules() throws SQLException {
		try (
			final PreparedStatement stmt = dbIncoming.getPstmt(dbIncoming.getQuery().getAllSchedules());
			final ResultSet rs = stmt.executeQuery();
		) {
			while (rs.next()) {
				final Schedule schedule = new Schedule();
				schedule.setArrivalTime(rs.getLong("time_a"));
				schedule.setDepartureTime(rs.getLong("time_d"));
				schedule.setRouteID(rs.getInt("route_id"));
				schedule.setServiceID(rs.getInt("service_id"));
				schedule.setSource(rs.getInt("source"));
				schedule.setTarget(rs.getInt("target"));

				if (serviceCodes != null) {
					serviceCodes.add(schedule.getServiceID());
				}

				if (routeCodes != null) {
					routeCodes.add(schedule.getRouteID());
				}

				schedules.add(schedule);
			}
		}
	}

	private void printStatistics() {
		final Map<String, Map<Long, Map<String, List<Double>>>> mapDataset = sampleInformation.getInformation();
		final Map<Long, Map<String, List<Double>>> mapDuration = (mapDataset == null) ? null : mapDataset.get(TEST_DATASET);
		final Map<String, List<Double>> mapAlgorithms = (mapDuration == null) ? null : mapDuration.get(null);
		final List<Double> runtimes = (mapAlgorithms == null) ? null : mapAlgorithms.get(currentMethodName);
		if (runtimes == null) {
			return;
		}

		LOGGER.info("   overall time (in nanosec): {}", System.nanoTime() - startTimeOverall);
		LOGGER.info("   average time (for one execution; in nanosec): {}", MathHelper.average(runtimes));
		LOGGER.info("   median time (for one execution; in nanosec): {}", MathHelper.median(runtimes));
	}

	private Schedule getRandomValuesFromExistentSchedule(final Schedule defaultValues) {
		if (USE_RANDOM || defaultValues == null)  {
			final Schedule randomValues = new Schedule();
			final Schedule[] schedulArr = this.schedules.toArray(new Schedule[this.schedules.size()]);

			final Schedule schedule = schedulArr[getRandom(schedulArr.length)];
			randomValues.setSource(schedule.getSource());
			randomValues.setTarget(schedule.getTarget());
			randomValues.setDepartureTime(schedule.getDepartureTime());
			// CHECKSTYLE:OFF MagicNumber
			randomValues.setArrivalTime(randomValues.getDepartureTime() + getRandom(1, 10000));
			// CHECKSTYLE:ON MagicNumber
			return randomValues;
		}

		return new Schedule(defaultValues);
	}

	// Private static methods

	@SuppressFBWarnings(
		value = "UC_USELESS_VOID_METHOD",
		justification = "This is only useless if the USE_CACHE_CLEAR flag is set to false"
	)
	private static void clearCache(final Database db) {
		if (!USE_CACHE_CLEAR) {
			return;
		}

		db.clearCacheMap();
		TestHelper.clearDatabaseCache();
	}

	private static int getRandom(final int bound) {
		return getRandom(0, bound);
	}

	private static int getRandom(final int lowerBound, final int upperBound) {
		return RANDOM_GENERATOR.nextInt(upperBound) + lowerBound;
	}

	private static LocalDate getRandomDate(final LocalDate defaultDate) {
		final LocalDate d = (defaultDate == null) ? LocalDate.now() : defaultDate;

		if (!USE_RANDOM && defaultDate != null) {
			return d;
		}

		LocalDate randomDate = null;
		boolean error = true;
		do {
			try {
				// CHECKSTYLE:OFF MagicNumber
				randomDate = d.withMonth(getRandom(1, 12)).withDayOfMonth(getRandom(1, 31));
				// CHECKSTYLE:ON MagicNumber
				error = false;
			} catch (final DateTimeException dte) {
				error = true;
			}
		} while (error);

		return randomDate;
	}

	private static int getRandomId(final Integer[] arr, final int defaultId) {
		if (USE_RANDOM || defaultId < 0)  {
			return arr[getRandom(arr.length)];
		}

		return defaultId;
	}

	/**
	 * Gets the maximal radius that should be used for range queries. This takes care about the available memory when calculating the radius. To do so the database is accessed (using
	 * {@link Database#getRangeByNodeCount(int, int)}) if (and only if) the memory is limited.
	 *
	 * @param node The node for which the radius should be calculated
	 * @return the radius that should be used for range queries (max == remaining distance; actual value depends on free memory)
	 */
	private static double getLoadableRadius(final Node node) {
		final double r = (DURATION - node.getDistance()) * (WALKING_SPEED / AbstractAlgorithm.CONVERSION_FACTOR_TO_METER_PER_SEC);

		// loadable radius must be positive... so we return node.getDistance as a minimum
		return Math.max(r, node.getDistance());
	}

	private static boolean shouldRuntimeBeAdded(final int i, final ResultSet r) throws SQLException {
		if (i < USE_VALUES_AFTER) {
			return false;
		}

		if (USE_ONLY_QUERIES_WITH_RESULTS && !(r == null || r.next())) {
			return false;
		}

		return true;
	}
}
